#include <iostream>
#include <QtGui>
#include <QApplication>

#include "MainWindow.h"

void noMessageOutput(QtMsgType type, const char *msg) {
	/* do nothing, i.e. discard Qt messages completely */
}

MainWindow::MainWindow() : QMainWindow(0, Qt::Window | Qt::WindowTitleHint | Qt::WindowCloseButtonHint | Qt::CustomizeWindowHint) {
	setupUi(this);

	this->setMaximumWidth(690);

	Push1_Tip = "Open a DEVIL RAW floppy image.";
	Push2_Tip = "Open a ZX Spectrum .TAP file.";
	Push3_1_Tip = "<nobr>Extract (multiple) selected catalog items to a .TAP file.</nobr><br>The .TAP file will be created and saved to a location of your choice under a name you choose.";
	Push3_2_Tip = "<nobr>Extract the (single) selected data block in the currently open .TAP file</nobr><br>to a .BIN file. Only one block must be selected in the list. Only the actual data is extracted, starting with the first byte after the flag byte and ending with the last byte before the checksum byte. In other words, the first two bytes containing the block length, the flag byte and the checksum byte will NOT be extracted.";
	Push4_1_Tip = "<nobr>Add new catalog entries (data blocks) to currently open floppy image</nobr><br>from an existing .TAP file containing Spectrum data. The .TAP file must NOT contain any non-writable headerless blocks. The floppy image must have a large enough continuous free space to accomodate all data in the .TAP file";
	Push4_2_Tip = "<nobr>Insert a new headerless block into the currently open .TAP file</nobr><br>from the contents of a binary (.bin) file of max. 65535 bytes. Max. 1 block must be selected in the list. If one block is selected, the data will be inserted as a headerless block right before it. If no block is selected, the data will be appended to the end of the .TAP file. You will have to choose a value for the flag byte of the new headerless block.";
	Push5_1_Tip = "<nobr>Rename currently selected catalog entry.</nobr><br>Only one entry must be selected in the list.";
	Push5_2_Tip = "<nobr>Rename currently selected header block.</nobr><br>Only one block must be selected in the list and it must be of type 'Head'.";
	Push6_1_Tip = "Delete all selected catalog entries.";
	Push6_2_Tip = "<nobr>Delete the (single) selected data block in the currently open .TAP file</nobr><br>Only one block must be selected in the list.";
	Push7_1_Tip = "<nobr>Change attributes for all (multiple) selected catalog entries.</nobr>";
	Push7_2_Tip = "<nobr>Create and insert a header block in the currently open .TAP file</nobr><br>before the (single) selected headerless block in the list. Only one block must be selected in the list and it must be a headerless block with a flag value = FF.";
	Push8_Tip = "<nobr>Create a new (empty, formatted) DEVIL floppy image file.</nobr>";
	Push9_Tip = "<nobr>Create a new .TAP file from the contents of a binary (.bin) file.</nobr><br>The new .TAP file will contain only one data block, which will be headerless and will have a flag byte with a value you choose.";
	Push10_Tip = "<nobr>Interleave a non~interleaved floppy image (.dimg) existing in your filesystem.</nobr><br>The interleaved floppy image (.img) will be saved in a location of your choice.";
	Push11_Tip = "<nobr>Deinterleave the currently open floppy image.</nobr><br>The deinterleaved floppy image (.dimg) will be saved in a location of your choice.";

	Check1_Tip = "<nobr>Display details in console</nobr><br>during operations.";
	Check2_Tip = "<nobr>Save a detailed operation log</nobr><br>during data extraction.";
	Check3_Tip = "<nobr>Display low-level details in console</nobr><br>during certain operations.";
	Check4_Tip = "<nobr>Display/Hide legend info box</nobr>";

	StatB1_1_tip = "<nobr>Number of selected data blocks/</nobr><br>Number of continuous ranges selected";
	StatB1_2_tip = "<nobr>Number of selected data blocks</nobr>";
	StatB2_1_tip = "<nobr>Binary size (in B) / disk size (in KB)</nobr><br>for all selected data blocks";
	StatB2_2_tip = "<nobr>Offset (address) from beginning of .TAP file</nobr><br>to beginning of selected block, in hex";

	nr_of_ranges = 0;

	QToolTip::add( pushButton1, Push1_Tip );
	QToolTip::add( pushButton2, Push2_Tip );
	QToolTip::add( pushButton3, Push3_1_Tip );
	QToolTip::add( pushButton4, Push4_1_Tip );
	QToolTip::add( pushButton5, Push5_1_Tip );
	QToolTip::add( pushButton6, Push6_1_Tip );
	QToolTip::add( pushButton7, Push7_1_Tip );
	QToolTip::add( pushButton8, Push8_Tip );
	QToolTip::add( pushButton9, Push9_Tip );
	QToolTip::add( pushButton10, Push10_Tip );
	QToolTip::add( pushButton11, Push11_Tip );

	QToolTip::add( checkBox1, Check1_Tip );
	QToolTip::add( checkBox2, Check2_Tip );
	QToolTip::add( checkBox3, Check3_Tip );
	QToolTip::add( checkBox4, Check4_Tip );

	pushButton3->setEnabled(false);
	pushButton4->setEnabled(false);
	pushButton5->setEnabled(false);
	pushButton6->setEnabled(false);
	pushButton7->setEnabled(false);
	pushButton11->setEnabled(false);
	checkBox4->setEnabled(false);

//	done here as demo for learning how to connect menu actions to slots in a program
	connect( menuExtractAction,	SIGNAL( activated() ), this, SLOT( extractFiles() ) );
	connect( menuInsertAction,	SIGNAL( activated() ), this, SLOT( insertFiles() ) );
	connect( menuRenameAction,	SIGNAL( activated() ), this, SLOT( renameFile() ) );
	connect( menuDeleteAction,	SIGNAL( activated() ), this, SLOT( deleteFiles() ) );
	connect( menuChange_AttrsAction,SIGNAL( activated() ), this, SLOT( changeAttrs() ) );

	frame1->hide();
	frame5->hide();

	textLabel3s = new QLabel( " 00 / 00 ", this, "textLabel3s" );
	textLabel3s->setAlignment(Qt::AlignHCenter);
	textLabel3s->setMinimumSize(textLabel3s->sizeHint());
	textLabel3s->clear();
//	QToolTip::add( textLabel3s, "<nobr>Number of selected data blocks /</nobr><br>Number of continuous ranges selected" );

	textLabel4s = new QLabel( " 000000 B / 000 KB ", this, "textLabel4s" );
	textLabel4s->setAlignment(Qt::AlignHCenter);
	textLabel4s->setMinimumSize(textLabel4s->sizeHint());
	textLabel4s->clear();
//	QToolTip::add( textLabel4s, "<nobr>Binary size (in B) / disk size (in KB)</nobr><br>for all selected data blocks" );

	textLabel5s = new QLabel( this, "textLabel5s" );
//	textLabel5s->setText( QString("%1").arg(qApp->argv()[0]) );	// TEST: displays the full path of the executable in the status bar (needs #include <qapplication.h>)
	textLabel5s->setText( "" );
	textLabel5s->setAlignment(Qt::AlignLeft);

	textLabel3s->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
	textLabel4s->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);
	textLabel5s->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);

//	this->statusBar()->setLayout(new QBoxLayout(QBoxLayout::LeftToRight, this->statusBar()));
//	this->statusBar()->layout()->addWidget(textLabel3s);
//	this->statusBar()->layout()->addWidget(textLabel4s);
//	this->statusBar()->layout()->addWidget(textLabel5s);
	this->statusBar()->addWidget(textLabel3s);
	this->statusBar()->addWidget(textLabel4s);
	this->statusBar()->addWidget(textLabel5s, 1);
	this->statusBar()->layout()->setSpacing(1);

	listBox1->clear();

	QPalette palette;

	QBrush brush(QColor(255, 255, 0, 255));		// list text
	brush.setStyle(Qt::SolidPattern);
	QBrush brush1(QColor(0, 0, 0, 255));		// list background
	brush1.setStyle(Qt::SolidPattern);
	QBrush brush2(QColor(128, 0, 0, 255));		// selected list elements background
	brush2.setStyle(Qt::SolidPattern);

	palette.setBrush(QPalette::Active, QPalette::Text, brush);
	palette.setBrush(QPalette::Active, QPalette::Base, brush1);
	palette.setBrush(QPalette::Active, QPalette::Highlight, brush2);
	palette.setBrush(QPalette::Inactive, QPalette::Text, brush);
	palette.setBrush(QPalette::Inactive, QPalette::Base, brush1);
	palette.setBrush(QPalette::Inactive, QPalette::Highlight, brush2);

	listBox1->setPalette(palette);

	// ------------- specifically set font for tooltips --------------
	QFont toolTipFont( "Bitstream Vera Serif", 9, QFont::Normal );
	QToolTip::setFont( toolTipFont );
	// ------------- specifically set font for all application ---------------------------
	// (this will only affect widgets who don't have their individual fonts specified,
	//  like the individual menu items and various dialogs like File Open, File Save etc.)
	QFont appFont( "Bitstream Vera Sans", 9, QFont::Normal );
	QApplication::setFont ( appFont, TRUE );
// ===========================================================================

/******* no need for this, it's already taken care of by ui_MainWindow.h ********
 *	connect(pushButton1, SIGNAL(clicked()), this, SLOT(openImage()));
	connect(pushButton2, SIGNAL(clicked()), this, SLOT(openTap()));
	connect(pushButton3, SIGNAL(clicked()), this, SLOT(extractFiles()));
	connect(pushButton4, SIGNAL(clicked()), this, SLOT(insertFiles()));
	connect(pushButton5, SIGNAL(clicked()), this, SLOT(renameFile()));
	connect(pushButton6, SIGNAL(clicked()), this, SLOT(deleteFiles()));
	connect(pushButton7, SIGNAL(clicked()), this, SLOT(changeAttrs()));
	connect(pushButton8, SIGNAL(clicked()), this, SLOT(createImage()));
	connect(pushButton9, SIGNAL(clicked()), this, SLOT(createTape()));
	connect(pushButton10, SIGNAL(clicked()), this, SLOT(interleave()));
	connect(pushButton11, SIGNAL(clicked()), this, SLOT(deinterleave()));
	connect(checkBox1, SIGNAL(stateChanged(int)), this, SLOT(setVerbose(int)));
	connect(checkBox2, SIGNAL(stateChanged(int)), this, SLOT(setLogOperation(int)));
	connect(checkBox3, SIGNAL(stateChanged(int)), this, SLOT(setDebugMode(int)));
	connect(checkBox4, SIGNAL(stateChanged(int)), this, SLOT(showLegend(int)));
	connect(listBox1, SIGNAL(selectionChanged()), this, SLOT(updateSelectionSize()));
	connect(menuOpen_Floppy_ImageAction, SIGNAL(activated()), this, SLOT(openImage()));
	connect(menuOpen_TAP_FileAction, SIGNAL(activated()), this, SLOT(openTap()));
	connect(menuNew_FloppyAction, SIGNAL(activated()), this, SLOT(createImage()));
	connect(menuNew_TAPAction, SIGNAL(activated()), this, SLOT(createTape()));
	connect(menuInterleaveAction, SIGNAL(activated()), this, SLOT(interleave()));
	connect(menuDeInterleaveAction, SIGNAL(activated()), this, SLOT(deinterleave()));
	connect(helpHelpAction, SIGNAL(activated()), this, SLOT(help()));	*/
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton1->clicked() ("Open Floppy Image").
// It pops up the builtin file open dialog (see the Qt Reference Documentation,
// the QFileDialog Class Reference, the "Static Public Members" section)
void MainWindow::openImage() {

	int ret_code;
	QString s0 = "Loaded disk image '";				// part of status bar message to display when finished
	QString s;							// floppy image file full path

	// =============== get the full path of floppy image to open =================================================
	s = QFileDialog::getOpenFileName(
		0,
		"Choose a floppy image to open",
		"/home/knoppix",
		"Floppy disk images (*.img)",
		0,
		QFileDialog::DontUseNativeDialog);
	if (s==QString::null) {return;}

	QFileInfo f(s);							// get all info for the floppy image file in QFileInfo object f

	// =============== create the error message popup in case we need it or simply for testing ===================
	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("File Error!");

	// =============== checking disk image file name and size ====================================================
	if (DEVIL::dotcount(f.fileName().toAscii().constData()) > 1) {		// floppy image file should only have 1 dot in its name
		printf("'%s': No multiple dots allowed in disk image file name.\n\n", f.fileName().toAscii().constData());
		w4.textLabel1->setText(QString("").sprintf("'%s':\nNo multiple dots allowed\nin disk image file name.", f.fileName().toAscii().constData()));
		msg.exec();
		return;
	}
	w4.textLabel1->setText(QString("").sprintf("'%s':\nDisk image file name\nmust have the extension .img.", f.fileName().toAscii().constData()));
	if (DEVIL::dotcount(f.fileName().toAscii().constData()) == 0) {		// floppy image file name should end in ".img" so should contain 1 dot
		printf("'%s': Disk image file name must have the extension .img.\n\n", f.fileName().toAscii().constData());
		msg.exec();
		return;
	}
	if ( (DEVIL::dotcount(f.fileName().toAscii().constData()) == 1) &&		// floppy image file name should end in ".img"
	!(strstr(f.fileName().toAscii().constData(), ".img") && (DEVIL::dotpos(f.fileName().toAscii().constData()) == strlen(f.fileName().toAscii().constData())-4)) ) {
		printf("'%s': Disk image file name must have the extension .img.\n\n", f.fileName().toAscii().constData());
		msg.exec();
		return;
	}
	DEVIL::test_stream = fopen (s.toAscii().constData(), "r");
	if (DEVIL::test_stream == NULL) {
		printf ("Error %d opening disk image file in R/O mode: %m\n\n", errno);
		w4.textLabel1->setText(QString("").sprintf("Error %d opening disk image file in R/O mode:\n%m", errno));
		msg.exec();
		return;
	} else {fclose(DEVIL::test_stream);}
	if (f.size() != 737280) {
		printf ("Invalid DEVIL disk image file specified - file size = %u\n\n", (unsigned int)f.size());
		w4.textLabel1->setText(QString("").sprintf("Invalid DEVIL disk image file:\nfile size = %u", (unsigned int)f.size()));
		msg.exec();
		return;
	}

	// =============== store the full path of floppy image in DEVIL::diskimg_file_name ===========================
	if (DEVIL::diskimg_file_name_allocated) {free(DEVIL::diskimg_file_name);} else {DEVIL::diskimg_file_name_allocated=true;}
	asprintf(&DEVIL::diskimg_file_name,"%s",s.toAscii().constData());		// this allocates memory to diskimg_file_name

	// =============== release memory allocated to tap_file_name if any ==========================================
	if (DEVIL::tap_file_name_allocated) {free(DEVIL::tap_file_name);}
	DEVIL::tap_file_name_allocated=false;

	// =============== delete previous floppy items if present, before displaying new image contents =============
	if (DEVIL::nr_files > 0) {
//		for (int i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}			// replacing these 2 lines with a single one
//		for (unsigned int i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {free(DEVIL::file_list[i]);}				// file_list[] is a const size array (DEVIL catalog has fized size of 77)
		for (short i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back(); free(DEVIL::file_list[i]);}
		DEVIL::nr_files = 0; DEVIL::nr_hless = 0;
	}
	// =============== delete previous tape items if present, before displaying new image contents ===============
	if (DEVIL::tap_idx > 0) {
		for (int i=DEVIL::tap_idx-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}
		free(DEVIL::tap_index);
		DEVIL::tap_idx = 0;
	}
	// ===========================================================================================================
	ret_code = DEVIL::devildirdump();		// this will set global vars DEVIL::file_list[], DEVIL::nr_files and DEVIL::nr_hless

	if (ret_code) {		// ============ if any error occured during DEVIL::devildirdump() ====================
		switch (ret_code) {
			case 1: w4.textLabel1->setText("Error opening floppy image"); break;
			case 2: w4.textLabel1->setText("Invalid disk image: \nDirectory entries cannot be full of zeros"); break;
			case 4: w4.textLabel1->setText("Invalid disk image: \nInvalid data type byte in dir entry"); break;
			case 5: w4.textLabel1->setText("Invalid disk image: \nInvalid attribute byte in dir entry"); break;
			case 6: w4.textLabel1->setText("Invalid disk image: \nInvalid unused byte in dir entry"); break;
		}
		msg.exec();

		// if error occured accessing current floppy image, we need to close it, clear the file list and revert to startup state
		// =============== release memory allocated to diskimg_file_name if any =====================
		if (DEVIL::diskimg_file_name_allocated) {free(DEVIL::diskimg_file_name);}
		DEVIL::diskimg_file_name_allocated=false;
		// =============== delete previous floppy items if present ==================================
		if (DEVIL::nr_files > 0U) {
			if ((ret_code!=2) && (ret_code!=4) && (ret_code!=5) && (ret_code!=6)) {
//				for (int i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}	// replacing these 2 lines with a single one
//				for (unsigned char i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {free(DEVIL::file_list[i]);}
				for (short i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back(); free(DEVIL::file_list[i]);}
			}
			DEVIL::nr_files = 0; DEVIL::nr_hless = 0;
		}
		frame1->hide();
		frame5->hide();

		// =============== set context-dependent button tooltips ====================================
		QToolTip::add( pushButton3, Push3_1_Tip );
		QToolTip::add( pushButton4, Push4_1_Tip );
		QToolTip::add( pushButton5, Push5_1_Tip );
		QToolTip::add( pushButton6, Push6_1_Tip );
		QToolTip::add( pushButton7, Push7_1_Tip );
		QToolTip::add( textLabel3s, StatB1_1_tip );
		QToolTip::add( textLabel4s, StatB2_1_tip );

		// =============== disable context-dependent buttons and clear status bar ===================
		textLabel1->setText("  No    Type    Name       Length  Start VarStart VarN Attributes     ");

		pushButton3->setText("&Extract To .TAP");	menuExtractAction->setText("Extract selected catalog entries to a .TAP file");	menuExtractAction->setMenuText("&Extract To .TAP");
		pushButton4->setText("&Insert From .TAP");	menuInsertAction->setText("Insert catalog entries from a .TAP file");		menuInsertAction->setMenuText("&Insert From .TAP");
		pushButton5->setText("&Rename Cat. Entry");	menuRenameAction->setText("Rename (one single) selected catalog entry");	menuRenameAction->setMenuText("&Rename Catalog Entry");
		pushButton6->setText("&Delete Cat. Entries");	menuDeleteAction->setText("Delete multiple selected catalog entries");		menuDeleteAction->setMenuText("&Delete Catalog Entries");
		pushButton7->setText("Change &Attributes");	menuChange_AttrsAction->setText("Change disk attributes for multiple selected catalog entries");	menuChange_AttrsAction->setMenuText("Change &Attributes");

		pushButton3->setEnabled(false);		menuExtractAction->setEnabled(FALSE);		// disable "Extract To .TAP" button & menu
		pushButton4->setEnabled(false);		menuInsertAction->setEnabled(FALSE);		// disable "Insert From .TAP" button & menu
		pushButton5->setEnabled(false);		menuRenameAction->setEnabled(FALSE);		// disable "Rename Block" button & menu
		pushButton6->setEnabled(false);		menuDeleteAction->setEnabled(FALSE);		// disable "Delete Block(s)" button & menu
		pushButton7->setEnabled(false);		menuChange_AttrsAction->setEnabled(FALSE);	// disable "Change Attributes" button & menu
		pushButton11->setEnabled(false);	menuDeInterleaveAction->setEnabled(FALSE);	// disable "De-Interleave Image" button & menu

		checkBox4->setEnabled(false); checkBox4->setChecked(0);	// disable "Legend" checkbox
		textLabel3s->clear();textLabel4s->clear();
		updateStatusMessage("");
		// =============== reset title bar caption ==================================================
		this->setCaption("CoBra DEVIL Floppy Image Utility");
		// ==========================================================================================
		return;
	}			// ===================================================================================

	// =============== populate listbox with list of files in image loaded =======================================
	for (unsigned int i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {
//		listItems[i] = new Q3ListBoxText(listBox1, QString(DEVIL::file_list[i]).remove(QRegExp("[\\n\\t\\r]")) );
		listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::file_list[i]).remove(QRegExp("[\\n\\t\\r]")) ));
	}

	this->setMaximumWidth(690);

	// =============== set context-dependent button tooltips ====================================
	QToolTip::add( pushButton3, Push3_1_Tip );
	QToolTip::add( pushButton4, Push4_1_Tip );
	QToolTip::add( pushButton5, Push5_1_Tip );
	QToolTip::add( pushButton6, Push6_1_Tip );
	QToolTip::add( pushButton7, Push7_1_Tip );
	QToolTip::add( textLabel3s, StatB1_1_tip );
	QToolTip::add( textLabel4s, StatB2_1_tip );

	// =============== set context-dependent GUI text ============================================================
	textLabel1->setText("  No    Type    Name       Length  Start VarStart VarN Attributes     ");
	textLabel3s->clear();textLabel4s->clear();

	pushButton3->setText("&Extract To .TAP");	menuExtractAction->setText("Extract selected catalog entries to a .TAP file");	menuExtractAction->setMenuText("&Extract To .TAP");
	pushButton4->setText("&Insert From .TAP");	menuInsertAction->setText("Insert catalog entries from a .TAP file");		menuInsertAction->setMenuText("&Insert From .TAP");
	pushButton5->setText("&Rename Cat. Entry");	menuRenameAction->setText("Rename (one single) selected catalog entry");	menuRenameAction->setMenuText("&Rename Catalog Entry");
	pushButton6->setText("&Delete Cat. Entries");	menuDeleteAction->setText("Delete multiple selected catalog entries");		menuDeleteAction->setMenuText("&Delete Catalog Entries");
	pushButton7->setText("Change &Attributes");	menuChange_AttrsAction->setText("Change disk attributes for multiple selected catalog entries");	menuChange_AttrsAction->setMenuText("Change &Attributes");

	pushButton3->setEnabled(false);	menuExtractAction->setEnabled(FALSE);
	pushButton4->setEnabled(true);	menuInsertAction->setEnabled(TRUE);
	pushButton5->setEnabled(false);	menuRenameAction->setEnabled(FALSE);
	pushButton6->setEnabled(false);	menuDeleteAction->setEnabled(FALSE);
	pushButton7->setEnabled(false);	menuChange_AttrsAction->setEnabled(FALSE);
	pushButton11->setEnabled(true);	menuDeInterleaveAction->setEnabled(TRUE);

	checkBox4->setEnabled(true); checkBox4->setChecked(1);

	// =============== updating info box below GUI file list =====================================================
	textLabel10->setText(QString("").sprintf("%2d",77-DEVIL::free_dir_recs));
	textLabel11->setText(QString("").sprintf("%6d",DEVIL::total_size));
	textLabel12->setText(QString("").sprintf("%6d",DEVIL::nr_recs*9*1024));
	textLabel13->setText(QString("").sprintf("%2d",DEVIL::free_dir_recs));
	textLabel14->setText(QString("").sprintf("%2d",DEVIL::nr_files));
	textLabel15->setText(QString("").sprintf("%3d",DEVIL::free_dir_recs*9));
	for (unsigned char i=0; i<77; i++) {fragview[i]=(DEVIL::alloc_dir[i]?'I':'-');} fragview[77]='\0';
	textLabel26->setText(fragview);
	frame5->hide();		// hide .TAP info box
	frame1->show();		// show DEVIL info box

	// =============== print status message in console and status bar and update title bar =======================
	printf ("Loaded disk image file '%s'\n\n",s.toAscii().constData());
	this->updateStatusMessage( s0+f.fileName()+"'" );
	this->setCaption(QString("%1 - '%2'").arg("CoBra DEVIL Floppy Image Utility",f.fileName()));

//	msg.textLabel1->setText(QString("nr_hless = %1").arg(DEVIL::nr_hless)); msg.exec();		// test point

	return;

}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton2->clicked() ("Open .TAP File").
// It pops up the builtin file save dialog (see the Qt Reference Documentation,
// the QFileDialog Class Reference, the "Static Public Members" section)
void MainWindow::openTap() {

	int ret_code;
	QString s0 = "Loaded .TAP file '";				// part of status bar message to display when finished

	// =============== get the full path of .TAP file to open ====================================================
	QString s = QFileDialog::getOpenFileName(
		0,
		"Choose a .TAP file to open",
		"/home/knoppix",
		"Tape files (*.tap)",
		0,
		QFileDialog::DontUseNativeDialog);
	if (s==QString::null) {return;}

	QFileInfo f(s);							// get all info for the .TAP file in QFileInfo object f

	// =============== create the error message popup in case we need it or simply for testing ===================
	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("File Error!");

	// =============== checking .TAP file name and size ==========================================================
	if (DEVIL::dotcount(f.fileName().toAscii().constData()) > 1) {		// .TAP file should only have 1 dot in its name
		printf("'%s': No multiple dots allowed in .TAP file name.\n\n", f.fileName().toAscii().constData());
		w4.textLabel1->setText(QString("").sprintf("'%s':\nNo multiple dots allowed\nin .TAP file name.", f.fileName().toAscii().constData()));
		msg.exec();
		return;
	}
	w4.textLabel1->setText(QString("").sprintf("'%s':\n.TAP file name\nmust have the extension .tap", f.fileName().toAscii().constData()));
	if (DEVIL::dotcount(f.fileName().toAscii().constData()) == 0) {		// .TAP file name should end in ".tap" so should contain 1 dot
		printf("'%s': .TAP file name must have the extension .tap\n\n", f.fileName().toAscii().constData());
		msg.exec();
		return;
	}
	if ( (DEVIL::dotcount(f.fileName().toAscii().constData()) == 1) &&		// .TAP file name should end in ".tap"
	!(strstr(f.fileName().toAscii().constData(), ".tap") && (DEVIL::dotpos(f.fileName().toAscii().constData()) == strlen(f.fileName().toAscii().constData())-4)) ) {
		printf("'%s': .TAP file name must have the extension .tap\n\n", f.fileName().toAscii().constData());
		msg.exec();
		return;
	}
	DEVIL::test_stream = fopen (s.toAscii().constData(), "r");
	if (DEVIL::test_stream == NULL) {
		printf ("Error %d opening .TAP file in R/O mode: %m\n\n", errno);
		w4.textLabel1->setText(QString("").sprintf("Error %d opening .TAP file in R/O mode:\n%m", errno));
		msg.exec();
		return;
	} else {fclose(DEVIL::test_stream);}

	// =============== delete previous floppy items if present, before displaying new tape contents =============
	if (DEVIL::nr_files > 0) {
//		for (int i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}			// replacing these 2 lines with a single one
//		for (unsigned int i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {free(DEVIL::file_list[i]);}				// file_list[] is a const size array (DEVIL catalog has fized size of 77)
		for (short i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {printf("delete listItems[%d]\n", i); delete listItems[i]; listItems.pop_back(); free(DEVIL::file_list[i]);}
		DEVIL::nr_files = 0; DEVIL::nr_hless = 0;
	}
	// =============== delete previous tape items if present, before displaying new tape contents ===============
	if (DEVIL::tap_idx > 0) {
		for (int i=DEVIL::tap_idx-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}
		free(DEVIL::tap_index);
		DEVIL::tap_idx = 0;
	}
	// ===========================================================================================================
	ret_code = DEVIL::devilfilewrite(s.toAscii().constData(), 0, 0);		// this will set global vars DEVIL::tap_list[], DEVIL::tap_index[] and DEVIL::tap_idx

	if (ret_code==-1 || ret_code==-2) {		// ============ if any error occured during DEVIL::devildirdump() ====================
		switch (ret_code) {
			case -1: w4.textLabel1->setText("TAP file too short"); break;
			case -2: w4.textLabel1->setText("Error opening TAP file"); break;
		}
		msg.exec();

		// if error occured opening .TAP file, we need to close it, clear the file list and revert to startup state
		// =============== release memory allocated to tap_file_name if any =========================
		if (DEVIL::tap_file_name_allocated) {free(DEVIL::tap_file_name);}
		DEVIL::tap_file_name_allocated=false;

		frame1->hide();
		frame5->hide();

		// =============== set context-dependent button tooltips ====================================
		QToolTip::add( pushButton3, Push3_1_Tip );
		QToolTip::add( pushButton4, Push4_1_Tip );
		QToolTip::add( pushButton5, Push5_1_Tip );
		QToolTip::add( pushButton6, Push6_1_Tip );
		QToolTip::add( pushButton7, Push7_1_Tip );
		QToolTip::add( textLabel3s, StatB1_1_tip );
		QToolTip::add( textLabel4s, StatB2_1_tip );

		// =============== disable context-dependent buttons and clear status bar ===================
		textLabel1->setText("  No    Type    Name       Length  Start VarStart VarN Attributes     ");

		pushButton3->setText("&Extract To .TAP");	menuExtractAction->setText("Extract selected catalog entries to a .TAP file");	menuExtractAction->setMenuText("&Extract To .TAP");
		pushButton4->setText("&Insert From .TAP");	menuInsertAction->setText("Insert catalog entries from a .TAP file");		menuInsertAction->setMenuText("&Insert From .TAP");
		pushButton5->setText("&Rename Cat. Entry");	menuRenameAction->setText("Rename (one single) selected catalog entry");	menuRenameAction->setMenuText("&Rename Catalog Entry");
		pushButton6->setText("&Delete Cat. Entries");	menuDeleteAction->setText("Delete multiple selected catalog entries");		menuDeleteAction->setMenuText("&Delete Catalog Entries");
		pushButton7->setText("Change &Attributes");	menuChange_AttrsAction->setText("Change disk attributes for multiple selected catalog entries");	menuChange_AttrsAction->setMenuText("Change &Attributes");

		pushButton3->setEnabled(false);		menuExtractAction->setEnabled(FALSE);		// disable "Extract To .TAP" button & menu
		pushButton4->setEnabled(false);		menuInsertAction->setEnabled(FALSE);		// disable "Insert From .TAP" button & menu
		pushButton5->setEnabled(false);		menuRenameAction->setEnabled(FALSE);		// disable "Rename Block" button & menu
		pushButton6->setEnabled(false);		menuDeleteAction->setEnabled(FALSE);		// disable "Delete Block(s)" button & menu
		pushButton7->setEnabled(false);		menuChange_AttrsAction->setEnabled(FALSE);	// disable "Change Attributes" button & menu
		pushButton11->setEnabled(false);	menuDeInterleaveAction->setEnabled(FALSE);	// disable "De-Interleave Image" button & menu

		checkBox4->setEnabled(false); checkBox4->setChecked(0);	// disable "Legend" checkbox
		textLabel3s->clear();textLabel4s->clear();
		updateStatusMessage("");
		// =============== reset title bar caption ==================================================
		this->setCaption("CoBra DEVIL Floppy Image Utility");
		// ==========================================================================================
		return;
	}			// ===================================================================================

	// =============== populate listbox with list of blocks in .TAP file loaded ==================================
	for (unsigned int i=0; i<DEVIL::tap_idx; i++) {
		listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::tap_list[i]).remove(QRegExp("[\\n\\t\\r]")) ));
//		listItems[i] = new Q3ListBoxText(listBox1, QString(DEVIL::tap_list[i]).remove(QRegExp("[\\n\\t\\r]")) );
	}
	// =============== clear and release memory allocated to DEVIL::tap_list[] ====================
	for (int i=DEVIL::tap_idx-1; i>=0; i--) { free(DEVIL::tap_list[i]); }
	free(DEVIL::tap_list);
	// ============================================================================================

	this->setMaximumWidth(770);

	// =============== set context-dependent button tooltips ====================================
	QToolTip::add( pushButton3, Push3_2_Tip );
	QToolTip::add( pushButton4, Push4_2_Tip );
	QToolTip::add( pushButton5, Push5_2_Tip );
	QToolTip::add( pushButton6, Push6_2_Tip );
	QToolTip::add( pushButton7, Push7_2_Tip );
	QToolTip::add( textLabel3s, StatB1_2_tip );
	QToolTip::add( textLabel4s, StatB2_2_tip );

	// =============== set context-dependent GUI text ============================================================
	textLabel1->setText(" Form E  T#    D#  Flag  Size  Type     Name     Length  Start VarStart VarN   ");
	textLabel3s->clear();textLabel4s->clear();
	frame1->hide();		// hide DEVIL info box
	frame5->show();		// show .TAP info box

	pushButton3->setText("&Extract To .BIN");	menuExtractAction->setText("Extract selected block to a binary file");		menuExtractAction->setMenuText("&Extract To .BIN");
	pushButton4->setText("&Insert From .TAP/.BIN");	menuInsertAction->setText("Insert data block from a .TAP or binary file");	menuInsertAction->setMenuText("&Insert From .TAP/.BIN");
	pushButton5->setText("&Rename Header");		menuRenameAction->setText("Change name stored in selected header");		menuRenameAction->setMenuText("&Rename Header");
	pushButton6->setText("&Delete Block");		menuDeleteAction->setText("Delete selected data block");			menuDeleteAction->setMenuText("&Delete Block");
	pushButton7->setText("Create &Header");		menuChange_AttrsAction->setText("Create header for selected headerless block");	menuChange_AttrsAction->setMenuText("Create &Header");

	pushButton3->setEnabled(false);	menuExtractAction->setEnabled(FALSE);
	pushButton4->setEnabled(true);	menuInsertAction->setEnabled(TRUE);
	pushButton5->setEnabled(false);	menuRenameAction->setEnabled(FALSE);
	pushButton6->setEnabled(false);	menuDeleteAction->setEnabled(FALSE);
	pushButton7->setEnabled(false);	menuChange_AttrsAction->setEnabled(FALSE);
	pushButton11->setEnabled(false); menuDeInterleaveAction->setEnabled(FALSE);

	checkBox4->setEnabled(true); checkBox4->setChecked(1);

	// =============== print status message in console and status bar and update title bar =======================
	printf ("Loaded .TAP file '%s'\n\n", f.fileName().toAscii().constData());
	this->updateStatusMessage( s0+f.fileName()+"'" );
	this->setWindowTitle(QString("%1 - '%2'").arg("CoBra DEVIL Floppy Image Utility",f.fileName()));

	// =============== store .TAP file full path in global variable DEVIL::tap_file_name =========================
	if (DEVIL::tap_file_name_allocated) {free(DEVIL::tap_file_name);} DEVIL::tap_file_name_allocated=true;
	DEVIL::tap_file_name = (char*)malloc((s.length()+1)*sizeof(char));	// +1 for ending '\0'
	sprintf(DEVIL::tap_file_name, "%s", s.toAscii().constData());

	return;

}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton3->clicked() ("Extract To .TAP").
// Extracts data from currently open Devil floppy image to a new .TAP file - or a data block from currently open .TAP file to a .BIN file
void MainWindow::extractFiles() {

	int ret_code=0;
	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window

	if (textLabel1->text()[2] == 'N') {	// ################## IF FLOPPY IMAGE OPEN, EXTRACT TO .TAP ##################

		unsigned char sel_items=selected_items;

		/* ---------------------------- TESTING ----------------------------------- *
		msg.setWindowTitle("RangeList Contents:");
		msg.textLabel1->setText("");
		for (int i=0; i<nr_of_ranges; i++) {
			msg.textLabel1->setText(msg.textLabel1->text() + QString("").sprintf("%d - %d\n", RangeList[i][0], RangeList[i][1]));
		}
		msg.exec();
		* ------------------------------------------------------------------------ */

		// =============== get the directory to save the .TAP file in ================================================
		QString d = QFileDialog::getExistingDirectory(
			0,
			"Select directory to save .TAP file",
			"/home/knoppix",
			QFileDialog::DontUseNativeDialog);
		if (d==QString::null) {return;}			// if user clicked "Cancel" abort
//printf("\n\nSelected directory = %s\n\n", d.toAscii().constData()); std::cout << std::flush;					// test point
		// =============== get a file name for the .TAP file to be saved =============================================
		SaveNameDialog w9; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w9.setupUi(&dlg);		// create the SaveNameDialog window
		w9.setWindowTitle(".TAP file name:"); w9.textLabel3->setText(".tap");
		if (!w9.exec()) {return;}			// if OK clicked, DEVIL::save_file_name will contain the .TAP new file name

		// =============== allocate string to DEVIL::path_to_dest_dir_base ===========================================
		if (DEVIL::path_to_dest_dir_base_allocated) {free(DEVIL::path_to_dest_dir_base);}
		DEVIL::path_to_dest_dir_base_allocated=true;
		DEVIL::path_to_dest_dir_base=(char*)malloc((d.length()+1)*sizeof(char));
		asprintf(&DEVIL::path_to_dest_dir_base, "%s", d.toAscii().constData());		// set path to saving dir to user choice
//		DEVIL::path_to_dest_dir_base[strlen(DEVIL::path_to_dest_dir_base)-1]='\0';	// and delete the ending '/'	// no need anymore in Qt4, there's no ending '/'

		// =============== assemble full path of directory for saving ================================================
		DEVIL::diskimg_dir();								// set disk image base saving dir (in dest_dir_base)
												// as "path_to_dest_dir_base/diskimg_file_name" minus the ".img" extension
		// =============== check if any regular file exists with the same name as dir for saving =====================
		if (DEVIL::xis_dir(DEVIL::dest_dir_base) == -2) {
			printf ("Error: '%s' already exists but is not a directory\n", DEVIL::dest_dir_base);
			msg.setWindowTitle("Directory Error");
			w4.textLabel1->setText(QString("").sprintf("Error: '%s' already exists\nbut is not a directory", DEVIL::dest_dir_base));
			msg.exec();
			return;
		}

		// =============== copy contents of DEVIL::save_file_name to a local QString =================================
		QString save_file_name(DEVIL::save_file_name);

		// =============== clear memory allocated to DEVIL::save_file_name by the SaveNameDialog dialog ==============
		if (DEVIL::save_file_name_allocated) {free(DEVIL::save_file_name); DEVIL::save_file_name_allocated=false;}

		// =============== save all selected catalog items using as .TAP file name the local QString =================
		for (int i=0; i<nr_of_ranges; i++) {
			ret_code = DEVIL::devilfilecopy(RangeList[i][0], RangeList[i][1], save_file_name.toAscii().constData(), 0);
			if (ret_code) break;
		// =============== clear current range selection in GUI list of catalog entries ==============================
//			for (unsigned char p=RangeList[i][0]; p<=RangeList[i][1]; p++) {listBox1->setSelected(listItems[p],false);}
		}

		// =============== clear selection in GUI list of catalog entries ============================================
		for (unsigned char p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++) {listBox1->setSelected(listItems[p],false);}

		// =============== display succes / failure messages =========================================================
		if (ret_code) {msg.setWindowTitle("Error"); w4.textLabel1->setText(QString("").sprintf("Error %d writing to .TAP file", ret_code)); msg.exec();} else
		updateStatusMessage(QString("").sprintf("%d catalog entries successfully extracted.",sel_items));

	} else {				// ################## IF .TAP FILE OPEN, EXTRACT (SINGLE) SELECTED BLOCK TO .BIN ##################

		unsigned int i;
		char data_len[6]; unsigned short d_len=0;
		unsigned int tap_file_address=0;

		// =============== get the directory to save the .BIN file in ================================================
		QString d = QFileDialog::getExistingDirectory(
			0,
			"Select directory to save in",
			"/home/knoppix",
			QFileDialog::DontUseNativeDialog);	// ATTENTION!!!  this will return a QString ending in "/"	// NOT ANYMORE IN QT4 !!!
		if (d==QString::null) {return;}			// if user clicked "Cancel" abort

		// =============== get a file name for the .BIN file to be saved =============================================
		SaveNameDialog w9; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w9.setupUi(&dlg);		// create the SaveNameDialog window
		w9.setWindowTitle(".BIN file name:"); w9.textLabel3->setText(".bin");
		if (!w9.exec()) {return;}			// if OK clicked, DEVIL::save_file_name will contain the .BIN new file name (without ".bin")

		// =============== check if destination file already exists ==================================================
		if (DEVIL::xis_dir((d+QString("/")+QString(DEVIL::save_file_name)+QString(".bin")).toAscii().constData())!=-1) {printf("Error: file '%s' already exists. Disk image creation aborted.\n", (d+QString("/")+QString(DEVIL::save_file_name)+QString(".bin")).toAscii().constData()); return;}

		// =============== copy contents of DEVIL::save_file_name to a local QString ===================================
		QString save_file_name(DEVIL::save_file_name);

		// =============== clear memory allocated to DEVIL::save_file_name by the SaveNameDialog dialog =================
		if (DEVIL::save_file_name_allocated) {free(DEVIL::save_file_name); DEVIL::save_file_name_allocated=false;}

		// =============== find the .TAP block selected in the GUI list ==============================================
		for (i=0; i<DEVIL::tap_idx; i++)
			{ if (listBox1->isSelected(i)) {break;} }	// i stores the index of selected list item which is the same as the block index in .TAP file

		// =============== extract data block length from GUI file list ==============================================
		for (unsigned char j=0; j<5; j++) {data_len[j]=listItems[i]->text().toAscii().constData()[j+23];}  data_len[5]='\0';
		d_len=atoi(data_len)-2;					// size of selected block minus the first byte (flag byte) minus the final byte (checksum byte)
		tap_file_address = DEVIL::tap_index[i]+3;		// actual data begins after the first 2 bytes (length) and the third byte (flag)

		// =============== save .TAP data block using as .TAP file name the local QString ============================
		// DEVIL::tap_file_name already initialized by openTap(), extracttapblock() will open and close the .TAP file
		ret_code = DEVIL::extracttapblock((d+QString("/")+save_file_name+QString(".bin")).toAscii().constData(), tap_file_address, d_len);

		// =============== clear selection in GUI list of tape items (only 1 item was selected) ======================
		listBox1->setSelected(listItems[i],false);

		// =============== display succes / failure messages =========================================================
		if (!ret_code) {
			updateStatusMessage(QString("").sprintf( "1 .TAP data block successfully extracted to '%s'", (save_file_name+QString(".bin")).toAscii().constData() ));
		} else {
			msg.setWindowTitle("Error Saving");
			w4.textLabel1->setText(QString("").sprintf( "Error saving data to '%s'", (save_file_name+QString(".bin")).toAscii().constData() ));
			msg.exec();
		}
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton4->clicked() ("Open .TAP File").
// Writes data from a .TAP file to a Devil floppy image
void MainWindow::insertFiles() {	// <<===== CONVERSION IN PROGRESS

	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("Error");
	char ret_code;
	
	if (textLabel1->text()[2] == 'N') {	// ################## IF FLOPPY IMAGE OPEN, INSERT FROM .TAP FILE ##################

		int tap_block_counter;
		// =============== get the full path of .TAP file to open ====================================================
		QString s;
		s = QFileDialog::getOpenFileName(
			0,
			"Choose a .TAP file to insert from",
			"/home/knoppix",
			"Tape files (*.tap)",
			0,
			QFileDialog::DontUseNativeDialog);
		if (s==QString::null) {return;}
		QFileInfo f(s);

		// =============== read .TAP file contents and determine nr of disk blocks required ==========================
		DEVIL::tap_list_mode=1;		// need this to prevent creation of dynamic array of char* pointers tap_list[] in DEVIL::devilfilewrite()
		DEVIL::blocks_needed = DEVIL::devilfilewrite(s.toAscii().constData(), 0, 0);
		DEVIL::tap_list_mode=0; tap_block_counter = DEVIL::tap_idx;
		DEVIL::tap_idx=0;		// need to clear this to avoid invalid "delete" later on in openTap()
		if (DEVIL::blocks_needed == -1) {w4.textLabel1->setText(".TAP file too short, aborting."); msg.exec(); return;} else
		if (DEVIL::blocks_needed == -2) {w4.textLabel1->setText("Error opening .TAP file, aborting"); msg.exec(); return;} else
		if (DEVIL::blocks_needed == -3) {w4.textLabel1->setText("Non-transferable headerless blocks\nand dataless header blocks\nfound in .TAP file, aborting"); msg.exec(); return;} else
		if (DEVIL::blocks_needed == -4) {w4.textLabel1->setText("Non-transferable headerless blocks\nfound in .TAP file, aborting"); msg.exec(); return;} else
		if (DEVIL::blocks_needed == -5) {w4.textLabel1->setText("Dataless headers found\nin .TAP file, aborting"); msg.exec(); return;} else
		if (DEVIL::blocks_needed >  77) {w4.textLabel1->setText("Number of .TAP file data blocks\nexceeds maximum number of allocation units\navailable on a DEVIL floppy disk"); msg.exec(); return;} else

		// =============== if nr of blocks > 0, write .TAP file contents to floppy image =============================
		if (DEVIL::blocks_needed > 0) {
			printf("Total tracks required: %d\n-------------------------\n\n", DEVIL::blocks_needed);
			int i=0, j=0, k;
			while (i<77) {	// ========== scan DEVIL::alloc_dir[] to find big enough continuous free space =======
				while ((i<77) && (DEVIL::alloc_dir[i]==1)) {i++;}	// i stops at first free dir entry or at 77
				if (i==77) {w4.textLabel1->setText("Disk image full, writing aborted.\n"); msg.exec(); return;}
				else {	// ========== if a free dir entry was found, start counting next free space ==========
					k=i;	// k keeps the index of first free dir entry found
					while ((i<77) && (DEVIL::alloc_dir[i]==0) && (j < DEVIL::blocks_needed)) {
						i++;	// i stops at first allocated dir entry or at 77
						j++;	// j counts nr of continuous free dir entries
					}
					if (j == DEVIL::blocks_needed) { // === if enough free space already found, write data
						printf("Found enough continuous free space at block %d (track %d)\n\n", k, k+3);

						// =============== write .TAP file contents to floppy image ==================================================
						ret_code = DEVIL::devilfilewrite(s.toAscii().constData(), 1, k); DEVIL::tap_idx=0;

						/* -------------------------------- UPDATE GUI WITH NEW FLOPPY CONTENTS ----------------------------------- */
						// =============== clear selection in GUI list of floppy items ===============================================
						for (unsigned char p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++) {listBox1->setSelected(listItems[p],false);}

						// =============== delete previous floppy items if present, before displaying new image contents =============
						if (DEVIL::nr_files > 0) {
//							for (int i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}			// replacing these 2 lines with a single one
//							for (unsigned int i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {free(DEVIL::file_list[i]);}				// file_list[] is a const size array (DEVIL catalog has fized size of 77)
							for (short i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back(); free(DEVIL::file_list[i]);}
							DEVIL::nr_files = 0; DEVIL::nr_hless = 0; DEVIL::total_size = 0; }

						// ===========================================================================================================
						DEVIL::devildirdump();		// this will set global vars DEVIL::file_list[], DEVIL::nr_files and DEVIL::nr_hless

						// =============== populate listbox with list of files in image loaded =======================================
						for (unsigned short p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++)
							{ listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::file_list[p]).remove(QRegExp("[\\n\\t\\r]")) )); }

						// =============== enable/disable buttons ====================================================================
						pushButton3->setEnabled(false);	// disable "Extract To .TAP" button
						pushButton5->setEnabled(false);	// disable "Rename Block" button
						pushButton6->setEnabled(false);	// disable "Delete Block(s)" button
						pushButton7->setEnabled(false);	// disable "Change Attributes" button

						// =============== updating info box below GUI file list =====================================================
						textLabel10->setText(QString("").sprintf("%2d",77-DEVIL::free_dir_recs));
						textLabel11->setText(QString("").sprintf("%6d",DEVIL::total_size));
						textLabel12->setText(QString("").sprintf("%6d",DEVIL::nr_recs*9*1024));
						textLabel13->setText(QString("").sprintf("%2d",DEVIL::free_dir_recs));
						textLabel14->setText(QString("").sprintf("%2d",DEVIL::nr_files));
						textLabel15->setText(QString("").sprintf("%3d",DEVIL::free_dir_recs*9));

						// =============== updating allocation map in info box =======================================================
						for (unsigned char p=0; p<77; p++) {fragview[p]=(DEVIL::alloc_dir[p]?'I':'-');} fragview[77]='\0';
						textLabel26->setText(fragview);

						// =============== clear status bar selection info and display success message ===============================
						textLabel3s->setText("");
						textLabel4s->setText("");
						updateStatusMessage(QString("Data in '%1'").arg(f.fileName()) + QString(" (%1 blocks) successfully inserted").arg(tap_block_counter/2) );
						/* -------------------------------------------------------------------------------------------------------- */
						return;
					} else {j=0;}
				}
			}
			w4.textLabel1->setText("Not enough continuous free space available on disk, writing aborted.\n"); msg.exec(); return;
		}

	} else {				// ################## IF .TAP FILE OPEN, INSERT BINARY DATA FROM A .TAP OR .BIN FILE ##################

		QString err_str("");
		unsigned int i, tap_idx_backup, tap_block_counter;
		unsigned int tap_file_address=0;

		QFileInfo f(DEVIL::tap_file_name);			// get info of currently open .TAP file

		// =============== ask if inserting from .TAP or from .BIN ===================================================
		OptionDialog w7; //QDialog opt(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w7.setupUi(&opt);	// create the Option dialog
		w7.radioButton1->setText("Insert from .TAP file");
		w7.radioButton2->setText("Insert from .BIN file");
		if (!w7.exec()) {return;}				// this will set DEVIL::option_selected if OK is clicked

		if (DEVIL::option_selected) {		// ################ if inserting from .BIN file ######################

			// =============== select a source binary file having data to insert =========================================
			QString s;
			s = QFileDialog::getOpenFileName(
				0,
				"Choose a .BIN file to insert from",
				"/home/knoppix",
				"Binary files (*.bin)",
				0,
				QFileDialog::DontUseNativeDialog);
			if (s==QString::null) {return;}
			QFileInfo f1(s);

			// =============== check size of source binary file (should be < 65535) ======================================
			if (f1.size()>=65536) {
				printf("Binary file size must be less than 64 KB\n");
				w4.textLabel1->setText("Binary file size must be less than 64 KB"); msg.exec(); return;
			}

			// =============== get a value for flag byte of headerless block to be created in new .TAP file ==============
			FlagByteDialog w5; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w5.setupUi(&dlg);	// create the Flag Byte dialog
			if (!w5.exec()) {return;}		// this will set DEVIL::devil_tap_block_flag if OK is clicked

			// =============== find the .TAP block selected in the GUI list, if any ======================================
			for (i=0; i<DEVIL::tap_idx; i++) {
				if (listBox1->isSelected(i)) {break;}		// i stores the index of selected list item which is the same as the block index in .TAP file.
			}							// If no item selected, i ends up == DEVIL::tap_idx.
			// =============== if one block selected, insert data before it, if no block selected, add data to end of .TAP file
			if (listBox1->isSelected(i)) {tap_file_address = DEVIL::tap_index[i];} else {tap_file_address = f.size();}

			// =============== call function to insert data block ========================================================
			ret_code = DEVIL::insertbinblock(tap_file_address, s.toAscii().constData(), DEVIL::devil_tap_block_flag);

			if (ret_code) {w4.textLabel1->setText(QString("").sprintf("Error %d inserting binary data.", ret_code)); msg.exec();}

			// =============== display succes / failure messages =========================================================
			if (!ret_code) {
				updateStatusMessage(QString("Contents of '%1' successfully inserted as headerless block.").arg(f1.fileName()));
			} else {
				updateStatusMessage(QString("Data insertion failed."));
			}

		} else {				// ################ if inserting from .TAP file ######################

			// =============== select a source .TAP file having data to insert ===========================================
			QString s;
			s = QFileDialog::getOpenFileName(
				0,
				"Choose a .TAP file to insert from",
				"/home/knoppix",
				"Tape files (*.tap)",
				0,
				QFileDialog::DontUseNativeDialog);
			if (s==QString::null) {return;}
			QFileInfo f1(s);

			// =============== run a quick check on the .TAP file ========================================================
			DEVIL::tap_list_mode = 1; tap_idx_backup = DEVIL::tap_idx;
			ret_code = DEVIL::devilfilewrite(s.toAscii().constData(),0,0);
			DEVIL::tap_list_mode = 0; tap_block_counter = DEVIL::tap_idx; DEVIL::tap_idx = tap_idx_backup;
			switch (ret_code) {
				case -1: err_str = QString(".TAP file too short.");					break;
				case -2: err_str = QString("Error opening .TAP file.");					break;
				case -6: err_str = QString("Error reading block length.\nUnexpected end of file!");	break;
				case -7: err_str = QString("Block end missing.\nUnexpected end of file!");		break;
			}
			if ((ret_code==-1) || (ret_code==-2) || (ret_code==-6) || (ret_code==-7)) {w4.textLabel1->setText(err_str); msg.exec(); return;}

			// =============== find the .TAP block selected in the GUI list, if any ======================================
			for (i=0; i<DEVIL::tap_idx; i++) {
				if (listBox1->isSelected(i)) {break;}		// i stores the index of selected list item which is the same as the block index in .TAP file.
			}							// If no item selected, i ends up == DEVIL::tap_idx.
			// =============== if one block selected, insert data before it, if no block selected, add data to end of .TAP file
			if (listBox1->isSelected(i)) {tap_file_address = DEVIL::tap_index[i];} else {tap_file_address = f.size();}

			// =============== call function to insert data block ========================================================
			ret_code = DEVIL::inserttapfile(tap_file_address, s.toAscii().constData());

			if (ret_code) {w4.textLabel1->setText(QString("").sprintf("Error %d inserting binary data.", ret_code)); msg.exec();}

			// =============== display succes / failure messages =========================================================
			if (!ret_code) {
				updateStatusMessage(QString("Contents of '%1' successfully inserted").arg(f1.fileName()) + QString(" (%1 blocks).").arg(tap_block_counter));
			} else {
				updateStatusMessage(QString("Data insertion failed."));
			}

		}

		// =============== clear selection in GUI list of tape items =================================================
		for (unsigned short p=0; p<DEVIL::tap_idx; p++) {listBox1->setSelected(listItems[p],false);}

		// =============== delete previous tape items if present, before displaying new tape contents ================
		if (DEVIL::tap_idx > 0) {
			for (int p=DEVIL::tap_idx-1; p>=0; p--) {delete listItems[p]; listItems.pop_back();}
			free(DEVIL::tap_index);
			DEVIL::tap_idx = 0; }

		// =============== re-read the .TAP file currently open ======================================================
		DEVIL::devilfilewrite(DEVIL::tap_file_name, 0, 0);

		// =============== populate listbox with list of data blocks in .TAP file ====================================
		for (unsigned short p=0; p<DEVIL::tap_idx; p++) {
			listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::tap_list[p]).remove(QRegExp("[\\n\\t\\r]")) ));
		}
		// =============== clear and release memory allocated to dynamic char* tap_list[] by devilfilewrite() ========
		for (int p=DEVIL::tap_idx-1; p>=0; p--) { free(DEVIL::tap_list[p]); }
		free(DEVIL::tap_list);
		// ===========================================================================================================
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton5->clicked() ("Rename File").
// Renames the data block selected in the list.
void MainWindow::renameFile() {

	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	RenameDialog w8; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w8.setupUi(&dlg);	// create the rename dialog window

	if (textLabel1->text()[2] == 'N') {		// ################## IF FLOPPY IMAGE OPEN, RENAME CATALOG ENTRY FROM FLOPPY ##################

		int ret_code=0; QString err_str("");
		char lst[3]; lst[2]='\0';
		unsigned char lst_no;	// order number in file list
		unsigned char idx_no;	// order number in disk catalog
		char devil_file_name[11]; for (int i=0; i<11; i++) {devil_file_name[i]='\0';}

		// we could rename a few selected data blocks one after another, but I decided to only activate the Rename button if only one block is selected...
		for (unsigned char i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {
			if (listBox1->isSelected(listItems[i])) {	// if list item is selected
				lst[0] = DEVIL::file_list[i][1]; lst[1]=DEVIL::file_list[i][2];	// lst = index of file in file list as string
				lst_no = (unsigned char)atoi(lst);				// lst_no = index of file in file list (and in devildir1_mem[]) as number
				idx_no = DEVIL::devildir1_mem[lst_no*256];			// idx_no = index of file in Devil disk directory (first block of file) (and in devildir0_mem[])
				
				memcpy(&devil_file_name[0],&DEVIL::devildir0_mem[idx_no*256+1], 10);	// extract DEVIL catalog name of item - in devil_file_name (better get it from the primary source)
	
				w8.lineEdit1->setText(QString("").sprintf("%s", devil_file_name));	// set input field in RenameDialog to current name
	
				if (!w8.exec()) {return;}						// if Cancel pressed, abort rename operation
				ret_code = DEVIL::devilfilerename(DEVIL::devil_file_new_name, lst_no);
	
				// =============== deselect item in GUI list of floppy items =================================================
				listBox1->setSelected(listItems[i],false);
	
				switch (ret_code) {
					case 1: err_str = QString("Error opening floppy image file");			break;
					case 2: err_str = QString("Error writing to floppy image file");		break;
				}
				if (ret_code) {w4.textLabel1->setText(err_str); msg.exec(); break;}

				// =============== display success message ===================================================================
				updateStatusMessage( QString("").sprintf("Successfully renamed catalog entry '%s' to '%s'", devil_file_name, DEVIL::devil_file_new_name) );
			}
		}
	
		if (ret_code) {	// if error occured accessing current floppy image, we need to close it, clear the file list and revert to startup state
			// =============== release memory allocated to diskimg_file_name if any =====================
			if (DEVIL::diskimg_file_name_allocated) {free(DEVIL::diskimg_file_name);}
			DEVIL::diskimg_file_name_allocated=false;
			// =============== clear listbox items if any and hide disk info if present =================
			if (DEVIL::nr_files > 0U) {
//				for (unsigned short i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {delete listItems[i];}			// replacing these 2 lines with a single one
//				for (unsigned short i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {free(DEVIL::file_list[i]);}		// file_list[] is a const size array (DEVIL catalog has fized size of 77)
				for (short i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back(); free(DEVIL::file_list[i]);}
				DEVIL::nr_files = 0; DEVIL::nr_hless = 0;
				this->setCaption("CoBra DEVIL Floppy Image Utility");
			}
			frame1->hide();
			frame5->hide();
			// =============== disable context-dependent buttons and clear status bar ===================
			pushButton3->setEnabled(false);	// disable "Extract To .TAP" button
			pushButton4->setEnabled(false);	// disable "Insert From .TAP" button
			pushButton5->setEnabled(false);	// disable "Rename Block" button
			pushButton6->setEnabled(false);	// disable "Delete Block(s)" button
			pushButton7->setEnabled(false);	// disable "Change Attributes" button
			pushButton11->setEnabled(false);// disable "De-Interleave Image" button
			checkBox4->setEnabled(false);	// disable "Legend" checkbox
			textLabel3s->setText("");
			textLabel4s->setText("");
			updateStatusMessage("");
			// ==========================================================================================
			return;
		}

		/* -------------------------------- UPDATE GUI WITH NEW FLOPPY CONTENTS ----------------------------------- */
		// =============== delete previous floppy items if present, before displaying new image contents =============
		if (DEVIL::nr_files > 0) {
//			for (int p=DEVIL::nr_files+DEVIL::nr_hless-1; p>=0; p--) {delete listItems[p]; listItems.pop_back();}			// replacing these 2 lines with a single one
//			for (unsigned int p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++) {free(DEVIL::file_list[p]);}				// file_list[] is a const size array (DEVIL catalog has fized size of 77)
			for (short p=DEVIL::nr_files+DEVIL::nr_hless-1; p>=0; p--) {delete listItems[p]; listItems.pop_back(); free(DEVIL::file_list[p]);}
			DEVIL::nr_files = 0; DEVIL::nr_hless = 0; }

		// ===========================================================================================================
		DEVIL::devildirdump();		// this will set global vars DEVIL::file_list[], DEVIL::nr_files and DEVIL::nr_hless

		// =============== populate listbox with list of files in image loaded =======================================
		for (unsigned short p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++)
			{ listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::file_list[p]).remove(QRegExp("[\\n\\t\\r]")) )); }

		// =============== enable/disable buttons ====================================================================
		pushButton3->setEnabled(false);	// disable "Extract To .TAP" button
		pushButton5->setEnabled(false);	// disable "Rename Block" button
		pushButton6->setEnabled(false);	// disable "Delete Block(s)" button
		pushButton7->setEnabled(false);	// disable "Change Attributes" button

		// =============== clear status bar selection info ===========================================================
		textLabel3s->setText("");
		textLabel4s->setText("");
		/* -------------------------------------------------------------------------------------------------------- */

	} else {					// ################## IF .TAP FILE OPEN, RENAME BLOCK FROM .TAP ##################

		char ret_code;
		unsigned int i=0;
		unsigned char tap_buffer[21];
		char name[11];

		// =============== find the .TAP block selected in the GUI list, if any ======================================
		for (i=0; i<DEVIL::tap_idx; i++) {
			if (listBox1->isSelected(i)) {break;}		// i stores the index of selected list item which is the same as the block index in .TAP file.
		}							// If no item selected, i ends up == DEVIL::tap_idx.

		// =============== read the .TAP header selected in the GUI list (renaming only applies to headers) ==========
		ret_code = DEVIL::readtapdata(DEVIL::tap_index[i], 21, tap_buffer);
		if (ret_code) {w4.textLabel1->setText(QString("").sprintf("Error reading .TAP file.")); msg.exec(); return;}

		// =============== extract name in the header into name[] as 7-bit ASCII chars ===============================
		for (unsigned char j=0; j<10; j++) {name[j] = tap_buffer[j+4] & 0x7F;}  name[10]='\0';

		// =============== show the Rename dialog with current name filled in ========================================
		w8.lineEdit1->setText(QString("").sprintf("%s", name));			// set input field in RenameDialog to name[]
		if (!w8.exec()) {return;} // this saves name to DEVIL::devil_file_new_name 	// if Cancel pressed, abort rename operation

		// =============== get new name into tap_buffer[] ============================================================
		for (unsigned char j=0; j<10; j++) {tap_buffer[j+4] = DEVIL::devil_file_new_name[j];}	// copy renamed back to tap_buffer[]

		// =============== recalculate new checksum for header block =================================================
		tap_buffer[20]=0; for (unsigned char j=2; j<20; j++) {tap_buffer[20]^=tap_buffer[j];}	// recalculate checksum

		// =============== write header block back to .TAP file ======================================================
		ret_code = DEVIL::writetapdata(DEVIL::tap_index[i], 21, tap_buffer);		// write tap_buffer[] back to .TAP file

		if (ret_code) {w4.textLabel1->setText(QString("").sprintf("Error writing to .TAP file.")); msg.exec(); return;}
		updateStatusMessage( QString("").sprintf("Successfully renamed .TAP data block '%s' to '%s'", name, DEVIL::devil_file_new_name) );

		// =============== clear selection in GUI list of tape items =================================================
		for (unsigned short p=0; p<DEVIL::tap_idx; p++) {listBox1->setSelected(listItems[p],false);}

		// =============== delete previous tape items if present, before displaying new tape contents ================
		if (DEVIL::tap_idx > 0) {
			for (int p=DEVIL::tap_idx-1; p>=0; p--) {delete listItems[p]; listItems.pop_back();}
			free(DEVIL::tap_index);
			DEVIL::tap_idx = 0; }

		// =============== re-read the .TAP file currently open ======================================================
		DEVIL::devilfilewrite(DEVIL::tap_file_name, 0, 0);

		// =============== populate listbox with list of data blocks in .TAP file ====================================
		for (unsigned short p=0; p<DEVIL::tap_idx; p++) {
			listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::tap_list[p]).remove(QRegExp("[\\n\\t\\r]")) ));
		}
		// =============== clear and release memory allocated to dynamic char* tap_list[] by devilfilewrite() ========
		for (int p=DEVIL::tap_idx-1; p>=0; p--) { free(DEVIL::tap_list[p]); }
		free(DEVIL::tap_list);
		// ===========================================================================================================
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton6->clicked() ("Delete Block(s)").
// Deletes the data blocks selected in the list.
void MainWindow::deleteFiles() {

	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("Error");
	Ui::ConfirmDialog w1; QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w1.setupUi(&dlg);	// create the confirmation dialog
	w1.pushButton1->setText("Yes"); w1.pushButton2->setText("No");

	if (textLabel1->text()[2] == 'N') {		// ################## IF FLOPPY IMAGE OPEN, DELETE SELECTED CATALOG ENTRIES FROM FLOPPY ##################

		int ret_code, sel_items=selected_items;

		w1.textLabel1->setText( QString("").sprintf("%d selected catalog entries will be deleted.\nAre you sure you want to continue?", sel_items) );
		if (!dlg.exec()) return;
	
		/* ---------------------------- TESTING ----------------------------------- *
		msg.setWindowTitle("RangeList Contents:");
		msg.textLabel1->setText("");
		for (int i=0; i<nr_of_ranges; i++) {
			msg.textLabel1->setText(msg.textLabel1->text() + QString("").sprintf("%d - %d\n", RangeList[i][0], RangeList[i][1]));
		}
		msg.exec();
		* ------------------------------------------------------------------------ */

		for (int i=0; i<nr_of_ranges; i++) {
			ret_code = DEVIL::devilfiledelete(RangeList[i][0], RangeList[i][1]);					// delete all selected data blocks
		}

		/* -------------------------------- UPDATE GUI WITH NEW FLOPPY CONTENTS ----------------------------------- */
		// =============== clear selection in GUI list of floppy items ===============================================
		for (unsigned short p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++) {listBox1->setSelected(listItems[p],false);}

		// =============== delete previous floppy items if present, before displaying new image contents =============
		if (DEVIL::nr_files > 0) {
//			for (int i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back();}	// replacing these 2 lines with a single one
//			for (unsigned int i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {free(DEVIL::file_list[i]);}		// file_list[] is a const size array (DEVIL catalog has fized size of 77)
			for (short i=DEVIL::nr_files+DEVIL::nr_hless-1; i>=0; i--) {delete listItems[i]; listItems.pop_back(); free(DEVIL::file_list[i]);}
			DEVIL::nr_files = 0; DEVIL::nr_hless = 0; DEVIL::total_size = 0; }

		// ===========================================================================================================
		DEVIL::devildirdump();		// this will set global vars DEVIL::file_list[] and DEVIL::nr_files		// update DEVIL::alloc_dir[] for defrag_disk() to see next
		DEVIL::defrag_disk();												// defragment free space on disk
		DEVIL::total_size = 0;		// total_size needs to be reset when running devildirdump() twice (to avoid displaying wrong value for total file size in GUI)
		DEVIL::devildirdump();												// update DEVIL::alloc_dir[] again after defragmentation

		// =============== populate listbox with list of files in image loaded =======================================
		for (unsigned short p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++)
			{ listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::file_list[p]).remove(QRegExp("[\\n\\t\\r]")) )); }

		// =============== enable/disable buttons ====================================================================
		pushButton3->setEnabled(false);	// disable "Extract To .TAP" button
		pushButton5->setEnabled(false);	// disable "Rename Block" button
		pushButton6->setEnabled(false);	// disable "Delete Block(s)" button
		pushButton7->setEnabled(false);	// disable "Change Attributes" button

		// =============== updating info box below GUI file list =====================================================
		textLabel10->setText(QString("").sprintf("%2d",77-DEVIL::free_dir_recs));
		textLabel11->setText(QString("").sprintf("%6d",DEVIL::total_size));
		textLabel12->setText(QString("").sprintf("%6d",DEVIL::nr_recs*9*1024));
		textLabel13->setText(QString("").sprintf("%2d",DEVIL::free_dir_recs));
		textLabel14->setText(QString("").sprintf("%2d",DEVIL::nr_files));
		textLabel15->setText(QString("").sprintf("%3d",DEVIL::free_dir_recs*9));

		// =============== updating allocation map in info box =======================================================
		for (unsigned char p=0; p<77; p++) {fragview[p]=(DEVIL::alloc_dir[p]?'I':'-');} fragview[77]='\0';
		textLabel26->setText(fragview);

		// =============== clear status bar selection info and display success message ===============================
		textLabel3s->setText("");
		textLabel4s->setText("");
		updateStatusMessage( QString("").sprintf("%d catalog entries successfully deleted.", sel_items) );
		/* -------------------------------------------------------------------------------------------------------- */

	} else {					// ################## IF .TAP FILE OPEN, DELETE SELECTED SINGLE DATA BLOCK FROM .TAP ##################

		char ret_code;
		unsigned int i=0;

		w1.textLabel1->setText( "Selected data block will be deleted.\nAre you sure you want to continue?" );
		if (!dlg.exec()) return;

		// =============== find the .TAP block selected in the GUI list, if any ======================================
		for (i=0; i<DEVIL::tap_idx; i++) {
			if (listBox1->isSelected(i)) {break;}		// i stores the index of selected list item which is the same as the block index in .TAP file.
		}							// If no item selected, i ends up == DEVIL::tap_idx.

		// =============== delete the .TAP block selected in the GUI list ============================================
		ret_code = DEVIL::tapblockdelete(DEVIL::tap_index[i]);

		if (ret_code) {w4.textLabel1->setText(QString("").sprintf("Error %d deleting data block %04d", ret_code, i)); msg.exec(); return;}
		updateStatusMessage( QString("").sprintf("Successfully deleted .TAP data block %04d", i) );

		// =============== clear selection in GUI list of tape items (only 1 item was selected) ======================
		listBox1->setSelected(listItems[i],false);

		// =============== delete previous tape items if present, before displaying new tape contents ================
		if (DEVIL::tap_idx > 0) {
			for (int p=DEVIL::tap_idx-1; p>=0; p--) {delete listItems[p]; listItems.pop_back();}
			free(DEVIL::tap_index);
			DEVIL::tap_idx = 0; }

		// =============== re-read the .TAP file currently open ======================================================
		DEVIL::devilfilewrite(DEVIL::tap_file_name, 0, 0);

		// =============== populate listbox with list of data blocks in .TAP file ====================================
		for (unsigned short p=0; p<DEVIL::tap_idx; p++) {
			listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::tap_list[p]).remove(QRegExp("[\\n\\t\\r]")) ));
		}
		// =============== clear and release memory allocated to dynamic char* tap_list[] by devilfilewrite() ========
		for (int p=DEVIL::tap_idx-1; p>=0; p--) { free(DEVIL::tap_list[p]); }
		free(DEVIL::tap_list);
		// ===========================================================================================================
	}
}

void MainWindow::changeAttrs() {

	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("Error");

	if (textLabel1->text()[2] == 'N') {		// ################## IF FLOPPY IMAGE OPEN, CHANGE SELECTED CATALOG ENTRIES ATTRS ##################

//		char attr1[4], attr2[4];
		unsigned char attr=0;
		int sel_items=selected_items;

		DevAttrDialog w3(this); //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w3.setupUi(&dlg);	// create the Change Attributes dialog window

		if (!w3.exec()) return;

/*		// -------------- using devilchattr() instead of devilfilesetattr() ----------------
		if (DEVIL::R_W) {strcpy(attr1,"-r1");} else {strcpy(attr1,"-r0");}
		if (DEVIL::SYS) {strcpy(attr2,"-s1");} else {strcpy(attr2,"-s0");}

		for (int i=0; i<nr_of_ranges; i++) {
			DEVIL::devilfilesetattr(RangeList[i][0], RangeList[i][1], attr1);
			DEVIL::devilfilesetattr(RangeList[i][0], RangeList[i][1], attr2);
		}
*/
		attr = (DEVIL::R_W ? 2 : 0) | (DEVIL::SYS ? 1 : 0);
		for (int i=0; i<nr_of_ranges; i++) {
			for (int j=RangeList[i][0]; j<=RangeList[i][1]; j++) {DEVIL::devilfilechattr(attr, j);}
		}

		/* -------------------------------- UPDATE GUI WITH NEW FLOPPY CONTENTS ----------------------------------- */
		// =============== clear selection in GUI list of floppy items ===============================================
		for (unsigned short p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++) {listBox1->setSelected(listItems[p],false);}

		// =============== delete previous floppy items if present, before displaying new image contents =============
		if (DEVIL::nr_files > 0) {
//			for (int p=DEVIL::nr_files+DEVIL::nr_hless-1; p>=0; p--) {delete listItems[p]; listItems.pop_back();}			// replacing these 2 lines with a single one
//			for (unsigned int p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++) {free(DEVIL::file_list[p]);}				// file_list[] is a const size array (DEVIL catalog has fized size of 77)
			for (short p=DEVIL::nr_files+DEVIL::nr_hless-1; p>=0; p--) {delete listItems[p]; listItems.pop_back(); free(DEVIL::file_list[p]);}
			DEVIL::nr_files = 0; DEVIL::nr_hless = 0;}

		// ===========================================================================================================
		DEVIL::devildirdump();		// this will set global vars DEVIL::file_list[], DEVIL::nr_files and DEVIL::nr_hless

		// =============== populate listbox with list of files in image loaded =======================================
		for (unsigned short p=0; p<DEVIL::nr_files+DEVIL::nr_hless; p++)
			{ listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::file_list[p]).remove(QRegExp("[\\n\\t\\r]")) )); }

		// =============== enable/disable buttons ====================================================================
		pushButton3->setEnabled(false);	// disable "Extract To .TAP" button
		pushButton5->setEnabled(false);	// disable "Rename Block" button
		pushButton6->setEnabled(false);	// disable "Delete Block(s)" button
		pushButton7->setEnabled(false);	// disable "Change Attributes" button

		// =============== clear status bar selection info and display success message in status bar =================
		textLabel3s->setText("");
		textLabel4s->setText("");
		updateStatusMessage( QString("").sprintf("Successfully changed attributes for %d catalog entries.", sel_items) );
		/* -------------------------------------------------------------------------------------------------------- */

	} else {					// ################## IF .TAP FILE OPEN, CREATE HEADER FOR SELECTED HEADERLESS ##################

		unsigned char header_buffer[21];
		unsigned int i;
		char data_len[6]; unsigned short d_len=0;
		int ret_code;

//		Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
//		msg.setWindowTitle("Error");

		CreateHeader w2; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w2.setupUi(&dlg);	// create the Create Header dialog window
		if (!w2.exec()) {return;}			// if OK clicked, we have  DEVIL::devil_file_new_name and DEVIL::devil_file_new_addr  variables set

		/* +++++++++++++++ fill out header_buffer[] with proper data before writing it to the .TAP file +++++++++++ */
		header_buffer[0]=0x13; header_buffer[1]=0x00;	// block length = 13h for headers
		header_buffer[2]=0x00;				// flag byte = 00h for headers
		header_buffer[3]=0x03;				// data type = 03h for Code
		for (unsigned char j=0; j<10; j++) {header_buffer[j+4]=DEVIL::devil_file_new_name[j];}
		header_buffer[16]=DEVIL::devil_file_new_addr%256;	// code loading address low byte
		header_buffer[17]=DEVIL::devil_file_new_addr/256;	// code loading address high byte

		// =============== find the .TAP block selected in the GUI list, if any ======================================
		for (i=0; i<DEVIL::tap_idx; i++) {
			if (listBox1->isSelected(i)) {break;}		// i stores the index of selected list item which is the same as the block index in .TAP file
		}							// i is used next when accessing listItems[i]
		// --------------- extract data length value from selected item in GUI list ----------------------------------
		for (unsigned char j=0; j<5; j++) {data_len[j]=listItems[i]->text().toAscii().constData()[j+23];}  data_len[5]='\0';
		d_len=atoi(data_len)-2;				// data length stored in header (size of selected headerless block minus the first 2 bytes)
		// -----------------------------------------------------------------------------------------------------------
		header_buffer[14]=d_len%256;			// DL0 byte
		header_buffer[15]=d_len/256;			// DL1 byte
		header_buffer[18]=0x00;				// P20 byte
		header_buffer[19]=0x80;				// P21 byte
		// --------------- compute header block checksum in the last byte of header_buffer[] -------------------------
		header_buffer[20]=0; for (unsigned char j=0; j<18; j++) {header_buffer[20] ^= header_buffer[j+2];}	// checksum for header block
		/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

		// =============== write (insert) contents of header_buffer[] into the .TAP file =============================
		//  DEVIL::tap_file_name already initialized by openTap(), inserttapblock() will open and close the .TAP file
		ret_code = DEVIL::inserttapblock(DEVIL::tap_index[i], 21, header_buffer);

		if (ret_code) {w4.textLabel1->setText(QString("").sprintf("Error %d creating header for data block %04d", ret_code, i)); msg.exec(); return;}
		updateStatusMessage( QString("").sprintf("Successfully created header for data block %04d", i) );

		// =============== clear selection in GUI list of tape items (only 1 item was selected) ======================
		listBox1->setSelected(listItems[i],false);

		/* -------------------------------- UPDATE GUI WITH NEW TAPE CONTENTS ------------------------------------- */
		// =============== delete previous tape items if present, before displaying new tape contents ================
		if (DEVIL::tap_idx > 0) {
			for (short p=DEVIL::tap_idx-1; p>=0; p--) {delete listItems[p]; listItems.pop_back();}
			free(DEVIL::tap_index);
			DEVIL::tap_idx = 0; }

		// =============== re-read the .TAP file currently open ======================================================
		DEVIL::devilfilewrite(DEVIL::tap_file_name, 0, 0);

		// =============== populate listbox with list of data blocks in .TAP file ====================================
		for (unsigned short p=0; p<DEVIL::tap_idx; p++) {
			listItems.push_back(new Q3ListBoxText(listBox1, QString(DEVIL::tap_list[p]).remove(QRegExp("[\\n\\t\\r]")) ));
		}

		// =============== clear and release memory allocated to dynamic char* tap_list[] by devilfilewrite() ========
		for (int p=DEVIL::tap_idx-1; p>=0; p--) { free(DEVIL::tap_list[p]); }
		free(DEVIL::tap_list);
		// ===========================================================================================================
		/* -------------------------------------------------------------------------------------------------------- */
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton8->clicked() ("Create Floppy Image").
// Creates an empty formatted Devil floppy image.
void MainWindow::createImage() {

	int ret_code;

	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("Error");

	// =============== get the directory to save the floppy image in =============================================
	QString d = QFileDialog::getExistingDirectory(
		0,
		"Select directory to save floppy image",
		"/home/knoppix",
		QFileDialog::DontUseNativeDialog);	// ATTENTION!!!  this will return a QString ending in "/"	// NOT ANYMORE IN QT4 !!!
	if (d==QString::null) {return;}			// if user clicked "Cancel" abort

	// =============== get a file name for the .IMG file to be saved =============================================
	SaveNameDialog w9; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w9.setupUi(&dlg);		// create the SaveNameDialog window
	w9.setWindowTitle("Floppy image file name:"); w9.textLabel3->setText(".img");
	if (!w9.exec()) {return;}			// if OK clicked, DEVIL::save_file_name will contain the .IMG new file name

	// =============== check if destination file already exists ==================================================
	if (DEVIL::xis_dir((d+QString("/")+QString(DEVIL::save_file_name)+QString(".img")).toAscii().constData())!=-1) {printf("Error: file '%s' already exists. Floppy image creation aborted.\n", (d+QString("/")+QString(DEVIL::save_file_name)+QString(".img")).toAscii().constData()); return;}

	// =============== copy contents of DEVIL::save_file_name to a local QString ===================================
	QString save_file_name(DEVIL::save_file_name);

	// =============== clear memory allocated to DEVIL::save_file_name by the SaveNameDialog dialog =================
	if (DEVIL::save_file_name_allocated) {free(DEVIL::save_file_name); DEVIL::save_file_name_allocated=false;}

	// =============== create the new floppy image with user chosen path & name ==================================
	ret_code = DEVIL::create_diskimg((d+QString("/")+save_file_name+QString(".img")).toAscii().constData());

	// =============== display success/failure message depending on return code ==================================
	if (!ret_code) {
		printf("DEVIL floppy image %s successfully created\n\n", (d+QString("/")+save_file_name+QString(".img")).toAscii().constData());
		updateStatusMessage( QString("").sprintf("DEVIL floppy image %s successfully created.", (save_file_name+QString(".img")).toAscii().constData()) );
	} else {
		switch (ret_code) {
			case 1: w4.textLabel1->setText("Error creating floppy image"); break;
			case 2: w4.textLabel1->setText("Error writing to floppy image");
		}
		msg.exec();
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton9->clicked() ("Create Tape File").
// Creates a .TAP file from the contents of a binary file (one headerless block with user chosen flag byte)
void MainWindow::createTape() {

	int ret_code;

	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("Error");

	// =============== select a source binary file having contents of new .TAP file ==============================
	QString s = QFileDialog::getOpenFileName(
		0,
		"Create .TAP - Select source .BIN file",
		"/home/knoppix",
		"Binary files (*.bin)",
		0,
		QFileDialog::DontUseNativeDialog);
	if (s==QString::null) {return;}

	// =============== check size of source binary file (should be < 65535) ======================================
	QFileInfo f1(s);
	if (f1.size()>=65536) {
		printf("Binary file size must be less than 64 KB\n");
		w4.textLabel1->setText("Binary file size must be less than 64 KB"); msg.exec(); return;
	}

	// =============== get a value for flag byte of headerless block to be created in new .TAP file ==============
	FlagByteDialog w5; //QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w5.setupUi(&dlg);	// create the Flag Byte dialog
	if (!w5.exec()) {return;}		// this will set DEVIL::devil_tap_block_flag if OK is clicked

	// =============== select a location and name for the new .TAP file ==========================================
	QString d = QFileDialog::getSaveFileName(
		0,
		"Save .TAP file as:",
		"/home/knoppix",
		"Tape files (*.tap)",
		0,
		QFileDialog::DontUseNativeDialog);
	if (d==QString::null) {return;}
//	if (strcmp(d.right(4),".tap")!=0) {d=d+QString(".tap");}
	if (!d.endsWith(".tap")) {d.append(".tap");}

	// =============== create the .TAP file from data in the source binary file ==================================
	ret_code = DEVIL::create_tapfile(s.toAscii().constData(), d.toAscii().constData(), DEVIL::devil_tap_block_flag);

	// =============== display success/failure message depending on return code ==================================
	if (!ret_code) {
		QFileInfo f2(d);
		printf(".TAP file '%s' successfully created\n", f2.fileName().toAscii().constData());
		updateStatusMessage( QString("").sprintf(".TAP file '%s' successfully created.", f2.fileName().toAscii().constData()) );
	} else {
		switch (ret_code) {
			case 1: w4.textLabel1->setText("Error opening .BIN file"); break;
			case 2: w4.textLabel1->setText("Error creating .TAP file");
		}
		msg.exec();
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton10->clicked() ("Interleave Image").
void MainWindow::interleave() {

	int ret_code;

	// =============== get the full path of floppy image to open =================================================
	QString s = QFileDialog::getOpenFileName(
		0,
		"Select a deinterleaved floppy image",
		"/home/knoppix",
		"Floppy disk images - deinterleaved (*.dimg)",
		0,
		QFileDialog::DontUseNativeDialog);
	if (s==QString::null) {return;}

	QFileInfo f(s);							// get all info for the floppy image file in QFileInfo object f

	// =============== create the error message popup in case we need it or simply for testing ===================
	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("File Error!");

	// =============== checking disk image file name and size ====================================================
	if (DEVIL::dotcount(f.fileName().toAscii().constData()) > 1) {		// floppy image file should only have 1 dot in its name
		printf("'%s': No multiple dots allowed in disk image file name.\n\n", f.fileName().toAscii().constData());
		w4.textLabel1->setText(QString("").sprintf("'%s':\nNo multiple dots allowed\nin disk image file names.", f.fileName().toAscii().constData()));
		msg.exec();
		return;
	}
	w4.textLabel1->setText(QString("").sprintf("'%s':\nDeinterleaved disk image file names\nmust have the extension .dimg.", f.fileName().toAscii().constData()));
	if (DEVIL::dotcount(f.fileName().toAscii().constData()) == 0) {		// floppy image file name should end in ".img" so should contain 1 dot
		printf("'%s': Deinterleaved disk image file names must have the extension .dimg.\n\n", f.fileName().toAscii().constData());
		msg.exec();
		return;
	}
	if ( (DEVIL::dotcount(f.fileName().toAscii().constData()) == 1) &&		// floppy image file name should end in ".img"
	!(strstr(f.fileName().toAscii().constData(), ".dimg") && (DEVIL::dotpos(f.fileName().toAscii().constData()) == strlen(f.fileName().toAscii().constData())-5)) ) {
		printf("'%s': Deinterleaved disk image file names must have the extension .dimg.\n\n", f.fileName().toAscii().constData());
		msg.exec();
		return;
	}
	DEVIL::test_stream = fopen (s.toAscii().constData(), "r");
	if (DEVIL::test_stream == NULL) {
		printf ("Error %d opening disk image file in R/O mode: %m\n\n", errno);
		w4.textLabel1->setText(QString("").sprintf("Error %d opening disk image file in R/O mode:\n%m", errno));
		msg.exec();
		return;
	} else {fclose(DEVIL::test_stream);}
	if (f.size() != 737280) {
		printf ("Invalid DEVIL disk image file specified - file size = %u\n\n", (unsigned int)f.size());
		w4.textLabel1->setText(QString("").sprintf("Invalid DEVIL disk image file:\nfile size = %u", (unsigned int)f.size()));
		msg.exec();
		return;
	}

	// =============== get the directory to save the interleaved floppy image in =================================
	QString d = QFileDialog::getExistingDirectory(
		0,
		"Select directory to save in:",
		"/home/knoppix",
		QFileDialog::DontUseNativeDialog);	// ATTENTION!!!  this will return a QString ending in "/"	// NOT ANYMORE IN QT4 !!!
	if (d==QString::null) {return;}			// If user clicked "Cancel" abort
	d+=QString("/");

	QFileInfo g(d);					// get all info for the target dir in QFileInfo object g

	// =============== interleave floppy image and save it in target directory chosen ============================
	ret_code = DEVIL::interleave(s.toAscii().constData(), d.toAscii().constData());

	// =============== display success/failure message depending on return code ==================================
	if (!ret_code) {
		printf("Interleaved image '%s' successfully saved\n\n", f.fileName().replace(".dimg",".img").toAscii().constData());
		updateStatusMessage( QString("").sprintf("Interleaved image '%s' successfully saved", f.fileName().replace(".dimg",".img").toAscii().constData()) );
	} else {
		switch (ret_code) {
			case 1: w4.textLabel1->setText("De-interleaved disk image file name\nmust have the extension .dimg"); break;
			case 2: w4.textLabel1->setText("Error opening de-interleaved disk image."); break;
			case 3: w4.textLabel1->setText("Target file already exists.\nFile creation aborted."); break;
			case 4: w4.textLabel1->setText("Error creating interleaved disk image"); break;
		}
		msg.exec();
	}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to pushButton11->clicked() ("De-Interleave Image").
void MainWindow::deinterleave() {

	int ret_code;
	QFileInfo f(DEVIL::diskimg_file_name);

	// =============== create the error message popup in case we need it or simply for testing ===================
	Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
	msg.setWindowTitle("File Error!");

	// =============== get the directory to save the interleaved floppy image in =================================
	QString d = QFileDialog::getExistingDirectory(
		0,
		"Select directory to save in:",
		"/home/knoppix",
		QFileDialog::DontUseNativeDialog);	// ATTENTION!!!  this will return a QString ending in "/"	// NOT ANYMORE IN QT4 !!!
	if (d==QString::null) {return;}			// If user clicked "Cancel" abort
	d+=QString("/");

	QFileInfo g(d);					// get all info for the target dir in QFileInfo object g

	// =============== deinterleave floppy image and save it in target directory chosen ==========================
	ret_code = DEVIL::deinterleave(d.toAscii().constData());

	// =============== display success/failure message depending on return code ==================================
	if (!ret_code) {
		printf("De-interleaved image '%s' successfully saved\n\n", f.fileName().replace(".img",".dimg").toAscii().constData());
		updateStatusMessage( QString("").sprintf("De-interleaved image '%s' successfully saved", f.fileName().replace(".img",".dimg").toAscii().constData()) );
	} else {
		switch (ret_code) {
			case 1: w4.textLabel1->setText("Interleaved disk image file name\nmust have the extension .img"); break;
			case 2: w4.textLabel1->setText("Error opening interleaved disk image."); break;
			case 3: w4.textLabel1->setText("Target file already exists.\nFile creation aborted."); break;
			case 4: w4.textLabel1->setText("Error creating deinterleaved disk image"); break;
		}
		msg.exec();
	}

}

// ################################### CHECKBOX ACTIONS ########################################

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// It is connected to checkBox1->stateChanged(int).
// Enables or disables verbose operation mode
void MainWindow::setVerbose(int verbose) {
	if (verbose) {DEVIL::verbose=1;} else {DEVIL::verbose=0;}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to checkBox2->stateChanged(int).
// Enables or disables the saving of a log file for certain operations
void MainWindow::setLogOperation(int log) {
	if (log) {DEVIL::log_on=1;} else {DEVIL::log_on=0;}
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to checkBox3->stateChanged(int).
// Enables or disables console output of low-level info during certain operations
void MainWindow::setDebugMode( int dbg ) {
	if (dbg) {DEVIL::debug=1; qInstallMsgHandler(0);} else {DEVIL::debug=0; qInstallMsgHandler(noMessageOutput);}
}

void MainWindow::showLegend(int on) {
	if (textLabel1->text()[2] == 'N') {		// ################## IF FLOPPY IMAGE OPEN ##################
		if (on) {frame1->show();} else {frame1->hide();}
	} else {					// ################## IF .TAP FILE OPEN #####################
		if (on) {frame5->show();} else {frame5->hide();}
	}
}

void MainWindow::help() {
	if (!DEVIL::hlp_open) {
		DEVIL::hlp_open = true;
//		HelpWindow w6; hlpp = new QDialog(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); hlpp->setAttribute(Qt::WA_DeleteOnClose); w6.setupUi(hlpp);			// create the help window
		hlpp = new HelpWindow;
		hlpp->show();
	} else {
		Ui::ErrorPopup w4; QDialog msg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w4.setupUi(&msg);	// create the error popup message window
		msg.setWindowTitle("Help Yourself!");
		w4.textLabel1->setText("The Help window is already open!");
		msg.exec();
	}
}

// ################################### STATUS BAR ACTIONS ######################################

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Used as a plain C++ function, called from various points to display status bar messages about current operations
// Updates the status bar message stored in textLabel5s.
void MainWindow::updateStatusMessage(QString message) {
	textLabel5s->setText( message );
}

// Custom slot added to MainWindow from the menu (Edit->Slots...) or by directly editing the MainWindow.ui XML file.
// Connected to listBox1->selectionChanged().
// It updates textLabel3s and textLabel4s when items are selected in the disk image file list
void MainWindow::updateSelectionSize() {
	if (textLabel1->text()[2] == 'N') {		// ################## IF FLOPPY IMAGE OPEN ##################
		selected_items=0;	// global to MainWindow
		nr_of_ranges=0;		// global to MainWindow
		int prev_selected=-1;
		int selected_bs=0, bs; char bs_c[7];  bs_c[6]='\0'; char type[5]; type[4]='\0';
		int selected_kbsd=0, kbsd; char kbsd_c[4]; kbsd_c[3]='\0';
		char cat_idx[3]; cat_idx[2]='\0';
		/* --------------------- get catalog entry ranges of selected entries -------------------------- */
		for (unsigned short i=0; i<DEVIL::nr_files+DEVIL::nr_hless; i++) {
			if (listBox1->isSelected(listItems[i])) {
				memcpy(&cat_idx[0], &DEVIL::file_list[i][1], 2);	// read catalog entry index as text in cat_idx
				memcpy(&type[0], &DEVIL::file_list[i][7], 4);		// read data type as text in type

				if (strcmp(type,"----")) {
					if (prev_selected==-1) {nr_of_ranges=1; RangeList[0][0]=atoi(cat_idx);} else
					if (prev_selected<atoi(cat_idx)-1) {RangeList[nr_of_ranges-1][1]=prev_selected; nr_of_ranges++; RangeList[nr_of_ranges-1][0]=atoi(cat_idx);}
				}
				for (int j=0; j<5; j++) {bs_c[j]=DEVIL::file_list[i][27+j];}
				bs=atoi(bs_c);
				selected_bs+=bs;
	//			for (int j=0; j<3; j++) {kbsd_c[j]=DEVIL::file_list[i][35+j];}
				kbsd = (bs/(9*1024) + (bs%(9*1024) ? 1 : 0)) * 9;
				selected_kbsd+=kbsd; selected_items++; if (strcmp(type,"----")) {prev_selected=atoi(cat_idx);}
			}
	//		if (i==DEVIL::nr_files-1) {RangeList[nr_of_ranges-1][1]=prev_selected;}
		}
		if (nr_of_ranges > 0) {RangeList[nr_of_ranges-1][1]=prev_selected;}
		/* --------------------------------------------------------------------------------------------- */
		if ((!selected_items) || (!nr_of_ranges)) {
			pushButton3->setEnabled(FALSE); menuExtractAction->setEnabled(FALSE);
			pushButton6->setEnabled(FALSE); menuDeleteAction->setEnabled(FALSE);
			pushButton7->setEnabled(FALSE); menuChange_AttrsAction->setEnabled(FALSE);
			textLabel3s->clear();textLabel4s->clear();
		} else {
			pushButton3->setEnabled(TRUE); menuExtractAction->setEnabled(TRUE);
			pushButton6->setEnabled(TRUE); menuDeleteAction->setEnabled(TRUE);
			pushButton7->setEnabled(TRUE); menuChange_AttrsAction->setEnabled(TRUE);
			textLabel3s->setText( QString(" ") + QString("%1").arg(selected_items,2) + QString(" / ") + QString("%1").arg(nr_of_ranges,2) + QString(" ") );
			textLabel4s->setText( QString(" ") + QString("%1").arg(selected_bs,6) + QString(" B / ") + QString("%1").arg(selected_kbsd,3) + QString(" KB ") );
		}
		if ((selected_items==1) && (strcmp(type,"----"))) {	// only activate "Rename" button if one item selected - non-headerless
			pushButton5->setEnabled(TRUE); menuRenameAction->setEnabled(TRUE);
		} else {
			pushButton5->setEnabled(FALSE); menuRenameAction->setEnabled(FALSE);
		}
	} else {					// ################## IF .TAP FILE OPEN ##################
		unsigned int selected_idx=0;
		char form[5], flag[3], hles[26];
		selected_items=0;	// global to MainWindow
		for (unsigned int i=0; i<DEVIL::tap_idx; i++) {
			if (listBox1->isSelected(i)) {
				selected_idx = i;
				selected_items++;
			}
		}
		if (selected_items) {
			textLabel3s->setText( QString(" ") + QString("%1").arg(selected_items) + QString(" ") );
		} else {textLabel3s->setText("");}
		if (selected_items<=1) {pushButton4->setEnabled(TRUE); menuInsertAction->setEnabled(TRUE);} else {pushButton4->setEnabled(FALSE); menuInsertAction->setEnabled(FALSE);}
		if (selected_items==1) {
			textLabel4s->setText( QString("").sprintf("Block start: %X", DEVIL::tap_index[selected_idx]) );

			for (unsigned char i=0; i< 4; i++) {form[i]=listItems[selected_idx]->text().toAscii().constData()[i];}     form[4]='\0';
			for (unsigned char i=0; i< 2; i++) {flag[i]=listItems[selected_idx]->text().toAscii().constData()[i+19];}  flag[2]='\0';
			for (unsigned char i=0; i<25; i++) {hles[i]=listItems[selected_idx]->text().toAscii().constData()[i+30];}  hles[25]='\0';

			pushButton3->setEnabled(TRUE);		// enable "Extract" button
			menuExtractAction->setEnabled(TRUE);	// enable "Extract" menu item
			pushButton6->setEnabled(TRUE);		// enable "Delete" button
			menuDeleteAction->setEnabled(TRUE);	// enable "Delete" menu item

			if (strcmp(form,"Head")==0) {pushButton5->setEnabled(TRUE); menuRenameAction->setEnabled(TRUE);} else {pushButton5->setEnabled(FALSE); menuRenameAction->setEnabled(FALSE);}	// enable/disable "Rename" button & menu item
			if (strcmp(form,"Data")==0 && strcmp(flag,"FF")==0 && strcmp(hles,"--------- headerless data")==0) {
				pushButton7->setEnabled(TRUE);	// --------- enable "Create Header" button
			} else {
				pushButton7->setEnabled(FALSE);	// --------- disable "Create Header" button
			}
		} else {
			pushButton3->setEnabled(FALSE);		// disable "Extract" button
			menuExtractAction->setEnabled(FALSE);	// disable "Extract" menu item
			pushButton5->setEnabled(FALSE);		// disable "Rename" button
			menuRenameAction->setEnabled(FALSE);	// disable "Rename" menu item
			pushButton6->setEnabled(FALSE);		// disable "Delete" button
			menuDeleteAction->setEnabled(FALSE);	// disable "Delete" menu item
			pushButton7->setEnabled(FALSE);		// disable "Create Header" button
			menuChange_AttrsAction->setEnabled(FALSE);	// disable "Create Header" menu item

			textLabel4s->setText("");
		}
	}
}

void MainWindow::closeEvent (QCloseEvent *event) {

	if (DEVIL::hlp_open) {
		hlpp->close();
	}

	Ui::ConfirmDialog w1; QDialog dlg(0, Qt::CustomizeWindowHint | Qt::WindowTitleHint); w1.setupUi(&dlg);	// create the confirmation dialog
	dlg.setWindowTitle("Are you sure?");
	w1.textLabel1->setText("Are you sure you want to quit?");
	w1.pushButton1->setText("Yes");
	if (dlg.exec()) {
		event->accept();
	} else {
		event->ignore();
	};
}
