#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <vector>

#include <QDialog>

#include "ui_MainWindow.h"

#include "ui_ConfirmDialog.h"
#include "CreateHeader/CreateHeader.h"
#include "DevAttrDialog/DevAttrDialog.h"
#include "ui_ErrorPopup.h"
#include "FlagByteDialog/FlagByteDialog.h"
#include "HelpWindow/HelpWindow.h"
#include "OptionDialog/OptionDialog.h"
#include "RenameDialog/RenameDialog.h"
#include "SaveNameDialog/SaveNameDialog.h"

//#include <QApplication>
#include <QLabel>
#include <QLineEdit>
#include <QStatusBar>
#include <QFileDialog>
#include <QString>
#include <QListWidget>
#include <QRegExp>
#include <QRadioButton>
#include <QFont>
#include <QToolTip>
#include <Qt3Support/Q3ListBoxText>
/*
#include <qcstring.h>	// required for QByteArray
 */
//#include <QFileInfo>

#include <sys/stat.h>	// required by stat type
#include <cstdlib>	// required by free() function
#include <cerrno>	// required by system var errno (?)
#include <cstring>

namespace DEVIL {

	extern struct stat st;
	extern unsigned long size;			// file size of filename argument provided
	extern unsigned char buffer[];			// used as write data when generating formatted DEVIL disk image

	extern bool path_to_dest_dir_base_allocated;
	extern char* path_to_dest_dir_base;		// Full path to dest_dir_base (excluding dest_dir_base) - allocated in main() and GUI

	extern bool dest_dir_base_allocated;
	extern char* dest_dir_base;			// disk image saving dir, set by diskimg_dir() to "path_to_dest_dir_base/diskimg_file_name"

	extern bool diskimg_file_name_allocated;
	extern char* diskimg_file_name;			// Full path of floppy image file currently open

	extern int diskimg_file_descrptr;		// disk image file descriptor
	extern FILE* test_stream;			// disk image file stream - used only for initial check in main()

	extern bool  save_file_name_allocated;
	extern char* save_file_name;

	extern bool tap_file_name_allocated;
	extern char* tap_file_name;			// Full path of .TAP file currently open, allocated in devilfilecopy() and openTap()

	extern char devil_file_name[11];
	extern char devil_file_new_name[11];
	extern unsigned short devil_file_new_addr;

	extern unsigned char R_W, SYS;

	extern unsigned int dotpos (const char* string);
	extern unsigned int dotcount (const char* string);
	extern int devildirdump();

	/* ############################### variables used by devildirdump() #################################### */
	extern unsigned char devildir0_mem[19712];	// 77 dir sectors x 1 entry/sector x 18 bytes/entry = 1386 bytes	/* stores the entire directory */
	extern unsigned char devildir1_mem[19712];	// 77 dir sectors x 1 entry/sector x 19 bytes/entry = 1463 bytes (one extra byte to store index of dir entry)	/* stores only used & relevant dir entries (not deleted and !=E4) */
	extern int nr_recs;				// nr of allocated dir entries (records)
	extern unsigned int nr_files;			// nr of distinct catalog entries (nr_files < nr_recs <= 77)
	extern unsigned int nr_hless;			// nr of headerless blocks (if any) attached to catalog entries
	extern int total_size;				// total size of data in files (in tape data blocks)
	extern char alloc_dir[77];			// map of free and allocated dir entries (records)
	extern unsigned char free_dir_recs;		// nr. of free directory entries (records)
	extern char* file_list[77*118];			// list of file details as text to use in Qt listBox		//# asprintf()-ed by devildirdump(); needs cleanup -> done in devildirdump()

	/* ############################### variables used by devilfilewrite() & writeblock() ###################### */
	extern unsigned int tap_idx;			// tape block counter (individual headers and non-headers)
	extern unsigned short dsk_idx;			// header counter
	extern char** tap_list;				// list of .TAP file blocks as text to use in Qt listBox
	extern unsigned int* tap_index;			// dynamic list of .TAP file block start addresses
	extern signed char blocks_needed;

	extern unsigned char tap_list_mode;		// 1 if -t option passed in command line, 0 if not

	extern bool hlp_open;				// true if help window open, false if not

	extern unsigned char devil_tap_block_flag;
	extern unsigned char range1, range2;		// range limits for files to extract

	extern unsigned char log_on;			// 0 if no log to generate, 1 if log
	extern unsigned char verbose;			// 0 if no detailed messages to generate, 1 if yes
	extern unsigned char debug;			// debug mode flag (print extra info): 0 = normal mode, 1 = debug mode

	extern int xis_dir (const char* d);
	extern int diskimg_dir ();
	extern int create_diskimg(const char* file_name);
	extern void defrag_disk();
	extern int devilfilecopy(unsigned char range1, unsigned char range2, const char* tap_name, bool overwrite);
	extern int extracttapblock(const char* bin_file_name, unsigned int tap_file_address, unsigned int data_length);
	extern int inserttapblock(unsigned int tap_file_address, unsigned int tap_block_length, unsigned char* buffer);
	extern int inserttapfile(unsigned int tap_file_address, const char* bin_file_name);
	extern int insertbinblock(unsigned int tap_file_address, const char* bin_file_name, unsigned char flag);
	extern int tapblockdelete(unsigned int tap_file_address);
	extern int readtapdata(unsigned int tap_file_address, unsigned int tap_data_length, unsigned char* tap_buffer);
	extern int writetapdata(unsigned int tap_file_address, unsigned int tap_data_length, unsigned char* tap_buffer);
	extern int create_tapfile(const char* bin_file_name, const char* tap_file_name, unsigned char flag);
	extern signed char devilfilewrite(const char* tap_file_name, unsigned char write, unsigned char block_no);
	extern int devilfilerename(char* new_name, unsigned char idx);
	extern int devilfilechattr(unsigned char attr, unsigned char idx1);
	extern int devilfiledelete(unsigned char range1, unsigned char range2);
	extern int devilfilesetattr(unsigned char range1, unsigned char range2, char* sys);
	extern int interleave(const char* dintl_file_name, const char* target_dir);
	extern int deinterleave(const char* target_dir);

	extern unsigned char option_selected;			// used by OptionDialog
}

class MainWindow : public QMainWindow, public Ui::MainWindow {
	Q_OBJECT
    private:
	QDialog* hlpp;	//HelpWindow* hlpp;

	QString Push1_Tip;
	QString Push2_Tip;
	QString Push3_1_Tip;
	QString Push3_2_Tip;
	QString Push4_1_Tip;
	QString Push4_2_Tip;
	QString Push5_1_Tip;
	QString Push5_2_Tip;
	QString Push6_1_Tip;
	QString Push6_2_Tip;
	QString Push7_1_Tip;
	QString Push7_2_Tip;
	QString Push8_Tip;
	QString Push9_Tip;
	QString Push10_Tip;
	QString Push11_Tip;

	QString Check1_Tip;
	QString Check2_Tip;
	QString Check3_Tip;
	QString Check4_Tip;

	QString StatB1_1_tip;
	QString StatB1_2_tip;
	QString StatB2_1_tip;
	QString StatB2_2_tip;

	QLabel* textLabel3s;	// will display number of selected list items
	QLabel* textLabel4s;	// will display total size of selected list items
	QLabel* textLabel5s;	// will display status bar messages

	//Q3ListBoxText* listItems[500];		// this should have a dynamic size, better implement it with vector
	std::vector<Q3ListBoxText*> listItems;

	unsigned char RangeList[40][2];
	int nr_of_ranges;
	unsigned char selected_items;		// number of selected items in listBox1, used by updateSelectionSize()

	char fragview[78];
    private slots:
	void openImage();
	void openTap();
	void extractFiles();
	void insertFiles();
	void renameFile();
	void deleteFiles();
	void changeAttrs();
	void createImage();
	void createTape();
	void interleave();
	void deinterleave();
	void setVerbose(int verbose);
	void setLogOperation(int log);
	void setDebugMode(int dbg);
	void showLegend(int on);
	void help();
	void updateStatusMessage(QString message);
	void updateSelectionSize();
	void closeEvent (QCloseEvent *event);
    public:
	MainWindow();
};

#endif
