// C headers
// _GNU_SOURCE required by asprintf()
// #define _GNU_SOURCE
#include <stdio.h>
#include <sys/stat.h>	/* required by stat type */
#include <stdlib.h>	/* required by exit() function */
#include <errno.h>	/* required by system var errno (?) */
#include <string.h>	/* required by string functions starting named "str*" */
#include <ctype.h>	/* required by isdigit() */
#include <fcntl.h>	/* required by low-level file access */
#include <unistd.h>	/* required by low-level file deletion function "unlink" */
#include <libgen.h>	/* required by dirname and basename */
#include <sys/types.h>	/* required by opendir() */
#include <dirent.h>	/* required by opendir() */



namespace DEVIL {

	#include "devil_usage.h"

	char exe_full_path[FILENAME_MAX];
	struct stat st;
	unsigned long size = 0;			// file size of filename argument provided
	unsigned char buffer[] = "\xE5\xE5\xE5\xE5\xE5\xE5\xE5\xE5\xE5\xE5";	// used as write data when generating formatted DEVIL disk image

	bool path_to_dest_dir_base_allocated=false;
	char* path_to_dest_dir_base;		// Absolute path to dest_dir_base (excluding dest_dir_base) - initialized in main() and GUI
						// - in command line mode it is initialized in main() to current directory (.)
						// - in GUI mode it is initialized to the dir selected by user when clicking "Extract to .TAP"
	bool dest_dir_base_allocated=false;
	char* dest_dir_base;			// .TAP file saving dir, set by diskimg_dir() to "path_to_dest_dir_base/diskimg_file_name"
						// before extracting to .TAP, where "diskimg_file_name" had the .img extension removed

	bool diskimg_file_name_allocated=false;	// Status flag to indicate if memory is allocated to "diskimg_file_name".
	char *diskimg_file_name;		// Full path of disk image file to operate on.
						// Initialized in main() to argv[2] and in MainWindow::openImage() to user choice

	int diskimg_file_descrptr;		// Disk image file descriptor
	FILE* test_stream;			// Disk image file stream - used only for initial check in main()

	char devil_file_name[11] = {'\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0'};
	char devil_file_new_name[11] = {'\0','\0','\0','\0','\0','\0','\0','\0','\0','\0','\0'};
	unsigned short devil_file_new_addr=0;	// code loading address when creating header for headerless .TAP file block

	bool save_file_name_allocated=false;
	char* save_file_name;			// Full path of file to be saved (without extension)

	bool tap_file_name_allocated=false;
	char* tap_file_name;			// Full path of .TAP file currently open, allocated in devilfilecopy() and MainWindow::openTap()

	bool tap_log_file_name_allocated=false;
	char* tap_log_file_name;

	int tap_file_descrptr;			// .TAP file descriptor
	FILE* tap_log_file_stream;		// .TAP log file stream

	int error_code;
	unsigned int i, j, k;
	unsigned char R_W, SYS;

	/* ############################### variables used by readblock() ##################################### */
	unsigned short track_length = 9216;	// length of a track in bytes (2 sides x 18 sectors x 256 bytes)
	unsigned short sstracklength = 4608;	// length of a track side in bytes (18 sectors x 256 bytes)
	int  xlt[18] = {0, 2, 4, 6, 8,10,12,14,16, 1, 3, 5, 7, 9,11,13,15,17};	// used for computing the start of sectors
	int xlt1[36] = {1,10, 2,11, 3,12, 4,13, 5,14, 6,15, 7,16, 8,17, 9,18,19,28,20,29,21,30,22,31,23,32,24,33,25,34,26,35,27,36};		// used for computing the physical sector number
	const int xlt2[36] = {0, 2, 4, 6, 8,10,12,14,16, 1, 3, 5, 7, 9,11,13,15,17,18,20,22,24,26,28,30,32,34,19,21,23,25,27,29,31,33,35};	// used for computing the start of sectors in an entire track

	/* ############################### variables used by devildirdump() #################################### */
//	unsigned char devildir0_mem[1386];	// 77 dir sectors x 1 entry/sector x 18 bytes/entry = 1386 bytes	/* stores the entire directory */
	unsigned char devildir0_mem[19712];	// 77 dir sectors x 1 entry/sector x 256 bytes/entry = 19712 bytes	/* stores the entire directory */
//	unsigned char devildir_mem[1463];	// 77 dir sectors x 1 entry/sector x 19 bytes/entry = 1463 bytes (one extra byte to store index of dir entry) */	/* stores only used dir entries (not deleted) */
	unsigned char devildir_mem[19712];	// 77 dir sectors x 1 entry/sector x 256 bytes/entry = 19712 bytes (first byte to store index of dir entry) */		/* stores only used dir entries (not deleted) */
//	unsigned char devildir1_mem[1463];	// 77 dir sectors x 1 entry/sector x 19 bytes/entry = 1463 bytes (one extra byte to store index of dir entry) */	/* stores only used & relevant dir entries (not deleted and !=E4) */
	unsigned char devildir1_mem[19712];	// 77 dir sectors x 1 entry/sector x 256 bytes/entry = 19712 bytes (first byte to store index of dir entry) */		/* stores only used & relevant dir entries (not deleted and !=E4) */
	int nr_recs=0;				// nr of allocated dir entries (records)
	unsigned int nr_files=0;		// nr of distinct catalog entries (nr_files <= nr_recs <= 77)
	unsigned int nr_hless=0;		// nr of headerless blocks (if any) attached to catalog entries
	int total_size=0;			// total size of data in files (in headed data blocks)
	char alloc_dir[77];			// map of free and allocated dir entries (records)
	unsigned char free_dir_recs=0;		// nr. of free directory entries (records)
	char* file_list[77*118];		// list of file details as text to use in Qt listBox (77 dir entries x max. 118 blocks/entry)		//# asprintf()-ed by devildirdump(); needs cleanup -> done in devildirdump()

	/* ############################### variables used by devilfilewrite() & writeblock() ###################### */
	unsigned char tap_file_buffer[0x10000];
	unsigned char track_buffer[9216];
	unsigned int tap_idx;			// tape block counter (individual headers and non-headers)
	unsigned short dsk_idx;			// headed block counter (data block together with a proper header block)
	char** tap_list;			// dynamic list of .TAP file blocks as text to use in Qt listBox
	unsigned int* tap_index;		// dynamic list of .TAP file block start addresses
	bool headerless=false;			// error flag, true if any headerless blocks found
	bool dataless=false;			// error flag, true if any header block found without its associated data block

	/* #################################################################################################### */

	unsigned char devil_tap_block_flag;	// flag byte value used when creating a new .TAP file
	unsigned char checksum, checksum_h;	//, checksum_d;		checksums for reading, header and data
	signed char blocks_needed;

	unsigned char range1, range2;		// range limits for files to extract
	char range1_c[5] = {'\0','\0','\0','\0','\0'}, range2_c[5] = {'\0','\0','\0','\0','\0'};	// range limits as 5-digit strings
	unsigned char dashpos;			// position of dash in string

	unsigned char create_mode=0;
	unsigned char defrag_mode=0;
	unsigned char extract_mode=0;
	unsigned char list_mode=0;
	unsigned char tap_list_mode=0;		// 1 if -t option passed in command line, 0 if not
	unsigned char write_mode=0;
	unsigned char delete_mode=0;
	unsigned char check_mode=0;
	unsigned char interleave_mode=0;
	unsigned char deinterleave_mode=0;
	unsigned char set_attr_mode=0;
	unsigned char log_on=0;			// 0 if no log to generate, 1 if log
	unsigned char verbose=0;		// 0 if no detailed messages to generate, 1 if yes
	unsigned char debug=0;			// debug mode flag (print extra info): 0 = normal mode, 1 = debug mode

	bool hlp_open = false;			// true if help window open, false if not

	unsigned char option_selected = 0;	// used by OptionDialog

	// Find the index of a dot char in a string
	unsigned int dotpos (const char* string) {
		char* e;
		int index;

		if (strchr(string, '.') == NULL) {return(0);}
		e = (char*)strchr(string, '.'); index = (int)(e - string);
		return(index);
	}

	// Count dots in a string
	unsigned int dotcount (const char* string) {
		unsigned short i, j=0;

		for (i=0; i<strlen(string); i++) {
			if (string[i] == '.') {j++;}
		}
		return(j);
	}

	// Test if dir "d" exists
	/* Return codes:
	 *  1 - Dir exists
	 * -1 - Dir does not exist
	 * -2 - Name exists but is not a dir
	 */
	int xis_dir (const char* d) {
		DIR *dirptr;

		if (access ( d, F_OK ) != -1) {			/* file exists */
			if ((dirptr = opendir (d)) != NULL) {
				closedir (dirptr); return 1;	/* d exists, and is dir */
			} else {return -2;}			/* d exists, but not dir */
		} else {return -1;}				/* d does not exist */
	}
	
	// Generate path of common directory for files extracted from disk image - in global string dest_dir_base.
	// "diskimg_file_name" and "path_to_dest_dir_base" must be already allocated.
	// If "dest_dir_base" is already allocated, it will be released first and then reallocated.
	/* Return codes:
	 * 0 - Success
	 * 1 - diskimg_file_name pointer not allocated
	 * 2 - path_to_dest_dir_base pointer not allocated
	 */
	int diskimg_dir () {
		unsigned short ix;
		char* dskimg_file_name;		// will store a backup of diskimg_file_name, to avoid corruption by basename()
		if (!diskimg_file_name_allocated) {printf("diskimg_dir(): diskimg_file_name pointer not allocated\n"); return 1;}
		if (!path_to_dest_dir_base_allocated) {printf("diskimg_dir(): path_to_dest_dir_base pointer not allocated\n"); return 2;}
		asprintf(&dskimg_file_name,"%s",diskimg_file_name);		// make backup copy

		ix = dotpos(basename(dskimg_file_name));			// get length of image file name without .img extension
		free(dskimg_file_name); asprintf(&dskimg_file_name,"%s",diskimg_file_name);	// make backup copy again

		if (dest_dir_base_allocated) free(dest_dir_base);
		dest_dir_base = (char*)malloc((strlen(path_to_dest_dir_base)+ix+2)*sizeof(char));
		dest_dir_base_allocated=true;

		sprintf(dest_dir_base, "%s/", path_to_dest_dir_base);		// start with absolute path to dest_dir_base
		strncat(dest_dir_base, basename(dskimg_file_name), ix);		// and then add the floppy image file name (without the .img extension) */
		free(dskimg_file_name);		// clear allocated backup of diskimg_file_name
		return 0;
	}

	// Check validity of a user-submitted range of catalog entries to operate on
	/* Return codes:
	 * 0 - Success
	 * 1 - Wrong syntax for file range
	 * 2 - Dash should be in between 2 numbers
	 * 3 - Multiple dashes in file range
	 * 4 - Order numbers can only have 2 digits
	 * 5 - First number > the second
	 * 6 - Order number has > 2 digits
	 * 7 - Wrong syntax for file range
	 */
	int range_check (char* arg) {
		dashpos=0;
		for (i=0; i<strlen(arg); i++) {
			if (!isdigit(arg[i]) && (arg[i] != '-')) {printf ("Wrong syntax for file range.\n"); return 1;}
			if ((arg[i] == '-') && ((i == 0) || (i == strlen(arg)-1))) {printf ("Wrong syntax for file range, dash should be in between 2 numbers.\n"); return 2;}
			if  (arg[i] == '-') {
				if (dashpos == 0) {dashpos = i;} else {printf ("No multiple dashes allowed in file range.\n"); return 3;}
			}
		}
		if ((dashpos > 2) || ((dashpos > 0) && (dashpos < strlen(arg)-3))) {printf("Catalog order numbers can only have 2 digits.\n"); return 4;}
		if (dashpos != 0) {	/* multiple catalog entries to be extracted */
			memcpy(&range1_c[0], &arg[0], dashpos); memcpy(&range2_c[0], &arg[0]+dashpos+1, strlen(arg)-dashpos-1);
			sscanf(range1_c, "%hhd", &range1); sscanf(range2_c, "%hhd", &range2);
			if (range1 >= range2) {printf("First number should be < the second.\n"); return 5;}
		} else {		/* one single catalog entry to be extracted */
			if (strlen(arg)>2) {printf("File order number can only have 2 digits.\n"); return 6;}
			for (i=0; i<strlen(arg); i++) { if (!isdigit(arg[i])) {printf ("Wrong syntax for file range.\n"); return 7;} }
			memcpy(&range1_c[0], &arg[0], strlen(arg));
			sscanf(range1_c, "%hhd", &range1); strcpy(range2_c, "-"); range2 = range1;
		}
		return 0;
	}

	/* Create an empty formatted floppy image */
	/* Return codes:
	 * 0 - Success
	 * 1 - File creation error
	 * 2 - Error writing to floppy image file
	 */
	int create_diskimg(const char* file_name) {
		int diskimg_file_descrptr, ret_code;

		umask(000);
		diskimg_file_descrptr = open (file_name, O_RDWR | O_CREAT, 00666);
		if (diskimg_file_descrptr == -1) {printf("File creation error %d: %m", errno); return 1;}
		for (int i=0; i<72*1024; i++) {ret_code=write (diskimg_file_descrptr, (void *) buffer, 10); if (ret_code!=10) break;}	/* writing 10 bytes at once */
		close(diskimg_file_descrptr);
		if (ret_code!=10) return 2;

		return 0;
	}


/*	######################################### STRUCTURE OF A DEVIL CATALOG ENTRY ##############################################

	CATALOG TAKES UP FIRST 3 TRACKS BOTH SIDES (TRACKS 0,1,2), HAS MAX. 77 ENTRIES. LOGICAL ALLOCATION UNIT IS 1 TRACK BOTH SIDES (9 KB).

	  00   01   02   03   04   05   06   07   08   09   10   11   12   13   14   15   16   17
	+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+
	| TY | N0 | N1 | N2 | N3 | N4 | N5 | N6 | N7 | N8 | N9 | D0 | D1 | A0 | A1 | B0 | B1 | SY |
	+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+----+

	Field meanings:

	TY = DATA TYPE (00 = program, 01 = number array, 02 = character array, 03 = code, E4 = >=2nd CATALOG ENTRY FOR SAME FILE, E5 = DELETED FILE)

	N0...N9 = DATA BLOCK NAME (10 characters, padded with trailing blanks)

	D0 = DATA BLOCK LENGTH (Low byte)	A0 = PARAMETER 1 (Low byte)	B0 = PARAMETER 2 (Low byte)
	D1 = DATA BLOCK LENGTH (High byte)	A1 = PARAMETER 1 (High byte)	B1 = PARAMETER 2 (High byte)

	SY = R/W-R/O, DIR-SYS ATTRIBUTES:	SY bit 1 = 0 FOR R/W,	SY bit 0 = 0 FOR DIR
						SY bit 1 = 1 FOR R/O	SY bit 0 = 1 FOR SYS

	########################################################################################################################### */


	/* Reads the CP/M directory (max. 77 entries of 18 bytes each) of the floppy image and stores:
	 * - all directory entries - in global var devildir0_mem[77x256]
	 * - all directory entries except deleted entries - in global var devildir_mem[77x256] (1st byte used as counter)
	 * - all directory entries except deleted entries and additional entries (byte0 = E4) - in global var devildir1_mem[77x256]
	 * Also it prints a listing of floppy contents to console */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Invalid disk image
	 * 3 - Correct exit in extract mode
	 * 4 - Invalid data type byte in dir entry
	 * 5 - Invalid attribute byte in dir entry
	 * 6 - Invalid unused byte in dir entry
	 */
	int devildirdump() {

		char *record, type[5], leng[6], par1[6], par2[6], /* varN[2], */ sys[8];
		char varN;
		unsigned char hless;				/* number of headerless blocks (if) attached after a headed block */
		unsigned char i;				/* counts all dir entries (0 to 76) */
		int diskimg_file_descrptr;			/* disk image file descriptor */
		unsigned char rb[256];				/* 256 byte read buffer */
		unsigned int dir_sect_start;
//		unsigned char dir_entry_idx;			/* counts all dir entries (0 to 76) */	// useless here, same as i
		char bad_unusareabyte_dirents[77*3+1];		/* stores numbers of dir entries (sectors) with non-zero bytes in unused area */
//		sprintf(bad_unusareabyte_dirents, "");		// this gives compiler warning (zero-length printf format string)
		bad_unusareabyte_dirents[0]='\0';
		char bad_attrbyte_dirents[77*3+1];		/* stores numbers of dir entries (sectors) with wrong attr bytes */
//		sprintf(bad_attrbyte_dirents, "");		// this gives compiler warning (zero-length printf format string)
		bad_attrbyte_dirents[0]='\0';
		char bad_datatypebyte_dirents[77*3+1];		/* stores numbers of dir entries (sectors) with wrong data type bytes */
//		sprintf(bad_datatypebyte_dirents, "");		// this gives compiler warning (zero-length printf format string)
		bad_datatypebyte_dirents[0]='\0';
		char bad_zeroentry_dirents[77*3+1];		/* stores numbers of dir entries (sectors) with 0x00 bytes in the first 18 bytes */
//		sprintf(bad_zeroentry_dirents, "");		// this gives compiler warning (zero-length printf format string)
		bad_zeroentry_dirents[0]='\0';

		/* ########################### OPEN DISK IMAGE FILE FOR READING ############################## */
		diskimg_file_descrptr = open (diskimg_file_name, O_RDONLY); if (diskimg_file_descrptr == -1) {printf("devildirdump() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}
		/* ########################################################################################### */

		for (i=0; i<77; i++) {alloc_dir[i]=0;}		/* clear dir allocation table */

		nr_recs=0; nr_files=0; nr_hless=0;
//		dir_entry_idx=0;
		free_dir_recs=77;
		if (debug) {printf ("\n                      DIRECTORY ENTRIES:\n===============================================================\n------------------- (logical sector order) --------------------\nRec#    Typ|             Name            | Len | Pa1 | Pa2 |At\n---------------------------------------------------------------\n");}

		for (i=0; i<77; i++) {
			dir_sect_start = (i/18)*sstracklength + xlt[i%18]*256;
			lseek (diskimg_file_descrptr, dir_sect_start, SEEK_SET);
			read (diskimg_file_descrptr, (void *) rb, 256);

			asprintf (&record, "%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X", \
				rb[0],rb[1],rb[2],rb[3],rb[4],rb[5],rb[6],rb[7],rb[8],rb[9],rb[10],rb[11],rb[12],rb[13],rb[14],rb[15],rb[16],rb[17]);	/* text hexdump of a 18-byte dir entry */
			// ======================== some disk format checking ============================
			if (strcmp(record, "00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00") == 0)
				{sprintf(bad_zeroentry_dirents, "%s %02d", bad_zeroentry_dirents, i); /* printf ("Invalid disk image - dir entries cannot be full of zeros\n"); return 2; */ }
			if (rb[0]>0x03 && rb[0]!=0xE4 && rb[0]!=0xE5) {sprintf(bad_datatypebyte_dirents, "%s %02d", bad_datatypebyte_dirents, i); /* printf("Invalid disk image - invalid data type byte in dir entry #%d\n", i); return 4; */ }
			if (rb[17]>0x03 && rb[17]!=0xE5) {sprintf(bad_attrbyte_dirents, "%s %02d", bad_attrbyte_dirents, i); /* printf("Invalid disk image - invalid attribute byte in dir entry #%d\n", i); return 5; */ }
			if (rb[0]<0x04 || rb[0]==0xE4) {
				// used to be p<255 but just found out these are reserved for possible headerless blocks (with flagbyte=FF) immediately following a headed block ...
				for (short p=18; p<20; p++) { if (rb[p]!=0) {sprintf(bad_unusareabyte_dirents, "%s %02d", bad_unusareabyte_dirents, i); break; /* printf("Invalid disk image - invalid unused byte in dir entry #%d\n", i); return 6; */} }
			}
			// ===============================================================================
			/* --------------- save each dir entry (even if empty) in the "devildir0_mem" variable ---------------- */
			memcpy((&devildir0_mem[0] + i*256), &rb[0], 256); // dir_entry_idx++;
			/* --------------- save each dir entry (if not deleted) in the "devildir_mem" variable ---------------- */
			/* --------------- set the 1st byte in front as index of catalogue records holding the ---------------- */
			/* --------------- order number of record in devildir0_mem (0 ... 76) --------------------------------- */
			// The 18th (offset 17, 11h), 19th (offset 18, 12h) and 20th (offset 19, 13h) bytes of dir entry are always 00
			// We shift the first 18 bytes to the right by 1 byte, overwriting the 19th which is always 00 anyway
			if ((rb[0]<=3) || (rb[0]==0xE4)) {
				devildir_mem[0+nr_recs*256]=i;					// record index
				memcpy((&devildir_mem[0] + nr_recs*256 + 1), &rb[0], 18);	// first 18 bytes of dir record (rb[18] is NOT copied !)
				memcpy((&devildir_mem[0] + nr_recs*256 + 19), &rb[19], 237);	// last 237 bytes of dir record (of which the 1st is always 00)
				nr_recs++;
				alloc_dir[i]=1;		/* update directory record allocation map */
				free_dir_recs--;	/* update the nr. of allocated records */
				if (debug) {printf (" %02d ---> %02X %2c %2c %2c %2c %2c %2c %2c %2c %2c %2c %02X %02X %02X %02X %02X %02X %02X\n", i, rb[0],rb[1],rb[2],rb[3],rb[4],rb[5],rb[6],rb[7],rb[8],rb[9],rb[10],rb[11],rb[12],rb[13],rb[14],rb[15],rb[16],rb[17]);}
			} else {if (debug) {printf (" %02d ---> %s\n", i, record);}}
			/* --------------- save each relevant dir entry (if not deleted) in the "devildir1_mem" variable ------ */
			if (rb[0]<=3) {
				devildir1_mem[0+nr_files*256]=i;				// record index
				memcpy((&devildir1_mem[0] + nr_files*256 + 1), &rb[0], 18);	// first 18 bytes of dir record (rb[18] is NOT copied !)
				memcpy((&devildir1_mem[0] + nr_files*256 + 19), &rb[19], 237);	// last 237 bytes of dir record (of which the 1st is always 00)
				nr_files++;
			}
			free(record);
		}
		if (debug) {
			printf("\n===============================================================\n\n\n");
			printf("--------------------------------------\n- DIRECTORY RECORD ALLOCATION TABLE --\n--------------------------------------\n");
			for (i=0; i<7; i++) {printf("trk. %02d - %d  %d  %d  %d  %d  %d  %d  %d  %d  %d\n", i*10, alloc_dir[i*10], alloc_dir[i*10+1], alloc_dir[i*10+2], alloc_dir[i*10+3], alloc_dir[i*10+4], alloc_dir[i*10+5], alloc_dir[i*10+6], alloc_dir[i*10+7], alloc_dir[i*10+8], alloc_dir[i*10+9]);}
			printf("trk. %02d - %d  %d  %d  %d  %d  %d  %d\n", 7*10, alloc_dir[7*10], alloc_dir[7*10+1], alloc_dir[7*10+2], alloc_dir[7*10+3], alloc_dir[7*10+4], alloc_dir[7*10+5], alloc_dir[7*10+6]);
			printf("--------------------------------------\n");
		}
		printf("\n");

		/* ################################## CLOSE DISK IMAGE FILE ################################## */
		close(diskimg_file_descrptr);
		/* ########################################################################################### */

		if (strlen(bad_zeroentry_dirents)) {printf("Invalid disk image - all-zero bytes in dir entries:%s\n", bad_zeroentry_dirents); return 2;}
		if (strlen(bad_datatypebyte_dirents)) {printf("Invalid disk image - invalid data type byte in dir entries:%s\n", bad_datatypebyte_dirents); return 4;}
		if (strlen(bad_attrbyte_dirents)) {printf("Invalid disk image - invalid attribute byte in dir entries:%s\n", bad_attrbyte_dirents); return 5;}
		if (strlen(bad_unusareabyte_dirents)) {printf("Invalid disk image - invalid unused byte in dir entries:%s\n", bad_unusareabyte_dirents); return 6;}

		if (list_mode) {
			printf("================================================================\n");
			printf(" CONTENTS OF: %s\n", basename(diskimg_file_name));
			printf("================================================================\n");
			printf(" ##    Type    Name       Length  Start VarStart VarN Attributes\n");
			printf("================================================================\n");
		}
		if (nr_recs == 0) {
			if (list_mode) {
				printf("|                                                              |\n");
				printf("|   This disk is empty (just formatted or all files deleted)   |\n");
				printf("|                                                              |\n");
				printf("================================================================\n\n");
			}
			if (extract_mode) {return 3;}
		} else {
			for (i=0; i<nr_files; i++) {
				switch (devildir1_mem[i*256+1]) {
					case 0 : strcpy(type,"Prog"); if (devildir1_mem[i*256+1+14]<0x80) {sprintf(par1, "%5d", devildir1_mem[i*256+1+13]+devildir1_mem[i*256+1+14]*256);} else {strcpy(par1, "     ");}  sprintf(par2, "%5d", devildir1_mem[i*256+1+15]+devildir1_mem[i*256+1+16]*256); varN = ' ';
					break;
					case 1 : strcpy(type,"NArr"); strcpy(par1, "     "); strcpy(par2, "     "); varN = devildir1_mem[i*256+1+14]-0x80+0x60;
					break;
					case 2 : strcpy(type,"CArr"); strcpy(par1, "     "); strcpy(par2, "     "); varN = devildir1_mem[i*256+1+14]-0xC0+0x60;
					break;
					case 3 : strcpy(type,"Code"); sprintf(par1, "%5d", devildir1_mem[i*256+1+13]+devildir1_mem[i*256+1+14]*256); strcpy(par2, "     "); varN = ' ';
				}
				switch (devildir1_mem[i*256+18]) {
					case 0 : strcpy(sys,"R/W DIR"); break;
					case 1 : strcpy(sys,"R/W SYS"); break;
					case 2 : strcpy(sys,"R/O DIR"); break;
					case 3 : strcpy(sys,"R/O SYS"); break;
					default: strcpy(sys,"??? ???");
				}
				total_size += devildir1_mem[i*256+1+11]+devildir1_mem[i*256+1+12]*256;
				sprintf(leng, "%5d", devildir1_mem[i*256+1+11]+devildir1_mem[i*256+1+12]*256);
				if (list_mode)   printf(" %2d    %s    %c%c%c%c%c%c%c%c%c%c  %s  %s  %s     %c    %s\n", \
					i, type, devildir1_mem[i*256+1+1], devildir1_mem[i*256+1+2], devildir1_mem[i*256+1+3], devildir1_mem[i*256+1+4], \
						 devildir1_mem[i*256+1+5], devildir1_mem[i*256+1+6], devildir1_mem[i*256+1+7], devildir1_mem[i*256+1+8], \
						 devildir1_mem[i*256+1+9], devildir1_mem[i*256+1+10], \
					leng, par1, par2, varN, sys);
				asprintf(&file_list[i+nr_hless], " %2d    %s    %c%c%c%c%c%c%c%c%c%c  %s  %s  %s     %c    %s\n", \
					i, type, devildir1_mem[i*256+1+1], devildir1_mem[i*256+1+2], devildir1_mem[i*256+1+3], devildir1_mem[i*256+1+4], \
						 devildir1_mem[i*256+1+5], devildir1_mem[i*256+1+6], devildir1_mem[i*256+1+7], devildir1_mem[i*256+1+8], \
						 devildir1_mem[i*256+1+9], devildir1_mem[i*256+1+10], \
					leng, par1, par2, varN, sys);
				hless=0;	// next we look for any headerless blocks attached to this catalog entry
//				printf("%02X %02X %02X %02X\n", devildir1_mem[i*256+1+11], devildir1_mem[i*256+1+12], devildir1_mem[i*256+20], devildir1_mem[i*256+21]);		// just checking
				if ((devildir1_mem[i*256+1+11]==devildir1_mem[i*256+20]) && (devildir1_mem[i*256+1+12]==devildir1_mem[i*256+21])) {
					while (hless <= 116) {
						if ((devildir1_mem[i*256+22+hless*2]) || (devildir1_mem[i*256+22+hless*2+1])) {
							total_size += devildir1_mem[i*256+22+hless*2] + 256*devildir1_mem[i*256+22+hless*2+1];
							if (list_mode) printf(" %2d    ----                %5d  -- headerless data -- %s\n", i, devildir1_mem[i*256+22+hless*2] + 256*devildir1_mem[i*256+22+hless*2+1], sys);
							nr_hless++;
							asprintf(&file_list[i+nr_hless], " %2d    ----                %5d  -- headerless data -- %s\n", i, devildir1_mem[i*256+22+hless*2] + 256*devildir1_mem[i*256+22+hless*2+1], sys);
							hless++;
						} else {break;}
					}
				}	// at this point hless = number of headerless blocks found attached to this headed block
				devildir1_mem[i*256+19]=hless;	// store hless in unused byte 19 of this catalog entry in devildir1_mem[]
			}		// at this point nr_hless = total number of headerless blocks found in catalog
			if (list_mode) {
				printf("================================================================\n");
				printf("|  Directory entries used:   %2d   |   Entries available:   %2d  |\n", 77-free_dir_recs, free_dir_recs);
				printf("|  Total bytes in files: %6d   |   Total file count:    %2d  |\n", total_size, nr_files);
				printf("|  Total bytes on disk:  %6d   |   Disk Free Space: %3d KB  |\n", nr_recs*9*1024, free_dir_recs*9);
				printf("================================================================\n");
				printf("|  For Prog: \"Start\" = Autostart Line Number                   |\n");
				printf("|            \"VarStart\" = Var Area offset from start           |\n");
				printf("|  For Code: \"Start\" = Code Block Start (Load) Address         |\n");
				printf("|  For XArr: \"VarN\"  = Name of the char/numeric array var      |\n");
				printf("================================================================\n\n");
			}
		}
		return 0;
	}

	/* Defragment the free space on disk. */
	void defrag_disk() {
		int diskimg_file_descrptr;
		unsigned char prev_alloc=1;
		unsigned char range1=0, range2=0;		// limits of unallocated range found
		unsigned char q=0;				// will store source block number in a shorter form
//		unsigned char k=0;
		bool expanded=false;				// flag activated when second hole encountered
		unsigned char dirent_buffer[256], track_buffer[track_length];	// read/write buffers for 1 sector and 1 track
		unsigned int dir_sect_start=0;			// start of a sector to read/write in the disk directory

		printf("Starting free space defragmentation on %s ...\n\n", diskimg_file_name);
		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("defrag_disk() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return;}

		printf("Disk space allocation map (1 = allocated block, 0 = free block):\n");
		printf("\n0---------1---------2---------3---------4---------5---------6---------7------\n");
		for (unsigned char i=0; i<77; i++) printf("%d", DEVIL::alloc_dir[i]?1:0); printf("\n");

		for (unsigned char i=0; i<77; i++) {
			if (prev_alloc && alloc_dir[i]) continue;
			if (!prev_alloc && !alloc_dir[i]) continue;
			if ((prev_alloc && !alloc_dir[i]) || (i==0 && !alloc_dir[i])) {prev_alloc=0; range1=i; continue;}	// start of unallocated range found
			if (!prev_alloc && alloc_dir[i]) {prev_alloc=1; range2=i-1;						// end of unallocated range found - time to act
				if (!expanded) printf("\nFound hole from %d to %d, caving in...\n", range1, range2);
				for (unsigned char j=range1; j<=76-(range2+1-range1); j++) {
					q = j+(range2+1-range1);							// source block number in a shorter form
					if (verbose) printf("relocating dir entry %d to %d... ", q, j);
					/* -------------------- relocate dir entry here --------------------- */
					dir_sect_start = (q/18)*sstracklength + xlt[q%18]*256;				// start of source sector with dir entry
					lseek (diskimg_file_descrptr, dir_sect_start, SEEK_SET);			// set position on it
					read (diskimg_file_descrptr, (void *) dirent_buffer, 256);			// read data from it
					dir_sect_start = (j/18)*sstracklength + xlt[j%18]*256;				// start of destination sector for dir entry
					lseek (diskimg_file_descrptr, dir_sect_start, SEEK_SET);			// set position on it
					write (diskimg_file_descrptr, (void *) dirent_buffer, 256);			// write dir entry data to new location
					for (int k=0; k<256; k++) dirent_buffer[k]=0xE5;				// load dir entry buffer with format bytes
					dir_sect_start = (q/18)*sstracklength + xlt[q%18]*256;				// start of source sector with dir entry
					lseek (diskimg_file_descrptr, dir_sect_start, SEEK_SET);			// set position on it ...
					write (diskimg_file_descrptr, (void *) dirent_buffer, 256);			// ... and overwrite it with format bytes
					/* ------------------------------------------------------------------ */
					if (verbose) printf("done.\n");
					if (verbose) printf("--> relocating track %d to %d... ", q, j);
					/* -------------------- relocate track here ------------------------- */
					lseek (diskimg_file_descrptr, track_length*(3+q), SEEK_SET);
					read (diskimg_file_descrptr, (void *) track_buffer, track_length);		// read track data from original location
					lseek (diskimg_file_descrptr, track_length*(3+j), SEEK_SET);
					write (diskimg_file_descrptr, (void *) track_buffer, track_length);		// write track data to new location
					for (int k=0; k<track_length; k++) track_buffer[k]=0xE5;			// load track buffer with format bytes
					lseek (diskimg_file_descrptr, track_length*(3+q), SEEK_SET);			// go back to original location ...
					write (diskimg_file_descrptr, (void *) track_buffer, track_length);		// ... and overwrite it with format bytes
					/* ------------------------------------------------------------------ */
					if (verbose) printf("done.\n");

					alloc_dir[j]=1; alloc_dir[q]=0;	// change alloc_dir[] to reflect the relocation

					range1++; range2++; i++;
					if (range2 == 76) {printf("\nDefragmentation completed.\n\n"); i=76; break;}

//					k=1;
					if (!alloc_dir[i]) {		// if we reached into another hole, extend the range
						printf("found another hole at %d", i);
						while (i <= 76) {if (!alloc_dir[i]) {range2++; i++;} else break;}
						printf (", expanding to %d\n", range2); expanded=true;
						if (range2 == 76) {printf("\nDefragmentation completed.\n\n"); i=76; break;}
					}
				}
			}
		}


		close (diskimg_file_descrptr);

		if (!range1 && !range2) {printf("\nThis disk is empty, nothing to defragment.\n\n");} else
		if (range1 && !range2) {printf("\nFree space on this disk is not fragmented, nothing to defragment.\n\n");} else {
			printf("Disk space allocation map (1 = allocated block, 0 = free block):\n");
			printf("\n0---------1---------2---------3---------4---------5---------6---------7------\n");
			for (unsigned char i=0; i<77; i++) printf("%d", DEVIL::alloc_dir[i]?1:0); printf("\n");
		}
	}


	/* Reads the first "data_size" bytes from one block (a full track, both sides of it) from the disk image and saves it in a .TAP file.
	 * This function will open both files, the .TAP file will be already created because the first data header
	 * will have been saved to it by devilfilecopy() before the actual block data saved by this function.
	 * Global var "checksum" is updated as data is being read */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error opening .TAP file for writing
	 */
	int readblock(unsigned char block_no, unsigned short data_size) {	/* data_size must be <= 9216, make sure when calling this function */
		int diskimg_file_descrptr, tap_file_descrptr;
		unsigned short count;
		unsigned char read_buffer[256];
		int i, j;
		if (debug) {printf("readblock(): reading %d bytes from block/track %d:\n", data_size, block_no);}
//		if (log_on && (tap_log_file_stream != NULL)) {fprintf(tap_log_file_stream, "reading %d bytes from block/track %d:\n", data_size, block_no);}
		diskimg_file_descrptr = open (diskimg_file_name, O_RDONLY); if (diskimg_file_descrptr == -1) {printf("readblock() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}
//		umask(000);
		tap_file_descrptr = open (tap_file_name, O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("readblock() error: tap_file_descrptr ERROR %d: %m\n", errno); return 2;}
		for (i=0; i<data_size/256; i++) {
			count = 256;
			lseek (diskimg_file_descrptr, track_length*(3+block_no)+xlt2[i]*256, SEEK_SET);
			read (diskimg_file_descrptr, (void *) read_buffer, count);
			for (j=0; j<count; j++) {checksum ^= read_buffer[j];}		/* update checksum */
			error_code = write (tap_file_descrptr, (void *) read_buffer, count); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
		}

		if (data_size % 256) {
			count = data_size % 256;
			lseek (diskimg_file_descrptr, track_length*(3+block_no)+xlt2[i]*256, SEEK_SET);
			read (diskimg_file_descrptr, (void *) read_buffer, count);
			for (j=0; j<count; j++) {checksum ^= read_buffer[j];}		/* update checksum */
			error_code = write (tap_file_descrptr, (void *) read_buffer, count); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
		}

		close(diskimg_file_descrptr); close(tap_file_descrptr);
		return 0;
	}


	/* Reads "data_size" bytes from one block (block = a full track, both sides of it) from the disk image and saves it in a .TAP file.
	 * The bytes are read starting after "offset" bytes from the beginning of the sector, so make sure offset+data_size <= 9216 before calling this function.
	 * This function will open both files, the .TAP file will be already created because the first data header
	 * will have been saved to it by devilfilecopy() before the actual block data saved by this function.
	 * Global var "checksum" is updated as data is being read */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error opening .TAP file for writing
	 */
	int readblock1(unsigned char block_no, unsigned short data_size, unsigned short offset) {	/* offset+data_size must be <= 9216, make sure when calling this function */
		int diskimg_file_descrptr, tap_file_descrptr;
		unsigned short count, data_size1=0;
		unsigned char read_buffer[256];
		int i, j;
		if (offset + data_size > 9216) {printf("readblock1() error: data to read exceeds end of track\n"); return 3;}
		if (debug) {printf("readblock(): reading %d bytes from block/track %d starting at block address %d:\n", data_size, block_no, offset);}
//		if (log_on && (tap_log_file_stream != NULL)) {fprintf(tap_log_file_stream, "reading %d bytes from block/track %d:\n", data_size, block_no);}
		diskimg_file_descrptr = open (diskimg_file_name, O_RDONLY); if (diskimg_file_descrptr == -1) {printf("readblock1() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}
//		umask(000);
		tap_file_descrptr = open (tap_file_name, O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("readblock1() error: tap_file_descrptr ERROR %d: %m\n", errno); return 2;}

		if (offset) {
			count = 256 - (offset % 256);
			lseek (diskimg_file_descrptr, track_length*(3+block_no)+xlt2[offset/256]*256+offset%256, SEEK_SET);
			read (diskimg_file_descrptr, (void *) read_buffer, count);
			for (j=0; j<count; j++) {checksum ^= read_buffer[j];}		/* update checksum */
			error_code = write (tap_file_descrptr, (void *) read_buffer, count); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
			data_size1 = data_size - count;
		} else {data_size1 = data_size;}

		for (i=(offset ? offset/256+1 : 0); i<(data_size+offset)/256; i++) {
			count = 256;
			lseek (diskimg_file_descrptr, track_length*(3+block_no)+xlt2[i]*256, SEEK_SET);
			read (diskimg_file_descrptr, (void *) read_buffer, count);
			for (j=0; j<count; j++) {checksum ^= read_buffer[j];}		/* update checksum */
			error_code = write (tap_file_descrptr, (void *) read_buffer, count); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
		}

		if (data_size1 % 256) {
			count = data_size1 % 256;
			lseek (diskimg_file_descrptr, track_length*(3+block_no)+xlt2[i]*256, SEEK_SET);
			read (diskimg_file_descrptr, (void *) read_buffer, count);
			for (j=0; j<count; j++) {checksum ^= read_buffer[j];}		/* update checksum */
			error_code = write (tap_file_descrptr, (void *) read_buffer, count); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
		}
//		printf ("readblock1() wrote %d bytes\n", 256-(offset%256) + (data_size/256-1)*256 + data_size1%256);	// test point
		close(diskimg_file_descrptr); close(tap_file_descrptr);
		return 0;
	}


	/* Writes "data_size" bytes to the block (a full track, both sides of it) specified by "block_no" in the disk image.
	 * Data is written starting at the beginning of the block (track), so make sure data_size <= 9216 before calling this function.
	 * Data to be written must already exist in "track_buffer".
	 * devilfilewrite() will have copied data from tap_file_buffer[] to track_buffer[] before calling this function.
	 * This function will open diskimg_file_name and close it when done. */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error opening .TAP file for writing
	 */
	int writeblock(unsigned char block_no, unsigned short data_size) {
		int diskimg_file_descrptr;
		unsigned short count;
		unsigned char write_buffer[256];
		int i;
		if (debug) {printf("writeblock(): writing %d bytes to track %d\n", data_size, block_no+3);}
		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("writeblock() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}
		for (i=0; i<data_size/256; i++) {
			count = 256;
			memcpy(&write_buffer[0], &track_buffer[i*256], count);	/* the first 3 bytes in tap_file_buffer are protocol bytes, not part of actual data */
			/* -------------------------------------------------------------------------------------------------- */
			error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, count, 9216*(3+block_no)+xlt2[i]*256); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
			fsync (diskimg_file_descrptr);
			/* -------------------------------------------------------------------------------------------------- */
		}
		if (data_size % 256) {
			count = data_size % 256;
			memcpy(&write_buffer[0], &track_buffer[i*256], count);	/* the first 3 bytes in tap_file_buffer are protocol bytes, not part of actual data */
			/* -------------------------------------------------------------------------------------------------- */
			error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, count, 9216*(3+block_no)+xlt2[i]*256); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
			fsync (diskimg_file_descrptr);
			/* -------------------------------------------------------------------------------------------------- */
		}
		close(diskimg_file_descrptr);
		return 0;
	}


	/* Writes "data_size" bytes to the block (a full track, both sides of it) specified by "block_no" in the disk image.
	 * Data is written starting at beginning of sector "sector" (sector=1...36) of the block (track), so make sure (sector-1)*256+data_size<=9216 before calling this function.
	 * Data to be written must already exist in "track_buffer".
	 * devilfilewrite() will copy data from tap_file_buffer[] to track_buffer[] before calling this function.
	 * This function will open diskimg_file_name and close it when done. */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error opening .TAP file for writing
	 */
	int writeblock1(unsigned char block_no, unsigned short data_size, unsigned char sector) {
		int diskimg_file_descrptr;
		unsigned short count;
		unsigned char write_buffer[256];
		int i;
		if ((sector-1)*256+data_size > 9216) {printf("writeblock1() error: data to write exceeds end of track\n"); return 3;}
		if (debug) {printf("writeblock1(): writing %d bytes to track %d starting at sector %d:\n", data_size, block_no+3, sector);}
		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("writeblock1() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}
		for (i=sector-1; i<sector-1+data_size/256; i++) {
			count = 256;
			memcpy(&write_buffer[0], &track_buffer[i*256], count);
			/* -------------------------------------------------------------------------------------------------- */
			error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, count, 9216*(3+block_no)+xlt2[i]*256); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
			fsync (diskimg_file_descrptr);
			/* -------------------------------------------------------------------------------------------------- */
		}
		if (data_size % 256) {
			count = data_size % 256;
			memcpy(&write_buffer[0], &track_buffer[i*256], count);
			/* -------------------------------------------------------------------------------------------------- */
			error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, count, 9216*(3+block_no)+xlt2[i]*256); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
			fsync (diskimg_file_descrptr);
			/* -------------------------------------------------------------------------------------------------- */
		}
		close(diskimg_file_descrptr);
		return 0;
	}


	/* Saves a single catalog entry or a range of entries to a .TAP file according to the .TAP file format.
	 * If tap_name contains an empty string, the name of the .TAP file will be the name of the first catalog entry in the given range.
	 * If tap_name contains a non-empty string, the string will be used as the .TAP file name (for GUI extraction of multiple ranges to the same TAP file).
	 * The .TAP file will be created in this function, when the first header is generated and saved to it.
	 * Then the .TAP file is closed and will later be opened and closed (along with the disk image file)
	 * by the readblock() function which is called in here. */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error creating disk image directory
	 * 2 - Target name already exists but is not a directory
	 * 3 - Existing TAP file could not be deleted
	 * 4 - Existing TAP file could not be deleted, file R/O
	 * 5 - Log file could not be created
	 */
	int devilfilecopy(unsigned char range1, unsigned char range2, const char* tap_name, bool overwrite) {

		unsigned char header[24];		/* save buffer holding header data + first 3 bytes of data block (data block length + flag (FF)) */
		header[0]=0x13; header[1]=0x00;		/* header size is always 19 bytes (0x0013) */
		header[2]=0x00;				/* flag byte = 00 for headers */
		unsigned short leftover_size, prev_leftover_size, data_length;
		char name[11], name1[11];		/* standard 10 char Spectrum program name */
		unsigned char params[6], params1[6];
		unsigned char dir_index;
		unsigned char hless_blocks;		/* number of headerless blocks found attached to current catalog entry */
		unsigned char hless_ix;			/* headerless block counter */
		unsigned short hless_length;		/* length of current headerless block (as read from disk) */
		unsigned short hless_length_tap;	/* length of current headerless block (to be stored in .TAP:   hless_length_tap=hless_length+2    because we add the flag & checksum) */
		unsigned char hless_start_track;	/* track (allocation unit) containing the start of current headerless block */
		unsigned short hless_start;		/* start of headerless data within current track */
		unsigned char hless_end_track;		/* track (allocation unit) containing the end of current headerless block */
		unsigned short hless_end;		/* end of headerless data within whatever track where it ends */

		/* ############################# CHECK FOR EXISTING DISK IMAGE DIR ###################################### */
//		diskimg_dir ();		/* set disk image base saving dir - in dest_dir_base */		// moved to main() and MainWindow::extractFiles()

		umask(000);
		switch (xis_dir(dest_dir_base)) {	/* check if disk image base directory exists */
							/* if not, create it: */
			case  1 : printf("Using already existing base directory %s\n", dest_dir_base);
			
			break;
			
			case -1 : if (mkdir (dest_dir_base, 0777) != 0) {printf ("Error %d creating disk image directory %s: %m\n", errno, dest_dir_base); return 1;}
				  else {if (verbose) {printf ("Disk image directory '%s' successfully created\n", dest_dir_base);} }
			break;
							/* if some file with same name exists, abort (should never happen but anyway) */
			case -2 : printf ("Error: '%s' already exists but is not a directory\n", dest_dir_base); return 2;
		}

		j = range1;
		/* ################### DETERMINE NAME OF .TAP FILE TO SAVE TO AND NAME OF ASSOCIATED LOG FILE #################### */
		if (!strlen(tap_name)) {			// if arg tap_name is an empty string
			dir_index = devildir1_mem[j*256];
			sprintf(name, "%c%c%c%c%c%c%c%c%c%c", \
							devildir0_mem[dir_index*256+1], devildir0_mem[dir_index*256+2], devildir0_mem[dir_index*256+3], devildir0_mem[dir_index*256+4], \
							devildir0_mem[dir_index*256+5], devildir0_mem[dir_index*256+6], devildir0_mem[dir_index*256+7], devildir0_mem[dir_index*256+8], \
							devildir0_mem[dir_index*256+9], devildir0_mem[dir_index*256+10]);
			for (short j=0; j<10; j++) {if (name[j]=='/') {name[j]='_';}}			/* replace '/' with '_' in the .TAP file name being saved to avoid Linux filesystem errors */
			for (short j=9; j>=0; j--) {if (name[j]==' ') {name[j]='\0';} else break;}	/* delete trailing spaces from file name */
		} else {
			snprintf(name, 11, tap_name);
		}

		// ========================== set full path for .TAP file in tap_file_name ==========================================
		if (tap_file_name_allocated) {free(tap_file_name);} tap_file_name_allocated=true;
		if (strlen(tap_name) <= 10) {	// we are running in command line mode
			tap_file_name = (char*)malloc((strlen(dest_dir_base)+strlen(name)+6)*sizeof(char));
			strcpy(tap_file_name, dest_dir_base); strcat(tap_file_name, "/");
			strcat(tap_file_name, name); strcat(tap_file_name, ".tap");	// using name[]		// set full path for .TAP file in tap_file_name
		} else {			// we are running in GUI
			tap_file_name = (char*)malloc((strlen(dest_dir_base)+strlen(tap_name)+6)*sizeof(char));
			strcpy(tap_file_name, dest_dir_base); strcat(tap_file_name, "/");
			strcat(tap_file_name, tap_name); strcat(tap_file_name, ".tap");	// using tap_name	// set full path for .TAP file in tap_file_name
		}

		if (tap_log_file_name_allocated) {free(tap_log_file_name);} tap_log_file_name_allocated=true;
		tap_log_file_name = (char*)malloc((strlen(tap_file_name)+5)*sizeof(char));
		strcpy(tap_log_file_name, tap_file_name); strcat(tap_log_file_name, ".log");	// set full path for .TAP log file in tap_log_file_name

		if (!strlen(tap_name) && overwrite) {			// if arg tap_name is an empty string and overwrite=true
			/* check if the .TAP file name already exists, and if so try to delete it. */
			/* if the .TAP file name exists and cannot be deleted, terminate with error */
			/* if the .TAP file name exists and is R/O, terminate with error */
			if (access (tap_file_name, F_OK) != -1 ) {
				if (access (tap_file_name, W_OK) != -1) {
					if (unlink(tap_file_name)==0) { printf ("Existing %s file successfully deleted\n", tap_file_name ); }
					else {printf ("Existing %s file could not be deleted. File error %d: %m\n\n", tap_file_name, errno); return 3;}
				} else {printf ("Existing %s file could not be deleted, file R/O\n", tap_file_name ); return 4;}
			}
		}

		/* if all OK with the .TAP file name, open the associated log file */
		if (log_on) {
			tap_log_file_stream = fopen (tap_log_file_name, "a");
			if (tap_log_file_stream == NULL) {printf ("devilfilecopy() error: log file %s could not be created. file error %d: %m\n\n", tap_log_file_name, errno); return 5;}
		}

		printf("Saving catalog entries from %d to %d into file %s:\n", range1, range2, tap_file_name);
		if (log_on) {fprintf(tap_log_file_stream, "Saving catalog entries from %d to %d into file %s:\n", range1, range2, tap_file_name);}
		/* ############################################################################### */
		while (j <= range2) {		/* read all data blocks that make up the catalog range specified */
			dir_index = devildir1_mem[j*256];
			if (verbose) {printf("----------------------------------------------------------------------\nSaving data block for catalog entry #%d (starting with dir entry %d):\n", j, dir_index);}
			if (log_on) {fprintf(tap_log_file_stream, "----------------------------------------------------------------------\nSaving data block for catalog entry #%d (starting with dir entry %d):\n", j, dir_index);}
			switch (devildir0_mem[dir_index*256]) {
				case 0 : if (verbose) {printf("> Prog: ");} if (log_on) {fprintf(tap_log_file_stream, "> Prog: ");}
				break;
				case 1 : if (verbose) {printf("> NArr: ");} if (log_on) {fprintf(tap_log_file_stream, "> NArr: ");}
				break;
				case 2 : if (verbose) {printf("> CArr: ");} if (log_on) {fprintf(tap_log_file_stream, "> CArr: ");}
				break;
				case 3 : if (verbose) {printf("> Code: ");} if (log_on) {fprintf(tap_log_file_stream, "> Code: ");}
			}
			if (verbose) {
				printf("%c%c%c%c%c%c%c%c%c%c =",devildir0_mem[dir_index*256+1], devildir0_mem[dir_index*256+2], devildir0_mem[dir_index*256+3], devildir0_mem[dir_index*256+4], \
								devildir0_mem[dir_index*256+5], devildir0_mem[dir_index*256+6], devildir0_mem[dir_index*256+7], devildir0_mem[dir_index*256+8], \
								devildir0_mem[dir_index*256+9], devildir0_mem[dir_index*256+10]);
			}
			if (log_on) {
				fprintf(tap_log_file_stream, "%c%c%c%c%c%c%c%c%c%c =",	devildir0_mem[dir_index*256+1], devildir0_mem[dir_index*256+2], devildir0_mem[dir_index*256+3], devildir0_mem[dir_index*256+4], \
											devildir0_mem[dir_index*256+5], devildir0_mem[dir_index*256+6], devildir0_mem[dir_index*256+7], devildir0_mem[dir_index*256+8], \
											devildir0_mem[dir_index*256+9], devildir0_mem[dir_index*256+10]);
			}
			leftover_size = devildir0_mem[dir_index*256+11]+devildir0_mem[dir_index*256+12]*256;
			data_length = leftover_size + 2;	/* to be used in the data packing as size (flag byte (FF) plus checksum byte = 2 extra bytes for data block) */
			if (verbose) {printf(" %5d bytes\n", leftover_size);}
			if (log_on) {fprintf(tap_log_file_stream, " %5d bytes\n", leftover_size);}
			sprintf(name, "%c%c%c%c%c%c%c%c%c%c", \
							devildir0_mem[dir_index*256+1], devildir0_mem[dir_index*256+2], devildir0_mem[dir_index*256+3], devildir0_mem[dir_index*256+4], \
							devildir0_mem[dir_index*256+5], devildir0_mem[dir_index*256+6], devildir0_mem[dir_index*256+7], devildir0_mem[dir_index*256+8], \
							devildir0_mem[dir_index*256+9], devildir0_mem[dir_index*256+10]);
			memcpy(&params[0], &devildir0_mem[dir_index*256+11], 6);		/* params = 6 bytes containing length, par1 and par2 (used as reference to compare the 6 bytes on following dir entries) */
			i=dir_index;	/* i = counter for allocation units (blocks, tracks) in this catalog entry */
			strcpy(name1, name); memcpy(&params1[0], &params[0], 6);
/*			if (debug) {printf ("name = '%s'  name1 = '%s'\n", name, name1);}	*/

			/* ####################### save header data here ######################### */
			/* -------------------- fill out the header buffer ----------------------- */
			memcpy(&header[3], &devildir0_mem[i*256], 17);	/* fill out the data type, name, data length, par1, par2 */
			checksum_h = 0; for (k=2; k<20; k++) {checksum_h ^= header[k];} header[20] = checksum_h;
			header[21] = data_length % 256; header[22] = data_length / 256; header[23] = 0xFF;
			/* -------------------- write out the header buffer ---------------------- */		/* THE .TAP FILE IS CREATED HERE */
			umask(000);			/* ##################### OPEN AND CLOSE .TAP FILE ####################### */
			tap_file_descrptr = open (tap_file_name, O_CREAT | O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("devilfilecopy() error: tap_file_descrptr ERROR %d: %m\n", errno); return 6;}
			error_code = write (tap_file_descrptr, (void *) header, 24); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
			close(tap_file_descrptr);	/* ###################################################################### */
			/* ####################################################################### */

			checksum = 0xFF;	/* initialize read data checksum with the Spectrum data block flag (0xFF) */
			while ((leftover_size) && (strcmp(name, name1)==0) && ((i==dir_index) || ((devildir0_mem[i*256]==0xE4) && (memcmp(&params[0], &params1[0], 6) == 0)))) {	/* read all tracks for this data block */
				if (verbose) {printf ("Saving data from dir entry #%02d (%5d bytes)\n", i, ((leftover_size > 9216) ? 9216 : leftover_size));}
				if (log_on) {fprintf(tap_log_file_stream, "Saving data from dir entry #%02d (%5d bytes)\n", i, ((leftover_size > 9216) ? 9216 : leftover_size));}
				/* ####################### save block data here ########################## */
				/* -------------------- read track and save to tap file ------------------ */
				readblock(i, (leftover_size > 9216) ? 9216 : leftover_size);	/* readblock() also updates global var checksum */
				/* ####################################################################### */
				if (leftover_size > 9216) {leftover_size -= 9216; i++;} else {prev_leftover_size = leftover_size; leftover_size = 0;}
				sprintf(name1, "%c%c%c%c%c%c%c%c%c%c", \
							devildir0_mem[i*256+1], devildir0_mem[i*256+2], devildir0_mem[i*256+3], devildir0_mem[i*256+4], \
							devildir0_mem[i*256+5], devildir0_mem[i*256+6], devildir0_mem[i*256+7], devildir0_mem[i*256+8], \
							devildir0_mem[i*256+9], devildir0_mem[i*256+10]);
				memcpy(&params1[0], &devildir0_mem[i*256+11], 6);
			}	/* at this point i holds the index of last track read. */
			/* ------------------ now write the data checksum byte ---------------------- */
			tap_file_descrptr = open (tap_file_name, O_CREAT | O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("devilfilecopy() error: tap_file_descrptr ERROR %d: %m\n", errno); return 6;}
			error_code = write (tap_file_descrptr, (void *) &checksum, 1); if (error_code == -1) {printf("devilfilecopy() error: WRITE ERROR %d: %m\n", errno);}
			close(tap_file_descrptr);
			/* -------------------------------------------------------------------------- */
			/* while (devildir0_mem[i*256] == 0xE5) {i++;}	// skip over possible deleted tracks (blocks) */

			/* ------------ now check for possible headerless blocks attached to the catalog entry we just saved ------------- */
			hless_blocks = devildir1_mem[j*256+19];	/* read the number of headerless blocks stored in devildir1_mem[] for current catalog entry by devildirdump() */
			if (hless_blocks && verbose) printf("   -----------------\nFound %d headerless blocks attached to current catalog entry:\n   -----------------\n", hless_blocks);
			hless_ix = 0;
			while (hless_ix < hless_blocks) {
				hless_length = devildir1_mem[j*256+22+hless_ix*2] + devildir1_mem[j*256+22+hless_ix*2+1]*256;	/* get current headerless block length (as stored on disk) */
				hless_start_track = (((prev_leftover_size/256)*256 + (prev_leftover_size%256 ? 256 : 0)) >= 9216) ? i+1 : i;
				hless_start = ((prev_leftover_size/256)*256 + (prev_leftover_size%256 ? 256 : 0))%9216;		/* if existing, headerless data starts at beginning of the 256-byte sector after catalog entry data */
				hless_end_track = i + ( (hless_length <= 9216-hless_start) ? 0 : ((hless_length-(9216-hless_start))/9216 + ((hless_length-(9216-hless_start))%9216 ? 1 : 0)) );
				hless_end = (hless_length>9216-hless_start) ? ((hless_length-(9216-hless_start))%9216-1) : (hless_start+hless_length-1);
				if (verbose) {
					printf("Headerless block #%d starts in track %d at offset %d and ends in track %d at offset %d\n", hless_ix+1, hless_start_track+3, hless_start, hless_end_track+3, hless_end);
					printf("Headerless block #%d: %d bytes\n   -----------------\n", hless_ix+1, hless_length);
				}
				hless_length_tap = hless_length + 2;		// data length to be stored in .TAP (2 extra bytes, flag & checksum)
				header[0]=hless_length_tap%256;			// 2-byte data length ...
				header[1]=hless_length_tap/256;			// ... of this headerless block
				header[2]=0xFF;					// flag byte
				/* ------------ now write the first 3 bytes of headerless block ------------- */
				tap_file_descrptr = open (tap_file_name, O_CREAT | O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("devilfilecopy() error: tap_file_descrptr ERROR %d: %m\n", errno); return 6;}
				error_code = write (tap_file_descrptr, (void *) &header[0], 3); if (error_code == -1) {printf("devilfilecopy() error: WRITE ERROR %d: %m\n", errno);}
				close(tap_file_descrptr);
				/* ------------ now write the headerless block data ------------------------- */
				checksum=0xFF;	/* initialize read data checksum with the flag byte value (0xFF) */
				readblock1(hless_start_track, ((hless_start_track == hless_end_track) ? (hless_end - hless_start + 1) : (9216-hless_start)), hless_start);
				if (hless_end_track - hless_start_track > 1) {
					for (unsigned char trkcnt=hless_start_track+1; trkcnt<hless_end_track; trkcnt++) {readblock(trkcnt, 9216);}
					readblock(hless_end_track, hless_end+1);
				}
				if (hless_end_track - hless_start_track == 1) {
					readblock(hless_end_track, hless_end+1);
				}
				/* ------------ now write the headerless block data checksum byte ----------- */
				tap_file_descrptr = open (tap_file_name, O_CREAT | O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("devilfilecopy() error: tap_file_descrptr ERROR %d: %m\n", errno); return 6;}
				error_code = write (tap_file_descrptr, (void *) &checksum, 1); if (error_code == -1) {printf("devilfilecopy() error: WRITE ERROR %d: %m\n", errno);}
				close(tap_file_descrptr);
				/* -------------------------------------------------------------------------- */

				hless_ix++; i=hless_end_track; prev_leftover_size=hless_end+1;
			}
			/* ------------ at this point all headerless blocks should be saved ---------------------------------------------- */

			if (verbose) {printf ("----------------------------------------------------------------------\n");}
			if (log_on) {fprintf(tap_log_file_stream, "----------------------------------------------------------------------\n");}
			/* --------------------------------------------------------------------------------------------------------------- */
			j++;
		}
		if (log_on) {fclose(tap_log_file_stream);}
		printf ("Catalog entries %d to %d successfully saved as %s.\n\n", range1, range2, tap_file_name);

		return 0;
	}


	/* Extract a "data_length" long data block starting at "tap_file_address" in the currently open .TAP file
	 * into a new .BIN binary file named "bin_file_name" */
	int extracttapblock(const char* bin_file_name, unsigned int tap_file_address, unsigned int data_length) {
		int tap_file_descrptr, bin_file_descrptr;
		struct stat st;
		stat(tap_file_name, &st); size = st.st_size;
		unsigned char* tap_buffer = (unsigned char*)malloc(size*sizeof(unsigned char));	// allocate mem for temporary .TAP read/write buffer

		/* ################ OPEN BIN FILE ################# */
		bin_file_descrptr = open (bin_file_name, O_CREAT | O_WRONLY | O_APPEND, 00666); if (bin_file_descrptr == -1) {printf("extracttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return -1;}
		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDONLY); if (tap_file_descrptr == -1) {printf("extracttapblock() error: tap_file_descrptr ERROR %d: %m\n", errno); return -2;}
		/* ################################################ */

		printf("Extracting %d bytes of binary data from '%s' into '%s'\n", data_length, tap_file_name, bin_file_name); fflush(stdout);

		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		read (tap_file_descrptr, (void*)tap_buffer, data_length);
		write(bin_file_descrptr, (void*)tap_buffer, data_length);

		/* ################ CLOSE BIN FILE ################ */
		close(bin_file_descrptr);
		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		free(tap_buffer);							// de-allocate mem for temporary .TAP read/write buffer
		return 0;
	}


	/* Insert a new data block into the currently open .TAP file at a specified point (address)
	 * from a data buffer pointed to by the third parameter passed.
	 * Must make sure the buffer size is at least tap_block_length bytes or else segfault will happen */
	int inserttapblock(unsigned int tap_file_address, unsigned int tap_block_length, unsigned char* buffer) {
		int tap_file_descrptr;
		struct stat st;
		stat(tap_file_name, &st); size = st.st_size;
		unsigned char* tap_buffer = (unsigned char*)malloc(size*sizeof(unsigned char));	// allocate mem for temporary .TAP read/write buffer
												// used for shifting .TAP file data to make room for insertion
		printf("Inserting %d bytes at offset %d in file '%s' of size %lu.\n", tap_block_length, tap_file_address, tap_file_name, size);		// TESTING

		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDWR); if (tap_file_descrptr == -1) {printf("inserttapblock() error: tap_file_descrptr ERROR %d: %m\n", errno); return -2;}
		/* ################################################ */

		if (tap_file_address > size-1) {printf("Data insertion address beyond end of file."); return -3;}

		ftruncate(tap_file_descrptr, size+tap_block_length);			// add tap_block_length 0x00 bytes at end
		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		read (tap_file_descrptr, (void*)tap_buffer, size-tap_file_address);
		lseek(tap_file_descrptr, tap_file_address+tap_block_length, SEEK_SET);
		write(tap_file_descrptr, (void*)tap_buffer, size-tap_file_address);	// move   size+1-tap_file_address   bytes from tap_file_address to end
		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		write(tap_file_descrptr, (void*)buffer, tap_block_length);		// (over)write   tap_block_length   bytes at   tap_file_address from   buffer

		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		free(tap_buffer);							// de-allocate mem for temporary .TAP read/write buffer
		return 0;
	}


	/* Insert data into the currently open .TAP file at a specified point ("tap_file_address")
	 * from another entire .TAP file having the path specified by "bin_file_name".
	 * Data can be inserted somewhere in the middle (tap_file_address < size of .TAP file,
	 * and then original data starting at insertion point must be shifted over towards the end)
	 * or can be simply appended at the end, after the entire original data (tap_file_address = size of .TAP file).
	 * It uses global var DEVIL::tap_file_name */
	int inserttapfile(unsigned int tap_file_address, const char* bin_file_name) {
		int tap_file_descrptr, bin_file_descrptr;
		struct stat st1;	// for .TAP file currently open
		struct stat st2;	// for .BIN file where data is taken from
		stat(tap_file_name, &st1); unsigned int size1 = st1.st_size;		// size1 = .TAP file size
		stat(bin_file_name, &st2); unsigned int size2 = st2.st_size;		// size2 = .BIN file size

		if (tap_file_address > size1) {printf("Data insertion address beyond end of file."); return -1;}

		unsigned char* tap_buffer = (unsigned char*)malloc(size1*sizeof(unsigned char));	// allocate mem for temporary .TAP read/write buffer
													// used for shifting .TAP file data to make room for insertion
		unsigned char* bin_buffer = (unsigned char*)malloc((size2)*sizeof(unsigned char));	// allocate mem for temporary .BIN read/write buffer
													// used for storing .BIN data and writing to .TAP
		printf("Inserting data from '%s' in '%s' at offset %d.\n\n", bin_file_name, tap_file_name, tap_file_address);		// TESTING

		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDWR); if (tap_file_descrptr == -1) {printf("inserttapblock() error: tap_file_descrptr ERROR %d: %m\n", errno); return -3;}
		/* ################ OPEN BIN FILE ################# */
		bin_file_descrptr = open (bin_file_name, O_RDONLY); if (bin_file_descrptr == -1) {printf("inserttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return -4;}
		/* ################################################ */

		/* ################ READ BIN FILE ################# */
		lseek(bin_file_descrptr, 0, SEEK_SET);
		read (bin_file_descrptr, (void*)(bin_buffer), size2);
		/* ################ CLOSE BIN FILE ################ */
		close(bin_file_descrptr);
		/* ################################################ */

		ftruncate(tap_file_descrptr, size1+size2);				// add size2 bytes of 0x00 at end
		if (tap_file_address<size1) {				// shift ending block(s) only if insertion point not at EOF, to make room for new data
			lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
			read (tap_file_descrptr, (void*)tap_buffer, size1-tap_file_address);
			lseek(tap_file_descrptr, tap_file_address+size2, SEEK_SET);
			write(tap_file_descrptr, (void*)tap_buffer, size1-tap_file_address);	// move   size1-tap_file_address   bytes from tap_file_address to end
		}
		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		write(tap_file_descrptr, (void*)bin_buffer, size2);			// (over)write   size2   bytes at   tap_file_address   from BIN buffer

		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		free(tap_buffer);							// de-allocate mem for temporary .TAP read/write buffer
		free(bin_buffer);							// de-allocate mem for temporary .BIN read/write buffer
		return 0;
	}


	/* Insert a new data block into the currently open .TAP file at a specified point ("tap_file_address")
	 * from a .BIN file having the path specified by "bin_file_name", using "flag" as Spectrum data flag.
	 * Data can be inserted somewhere in the middle (tap_file_address < size of .TAP file,
	 * and then original data starting at insertion point must be shifted over towards the end)
	 * or can be simply appended at the end, after the entire original data (tap_file_address = size of .TAP file).
	 * It uses global var DEVIL::tap_file_name */
	int insertbinblock(unsigned int tap_file_address, const char* bin_file_name, unsigned char flag) {
		int tap_file_descrptr, bin_file_descrptr;
		struct stat st1;	// for .TAP file currently open
		struct stat st2;	// for .BIN file where data is taken from
		stat(tap_file_name, &st1); unsigned int size1 = st1.st_size;		// size1 = .TAP file size
		stat(bin_file_name, &st2); unsigned int size2 = st2.st_size;		// size2 = .BIN file size

		if (tap_file_address > size1) {printf("Data insertion address beyond end of file."); return -1;}
		if (size2>0xFFFF) {printf("Cannot insert data blocks bigger than 65535 bytes.\n"); return -2;}

		unsigned char* tap_buffer = (unsigned char*)malloc(size1*sizeof(unsigned char));	// allocate mem for temporary .TAP read/write buffer
													// used for shifting .TAP file data to make room for insertion
		unsigned char* bin_buffer = (unsigned char*)malloc((size2+4)*sizeof(unsigned char));	// allocate mem for temporary .BIN read/write buffer
													// used for storing .BIN data, calculating checksum and writing to .TAP
		printf("Inserting data from '%s' in '%s' at offset %d.\n\n", bin_file_name, tap_file_name, tap_file_address);		// TESTING

		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDWR); if (tap_file_descrptr == -1) {printf("inserttapblock() error: tap_file_descrptr ERROR %d: %m\n", errno); return -3;}
		/* ################ OPEN BIN FILE ################# */
		bin_file_descrptr = open (bin_file_name, O_RDONLY); if (bin_file_descrptr == -1) {printf("inserttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return -4;}
		/* ################################################ */

		/* ################ READ BIN FILE ################# */
		lseek(bin_file_descrptr, 0, SEEK_SET);
		bin_buffer[0]=(size2+2)%256; bin_buffer[1]=(size2+2)/256; bin_buffer[2]=flag;	// .TAP data block is preceded by 2-byte data length plus flag byte
		read (bin_file_descrptr, (void*)(bin_buffer+3), size2);
		bin_buffer[size2+3]=0; for (unsigned int j=2; j<size2+3; j++) {bin_buffer[size2+3] ^= bin_buffer[j];}	// compute data checksum in last BIN buffer byte
		/* ################ CLOSE BIN FILE ################ */
		close(bin_file_descrptr);
		/* ################################################ */

		ftruncate(tap_file_descrptr, size1+size2+4);				// add size2+4 bytes of 0x00 at end
		if (tap_file_address<size1) {				// shift ending block(s) only if insertion point not at EOF, to make room for new data
			lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
			read (tap_file_descrptr, (void*)tap_buffer, size1-tap_file_address);
			lseek(tap_file_descrptr, tap_file_address+size2+4, SEEK_SET);
			write(tap_file_descrptr, (void*)tap_buffer, size1-tap_file_address);	// move   size1-tap_file_address   bytes from tap_file_address to end
		}
		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		write(tap_file_descrptr, (void*)bin_buffer, size2+4);			// (over)write   size2+4   bytes at   tap_file_address   from BIN buffer

		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		free(tap_buffer);							// de-allocate mem for temporary .TAP read/write buffer
		free(bin_buffer);							// de-allocate mem for temporary .BIN read/write buffer
		return 0;
	}


	int tapblockdelete(unsigned int tap_file_address) {
		int tap_file_descrptr;
		unsigned int data_length=0;
		struct stat st;
		stat(tap_file_name, &st); size = st.st_size;
		unsigned char* tap_buffer = (unsigned char*)malloc(size*sizeof(unsigned char));	// allocate mem for temporary .TAP read/write buffer
												// used for shifting .TAP file data to make room for insertion
		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDWR); if (tap_file_descrptr == -1) {printf("extracttapblock() error: tap_file_descrptr ERROR %d: %m\n", errno); return -2;}
		/* ################################################ */

		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		read (tap_file_descrptr, (void*)tap_buffer, 2);		// read the 2-byte data length of block
		data_length = tap_buffer[0] + 256*tap_buffer[1];	// determine length of data block to delete (total length is data_length+2)

		printf("Deleting %d bytes of binary data starting at %d in '%s'\n", data_length+2, tap_file_address, tap_file_name); fflush(stdout);

		if (tap_file_address+2+data_length<size) {		// shift ending block(s) only if insertion point not at EOF, to make room for new data
			lseek(tap_file_descrptr, tap_file_address+2+data_length, SEEK_SET);
			read (tap_file_descrptr, (void*)tap_buffer, size-(tap_file_address+2+data_length));	// read in ending block(s) to be shifted left over deleted area
			lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
			write(tap_file_descrptr, (void*)tap_buffer, size-(tap_file_address+2+data_length));	// write it back over deleted area
		}
		ftruncate(tap_file_descrptr, size-(2+data_length));	// cut 2+data_length bytes from end

		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		free(tap_buffer);							// de-allocate mem for temporary .TAP read/write buffer
		return 0;
	}


	/* Read some data from the .TAP file currently open into a memory buffer.
	 * Must make sure memory area pointed by "tap_buffer" can accomodate "tap_data_length" bytes
	 * before calling this function or else segfault strikes! */
	int readtapdata(unsigned int tap_file_address, unsigned int tap_data_length, unsigned char* tap_buffer) {
		int tap_file_descrptr;
		unsigned int ret_code;

		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDONLY); if (tap_file_descrptr == -1) {printf("extracttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return 2;}
		/* ################################################ */

		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		ret_code = read (tap_file_descrptr, (void*)tap_buffer, tap_data_length);

		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		if (ret_code == tap_data_length) {return 0;} else
		if (ret_code == 0) {printf("End of file reached."); return 1;} else
				   {printf("Error reading .TAP file."); return 2;}

		return 3;
	}


	/* Write some data from a memory buffer into the .TAP file currently open.
	 * Must make sure writing will not go beyond end of file. */
	int writetapdata(unsigned int tap_file_address, unsigned int tap_data_length, unsigned char* tap_buffer) {
		int tap_file_descrptr;
		unsigned int ret_code;

		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_WRONLY); if (tap_file_descrptr == -1) {printf("extracttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return 2;}
		/* ################################################ */

		lseek(tap_file_descrptr, tap_file_address, SEEK_SET);
		ret_code = write(tap_file_descrptr, (void*)tap_buffer, tap_data_length);

		/* ################ CLOSE TAP FILE ################ */
		close(tap_file_descrptr);
		/* ################################################ */

		if (ret_code == tap_data_length) {return 0;}
		return 1;
	}


	int create_tapfile(const char* bin_file_name, const char* tap_file_name, unsigned char flag) {
		int bin_file_descrptr, tap_file_descrptr;
		struct stat st;
		stat(bin_file_name, &st); unsigned int size = st.st_size;		// size = .BIN file size
		unsigned char* bin_buffer = (unsigned char*)malloc((size+4)*sizeof(unsigned char));	// allocate mem for temporary .BIN read/write buffer
													// used for storing .BIN data, calculating checksum and writing to .TAP
		/* ################ OPEN BIN FILE ################# */
		bin_file_descrptr = open (bin_file_name, O_RDONLY); if (bin_file_descrptr == -1) {printf("inserttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return 1;}
		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_CREAT | O_WRONLY | O_APPEND, 00666); if (tap_file_descrptr == -1) {printf("extracttapblock() error: bin_file_descrptr ERROR %d: %m\n", errno); return 2;}
		/* ################################################ */

		lseek(bin_file_descrptr, 0, SEEK_SET);
		bin_buffer[0]=(size+2)%256; bin_buffer[1]=(size+2)/256; bin_buffer[2]=flag;
		read (bin_file_descrptr, (void*)(bin_buffer+3), size);
		bin_buffer[size+3]=0; for (unsigned int j=2; j<size+3; j++) {bin_buffer[size+3] ^= bin_buffer[j];}
		write(tap_file_descrptr, (void*)bin_buffer, size+4);
		return 0;
	}


	/* Reads a .TAP file and checks all packaged data for errors.
	 * If write == 0 and no errors are encountered returns the number of required tracks on disk.
	 * If write == 0 and errors are encountered, returns -1.
	 * If write != 0 writes all data in the .TAP file to disk, block by block (track by track).
	 * We need to make sure enough continuous free space exists on disk BEFORE writing data from the .TAP file to disk.
	 * That's why this function is meant to be called in 2 different ways: first with write=0 to get number of required tracks,
	 * then with write = 1 and block_no = first block available on disk in a big enough continuous free space. */
	/* If this function is called in non-write mode (write=0) and non-tap_list_mode (tap_list_mode=0), meaning is called
	 * from the GUI in MainWindow::openTap() to view a TAP file, a dynamic array of char* is created (tap_list[]) here
	 * and then is destroyed in MainWindow::openTap() after extracting the text list items for the file list in the GUI */
	/* Return codes:
	 * >=0 - Tracks needed (when not in write mode)
	 * -1 - TAP file too short
	 * -2 - Error opening TAP file
	 * -3 - headerless && dataless blocks found
	 * -4 - headerless blocks found
	 * -5 - dataless blocks found
	 * -6 - Error opening floppy image (when in write mode)
	 */
	signed char devilfilewrite(const char* tap_file_name, unsigned char write, unsigned char block_no) {
		struct stat st;
		char* tap_filename;			/* copy of arg "tap_file_name" for use with basename() */
		unsigned int tap_file_address;
		unsigned char tracks_needed;		/* total nr of tracks needed for entire .tap file */
		unsigned char tracks_needed1;		/* nr of tracks needed for one .tap data block (Prog, Code etc.) */
		unsigned char dirent_buffer[256], dirent_buffer1[256], block_nr;
		unsigned int i;
		char name[11], type[5];
		char start[6];				/* 5-digit start addr or start line as string */
		char varstart[6];			/* 5-digit start addr offset of Prog vars as string */
		char varname=' ';			/* array var name */
		unsigned short len=0, /* len1, */ count, par1, par2, data_len, data_len1;
		unsigned int size, block_start;
		unsigned char err;
		char** tmp;
		unsigned int* tmp1;
		bool header=false;			/* flag to indicate if previous header block exists */
		bool bad_name=false;			/* flag to indicate if header name contains non-ASCII chars (>127) */
		bool bad_dlen=false;			/* flag to indicate if 2-byte block data length could not be read */
		bool bad_dblk=false;			/* flag to indicate if block data (end part) is missing */
		bool chks_err=false;			/* flag to indicate if header block has failed checksum */
		bool chks_err1=false;			/* flag to indicate if headed block has failed checksum */
		bool pbx=false;				/* flag to indicate if previous headed block exists (when finding a headerless block with FF flag) */
		unsigned short pbx_dsk_idx;		/* keeps disk order number of last found headed block */
		char pbx_dskcidx[5];			/* keeps disk order number of last found headed block, as string */
		pbx_dskcidx[4]='\0';
		char hless_type[]="writable";		/* whether headerless block is writable (flag=FF and follows a headed block or another writable headerless) or non-writable */
		unsigned char pbx_dirents;		/* dir entries needed by previously found headed block plus all writable headerless blocks found after it */
		char pbx_dirents_c[3];			/* same thing as string */
		unsigned char pbx_last_sects;		/* sectors needed on last (partially allocated) track needed by previously found headed block: if last track fully allocated, pbx_last_sects==0 */
		unsigned char pbx_last_sect;		/* last sector previously allocated on last track previously allocated when parsing headerless blocks attached to a headed block */
		unsigned char pbx_start_block=0;	/* first allocation unit (0...77) (dir entry index) for previously found headed block */
		unsigned char pbx_last_track;		/* last alloc unit (0...77) previously allocated when parsing headerless blocks attached to a headed block */
		unsigned int pbx_hless_count;		/* store current number of headerless blocks attached to a headed block in write mode */
		unsigned char st_trk;			/* start track (3...79) when writing a writable headerless block to disk */
		unsigned char st_sec;			/* start sector (1...36) when writing a writable headerless block to disk */

		headerless=false;	// error flag, true if any non-writable headerless block (with flag != FF) found
		dataless=false;		// error flag, true if any header block found without its associated data block

		block_nr=block_no;	/* block_nr = current track to write to, will be updated live (block 0 = track 3, tracks 0,1,2 are the directory) */

		for (i=0; i<256; i++) {dirent_buffer[i]=0x00;}		/* bytes 21-255 in a dir entry are always zero and 0-20 will be overwritten with header info */
		stat(tap_file_name, &st); size = st.st_size;
		if (size < 5) {printf("%s: Invalid TAP file! File too short!\n", tap_file_name); return -1;}
		/* ################ OPEN TAP FILE ################# */
		tap_file_descrptr = open (tap_file_name, O_RDONLY); if (tap_file_descrptr == -1) {printf("devilfilewrite(): Error %d opening .TAP file: %m\n", errno); return -2;}
		/* ################################################ */
		tap_file_address = 0;
		tracks_needed = 0;
		err = 0; tap_idx = 0; dsk_idx = 0;
		asprintf(&tap_filename, "%s", tap_file_name);
		while (tap_file_address+1 < size) {
			if (!write && !tap_file_address) {
				printf("===========================================================================\n");
				printf(" CONTENTS OF: %s\n", basename(tap_filename)); free(tap_filename);
				printf("===========================================================================\n");
				printf("Form E  T#    D#  Flag  Size  Type     Name     Length  Start VarStart VarN\n");
				printf("===========================================================================\n");
			}
			block_start=tap_file_address;		// store start address of current block to read next
			/* ---------------- GET DATA LENGTH ---------------------- */
			count = 2; if (size<tap_file_address+count) {/*printf("Missing end of .TAP file: cannot read data length.");*/ bad_dlen=true; bad_dblk=true;}	else {bad_dlen=false;
			lseek (tap_file_descrptr, tap_file_address, SEEK_SET);
			read (tap_file_descrptr, (void *) tap_file_buffer, count);
			data_len=tap_file_buffer[0]+tap_file_buffer[1]*256; if (data_len<3) {printf("Bad block length (<3): data_len = %d", data_len); bad_dlen=true; close(tap_file_descrptr); break;}
			tap_file_address += count;
			/* ---------------- GET DATA BLOCK ----------------------- */
			count = data_len; if (size<tap_file_address+count) {/*printf("Missing end of .TAP file: cannot read data block.");*/ bad_dblk=true;}		else {bad_dblk=false;
			lseek (tap_file_descrptr, tap_file_address, SEEK_SET);
			read (tap_file_descrptr, (void *) tap_file_buffer, count);
			tap_file_address += count;
			/* ------------------------------------------------------- */
			checksum=0; for (unsigned int j=0; j<data_len-1U; j++) {checksum ^= tap_file_buffer[j];}
			if (checksum != tap_file_buffer[data_len-1]) {chks_err=true;} else {chks_err=false;}
//			if (size < tap_file_address+1+len) {printf("TAP file incomplete!\n"); err=1; break;}

			if ((data_len != 0x13) || (tap_file_buffer[0]!=0x00)) { header=false; } else { header=true; } } }

			if (!write && !tap_list_mode) {		// if this function is called from GUI in non-write mode
				/* --------- implementing/expanding dynamic array of char pointers ---------- */
				/* --------- to store text items displayed in the GUI file list ------------- */
				/* --------- and dynamic array of int to store block start addresses -------- */
				if (tap_idx==0) {
					tap_list = (char**)malloc(sizeof(char*));			// stores text items displayed in the GUI file list
					tap_index = (unsigned int*)malloc(sizeof(unsigned int));	// stores .TAP file start addresses of individual blocks
				} else {
					tmp = (char**)realloc(tap_list,(tap_idx+1)*sizeof(char*));
					tap_list=tmp;
					tmp1 = (unsigned int*)realloc(tap_index,(tap_idx+1)*sizeof(unsigned int));
					tap_index=tmp1;
				} /* ------------------------------------------------------------------------ */
				tap_index[tap_idx]=block_start;						// store start address for current block
			}

			if (!bad_dlen && !bad_dblk) {			// if 2-byte data length and the block data have been read ok
				if (header) {

					pbx=false;
					for (unsigned char i=2; i<=11; i++) { if (tap_file_buffer[i] > 127) {bad_name=true; break;} else {bad_name=false;} }
					/* ---------------------- GET name, length, par1, par2 --------------------------------------------- */
					memcpy(&name[0], &tap_file_buffer[2], 10); name[10]='\0';		// get name
					len = tap_file_buffer[12] + tap_file_buffer[13]*256; type[4] = '\0';	// get data length (for next data block)
					par1= tap_file_buffer[14] + tap_file_buffer[15]*256;			// get parameter 1
					par2= tap_file_buffer[16] + tap_file_buffer[17]*256;			// get parameter 2
					/* ---------------------- SET start, varstart, varname STRINGS FOR DISPLAY ------------------------- */
					switch (tap_file_buffer[1]) {
						case 0 : strcpy(type, "Prog"); if (par1<0x8000) {sprintf(start,"%5d",par1);} else {strcpy(start,"     ");} sprintf(varstart,"%5d",par2); varname=' '; break;
						case 1 : strcpy(type, "NArr"); strcpy(start,"     "); strcpy(varstart,"     "); varname=tap_file_buffer[15]-0x80+0x60; break;
						case 2 : strcpy(type, "CArr"); strcpy(start,"     "); strcpy(varstart,"     "); varname=tap_file_buffer[15]-0xC0+0x60; break;
						case 3 : strcpy(type, "Code"); sprintf(start,"%5d",par1); strcpy(varstart,"     "); varname=' ';
					}
					memcpy(&dirent_buffer[0], &tap_file_buffer[1], 17);	// copy header info to dir write buffer
					if (!write) {
									      printf("Head %c %04d        %02X  %5d  %s  %s  %5d   %s   %s    %c\n", (chks_err?'!':' '), tap_idx, tap_file_buffer[0], data_len, type, name, len, start, varstart, varname);
						if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
							asprintf(&tap_list[tap_idx], "Head %c %04d        %02X  %5d  %s  %s  %5d   %s   %s    %c\n", (chks_err?'!':' '), tap_idx, tap_file_buffer[0], data_len, type, name, len, start, varstart, varname);
						}			/* ------------------------------------------------------------ */
					}
				} else {
//					headerless=true;
					if ((tap_file_buffer[0]!=0xFF) || (!pbx) || (pbx_hless_count>116)) {	// headerless is non-writable if flag!=FF, or no previous headed block, or if 117 or more consecutive writable headerless blocks already found
						headerless=true; pbx=false; sprintf(pbx_dskcidx, "    ");
						sprintf(hless_type, "non-writ");
					} else {
						sprintf(pbx_dskcidx, "%04d", pbx_dsk_idx); sprintf(hless_type, "writable");
					}
					if (pbx) {
						pbx_last_track = pbx_start_block + 3 + tracks_needed1 - 1;	// last track currently already allocated (before writing this headerless)
						pbx_last_sect = pbx_last_sects;			/* 1...36 */	// last sector allocated on last track currently already allocated (before writing this headerless)
						pbx_last_sects += (data_len-2)/256 + ((data_len-2)%256 ? 1 : 0);	// sectors on last track of last headed block plus all sectors of this writable headerless block
						if (pbx_last_sects>36) {					// if this headerless extends into next track
							 pbx_dirents   += pbx_last_sects/36-1+(pbx_last_sects%36 ? 1 : 0);	// total tracks needed so far for last headed block & writable headerless blocks following
							tracks_needed  += pbx_last_sects/36-1+(pbx_last_sects%36 ? 1 : 0);	// total tracks needed so far for this .TAP file
							tracks_needed1 += pbx_last_sects/36-1+(pbx_last_sects%36 ? 1 : 0);	// total tracks needed so far for last headed block & writable headerless blocks following
							    block_nr   += pbx_last_sects/36-1+(pbx_last_sects%36 ? 1 : 0);	// update current block nr to write for future headed block
						}
						pbx_last_sects = ((pbx_last_sects%36) ? (pbx_last_sects%36) : 36) ;	// pbx_last_sects now updated to show nr of sectors (1...36) on last track after writing this headerless
						sprintf(pbx_dirents_c, "%2d", pbx_dirents);
					}
					if (!write) {
									      printf("Data %c %04d  %s  %02X  %5d  --------- headerless data (%s) -%c%s%c---\n", (chks_err?'!':' '), tap_idx, pbx_dskcidx, tap_file_buffer[0], data_len, hless_type, (pbx && verbose ? ' ' : '-'), (pbx && verbose ? pbx_dirents_c : "--"), (pbx && verbose ? ' ' : '-'));
						if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
							asprintf(&tap_list[tap_idx], "Data %c %04d  %s  %02X  %5d  --------- headerless data (%s) -%c%s%c---\n", (chks_err?'!':' '), tap_idx, pbx_dskcidx, tap_file_buffer[0], data_len, hless_type, (pbx && verbose ? ' ' : '-'), (pbx && verbose ? pbx_dirents_c : "--"), (pbx && verbose ? ' ' : '-'));
						}			/* ------------------------------------------------------------ */
						printf("---------------------------------------------------------------------------\n");
					} else {
						if (pbx && !headerless) {
							st_trk = (pbx_last_sect==36 ? pbx_last_track+1 : pbx_last_track);	// writing start track (3...79)
							st_sec = (pbx_last_sect==35) ? 36 : (pbx_last_sect+1)%36;		// writing start sector (1...36)
							printf("-----------------------------------------------------------------------------\n");
							printf("Writing headerless block:  size = %5d to disk from track %2d ps %2d (ls %2d):\n", data_len-2,
								st_trk,
								xlt1[st_sec-1],
								st_sec );
							printf("-----------------------------------------------------------------------------\n");
						/* -------------- write writable headerless block here ---------------- */
							diskimg_file_descrptr = open (diskimg_file_name, O_RDWR);
							if (diskimg_file_descrptr == -1) {printf("devilfilewrite() error: diskimg_file_descrptr ERROR %d: %m\n", errno);}
							/* ------------------- read first dir entry -------------------- */
							error_code = pread (diskimg_file_descrptr, (void *) dirent_buffer1, 256, track_length*(pbx_start_block/36)+xlt2[pbx_start_block%36]*256);
							if (error_code == -1) {printf("while reading first dir entry READ ERROR %d: %m\n", errno);}
							if (!pbx_hless_count) {	// if this is the first headerless for this headed block, we must copy headed block length from locations 0Bh, 0Ch to locations 14h, 15h of each dir entry
								dirent_buffer1[0x14]=dirent_buffer1[0x0B]; dirent_buffer1[0x15]=dirent_buffer1[0x0C];
							}
							/* ------------------- now store this headerless block length in proper locations of dir entry buffer -------------------- */
							/* ---------------- (headerless blocks attached to a headed block have lengths stored from byte 0x16 on) ----------------- */
							dirent_buffer1[0x16+pbx_hless_count*2]=(data_len-2)%256; dirent_buffer1[0x16+pbx_hless_count*2+1]=(data_len-2)/256;
							/* ------------------- write all dir entries for headed block + headerless blocks attached to it found so far ------------ */
							for (unsigned char pbx_ix=pbx_start_block; pbx_ix<pbx_start_block+pbx_dirents; pbx_ix++) {
								error_code = pwrite (diskimg_file_descrptr, (void *) dirent_buffer1, 256, track_length*(pbx_ix/36)+xlt2[pbx_ix%36]*256);
								if (error_code == -1) {printf("while writing dir entry #%d WRITE ERROR %d: %m\n", pbx_ix, errno);}
								if (pbx_ix==pbx_start_block) {dirent_buffer1[0]=0xE4;}	// next dir entries must have first byte = 0xE4
							}
							/* ------------------- write current headerless block in data area on disk ----------------------------------------------- */
							for (unsigned char pbx_dx=0; pbx_dx<(data_len-2)/256; pbx_dx++) {
								error_code = pwrite (diskimg_file_descrptr, (void *) &tap_file_buffer[1 + pbx_dx*256], 256,			/* actual data in buffer is after first byte (flag byte) */
										     track_length*st_trk + track_length*(((pbx_dx+st_sec-1)*256)/track_length) + xlt2[((pbx_dx+st_sec-1)%36)]*256);
//										     track_length*pbx_last_track + track_length*(((pbx_dx+pbx_last_sect)*256)/track_length) + xlt2[((pbx_dx+pbx_last_sect)%36)]*256);
								if (error_code == -1) {printf("while writing sector #%d of this headerless at %X WRITE ERROR %d: %m\n", pbx_dx, track_length*(pbx_last_track+1) + (pbx_dx*256)/track_length + xlt2[(pbx_dx%36)]*256, errno);}
							}
							if ((data_len-2)%256) {
								unsigned char pbx_dx=(data_len-2)/256;
								error_code = pwrite (diskimg_file_descrptr, (void *) &tap_file_buffer[1 + pbx_dx*256], (data_len-2)%256,	/* actual data in buffer is after first byte (flag byte) */
										     track_length*st_trk + track_length*(((pbx_dx+st_sec-1)*256)/track_length) + xlt2[((pbx_dx+st_sec-1)%36)]*256);
//										     track_length*pbx_last_track + track_length*(((pbx_dx+pbx_last_sect)*256)/track_length) + xlt2[((pbx_dx+pbx_last_sect)%36)]*256);
								if (error_code == -1) {printf("while writing last sector (#%d) of this headerless at %X WRITE ERROR %d: %m\n", pbx_dx, track_length*(pbx_last_track+1) + (pbx_dx*256)/track_length + xlt2[(pbx_dx%36)]*256, errno);}
							}
							/* ----------------------------------------------------------------------------------------------------------------------- */
							fsync (diskimg_file_descrptr);
							close (diskimg_file_descrptr);
						/* -------------------------------------------------------------------- */
						}
					}
					pbx_hless_count++;
				}
			} else {
				if (!bad_dlen /* && bad_dblk */) {	// if block data (end part) missing
					if (!write) {
									      printf("???? * %04d        ??  %5d  ---------- corrupted (missing) data ---------\n", tap_idx, data_len);
						if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
							asprintf(&tap_list[tap_idx], "???? * %04d        ??  %5d  ---------- corrupted (missing) data ---------\n", tap_idx, data_len);
						}			/* ------------------------------------------------------------ */
					}
				} else {				// if 2-byte data length could not be read completely
					if (!write) {
									      printf("???? ? %04d        ??  ?????  ---------- corrupted (missing) data ---------\n", tap_idx);
						if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
							asprintf(&tap_list[tap_idx], "???? ? %04d        ??  ?????  ---------- corrupted (missing) data ---------\n", tap_idx);
						}			/* ------------------------------------------------------------ */
					}
				}
			}

			tap_idx++;	// increment tape block counter (individual header and non-header blocks)

			/* ########################################################################################################## */
			if (!header && !bad_dlen && !bad_dblk) { continue; }		// If not a header, no point trying to read a headed data block after
			if (bad_dlen || bad_dblk) {
				printf("---------------------------------------------------------------------------\n"); break;
			}
			if (size==tap_file_address) {
				printf("===========================================================================\n"); break;
			}
			
			/* ########################################################################################################## */

			block_start=tap_file_address;
			/* ---------------- GET NEXT DATA LENGTH ----------------- */
			count = 2;
			if (size<tap_file_address+count) {					// if file ends before flag byte or 2-byte data length
				/*printf("Missing end of .TAP file: cannot read data length.");*/
				bad_dlen=true; bad_dblk=true;
			} else {
				bad_dlen=false;
				lseek (tap_file_descrptr, tap_file_address, SEEK_SET);
				read (tap_file_descrptr, (void *) tap_file_buffer, 3);		// read next 3 bytes (data length and flag byte)
				data_len1=tap_file_buffer[0]+tap_file_buffer[1]*256;		// data length of this block from flag to checksum byte
				if (data_len1<3) {printf("Bad block length (<3): data_len1 = %d", data_len); bad_dlen=true; close(tap_file_descrptr); break;}
				if ((data_len1 != len+2) || (tap_file_buffer[2] != 0xFF)) {dataless=true; pbx=false; continue;}	// if data length not matching header or flag != FF, restart the while loop
				// from this point on, data_len1 == len+2, meaning  len = length of actual data in this headed block
				tap_file_address += count;
				/* ---------------- GET NEXT DATA BLOCK ------------------ */
				count = data_len1;
				if (size<tap_file_address+count) {/*printf("Missing end of .TAP file: cannot read data block.");*/ bad_dblk=true;}
				else {
					bad_dblk=false;
//					for (int ix=0; ix<0x10000; ix++) {tap_file_buffer[ix]=0;}
					lseek (tap_file_descrptr, tap_file_address, SEEK_SET);
					read (tap_file_descrptr, (void *) tap_file_buffer, count);
					tap_file_address += count;
					/* ------------------------------------------------------- */
					checksum=0; for (unsigned int j=0; j<data_len1-1U; j++) {checksum ^= tap_file_buffer[j];}
					if (checksum != tap_file_buffer[data_len1-1]) {chks_err1=true;} else {chks_err1=false;}
					tracks_needed1 = (len%9216) ? len/9216+1 : len/9216;	// tracks (allocation units) needed by this headed block
					tracks_needed += tracks_needed1;
				}
			}

			if (!write && !tap_list_mode) {		// if this function is called from GUI in non-write mode
				/* ---------------- expanding dynamic array of char pointers ---------------- */
				/* ---------------- and dynamic array of ints ------------------------------- */
					tmp = (char**)realloc(tap_list,(tap_idx+1)*sizeof(char*));
					tap_list=tmp;
					tmp1 = (unsigned int*)realloc(tap_index,(tap_idx+1)*sizeof(unsigned int));
					tap_index=tmp1;
				/* -------------------------------------------------------------------------- */
				tap_index[tap_idx]=block_start;						// store start address for current block
			}

			if (!bad_dlen && !bad_dblk) {			// if 2-byte data length and the block data have been read ok
				if (!write) {
								      printf("Data %c %04d  %04d  FF  %5d", (chks_err1?'!':' '), tap_idx, dsk_idx, data_len1);
					if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
					    if (verbose)
						asprintf(&tap_list[tap_idx], "Data %c %04d  %04d  FF  %5d  ------------------> OK, tracks needed: %2d", (chks_err1?'!':' '), tap_idx, dsk_idx, data_len1, tracks_needed1);
					    else
						asprintf(&tap_list[tap_idx], "Data %c %04d  %04d  FF  %5d", (chks_err1?'!':' '), tap_idx, dsk_idx, data_len1);
					}			/* ------------------------------------------------------------ */
					if (verbose) {printf("  ------------------> OK, tracks needed: %2d\n", tracks_needed1);} else printf("\n");
					printf("---------------------------------------------------------------------------\n");
				}
			} else {
				if (!bad_dlen /* && bad_dblk */) {	// if block data (end part) missing
					if (!write) {
									      printf("Data * %04d  %04d  ??  %5d  ---------- corrupted (missing) data ---------\n", tap_idx, dsk_idx, data_len1);
						if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
							asprintf(&tap_list[tap_idx], "Data * %04d  %04d  ??  %5d  ---------- corrupted (missing) data ---------\n", tap_idx, dsk_idx, data_len1);
						}			/* ------------------------------------------------------------ */
						printf("---------------------------------------------------------------------------\n");
					}
				} else {				// if 2-byte data length could not be read completely
					if (!write) {
									      printf("Data ? %04d  %04d  ??  ?????  ---------- corrupted (missing) data ---------\n", tap_idx, dsk_idx);
						if (!tap_list_mode) {	/* ---------- store TAP entry in dynamic char* array ---------- */
							asprintf(&tap_list[tap_idx], "Data ? %04d  %04d  ??  ?????  ---------- corrupted (missing) data ---------\n", tap_idx, dsk_idx);
						}			/* ------------------------------------------------------------ */
						printf("---------------------------------------------------------------------------\n");
					}
				}
			}

			pbx_dsk_idx=dsk_idx;	// keep disk order number of this headed block (before incrementing it 2 lines below)
			tap_idx++;		// increment tape block counter for future tape block found (individual header and non-header blocks)
			dsk_idx++;		// increment Devil disk saving counter for future headed block found (counts only headed data blocks - headerless blocks are ignored)
			pbx=true;		// headed block is found at this point
			pbx_dirents = tracks_needed1;	// tracks needed for this headed block
			pbx_last_sects = len%9216 ? ((len%9216)/256 + (((len%9216)%256) ? 1 : 0)) : 36;	// sectors occupied (1...36) on last (partially allocated) track: if len is multiple of 9216, pbx_last_sects==36
			pbx_hless_count = 0;	// initialize number of headerless blocks attached to this headed block

			if (bad_dlen || bad_dblk) {	// if error reading the 2-byte data length or the block data, stop here
				printf("---------------------------------------------------------------------------\n"); break;
			}

			if (write) {
				printf("--------------------------------------------------------------\n");
				printf("Writing %s:  %s  size = %5d to disk from track %2d:\n", type, name, len, block_nr+3);
				printf("--------------------------------------------------------------\n");
				/* ################# WRITE THE FIRST DIR ENTRY FOR THIS HEADED DATA BLOCK #################### */
				diskimg_file_descrptr = open (diskimg_file_name, O_RDWR);
				if (diskimg_file_descrptr == -1) {printf("devilfilewrite() error: diskimg_file_descrptr ERROR %d: %m\n", errno); close (tap_file_descrptr); return -6;}

				if (verbose) {printf(" writing dir entry #01 to track %2d sector %2d\n", block_nr/36, xlt2[block_nr%36]+1);}
				pbx_start_block = block_nr;		// store index of first disk allocation unit for this headed block

				error_code = pwrite (diskimg_file_descrptr, (void *) dirent_buffer, 256, track_length*(block_nr/36)+xlt2[block_nr%36]*256);
				if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
				fsync (diskimg_file_descrptr);

				if (tracks_needed1 > 1) {	/* ################# WRITE THE REST OF THE DIR ENTRIES FOR THIS HEADED DATA BLOCK ################### */
					dirent_buffer[0]=0XE4;		/* flag for extra dir entries of the same data block */
					for (unsigned char i=1; i<tracks_needed1; i++) {
						if (verbose) {printf(" writing dir entry #%02d to track %2d sector %2d\n", i+1, (block_nr+i)/36, xlt2[(block_nr+i)%36]+1);}
						error_code = pwrite (diskimg_file_descrptr, (void *) dirent_buffer, 256, track_length*((block_nr+i)/36)+xlt2[(block_nr+i)%36]*256);
						if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}
					}
				}
				for (i=0; i<len/9216; i++) {	/* ################# WRITE ALL FULL TRACKS FOR THIS .TAP DATA BLOCK ################### */
					if (verbose) {printf(" Writing     block #%02d to track %2d\n", i+1, block_nr+3+i);}
					memcpy(&track_buffer[0], &tap_file_buffer[1+i*9216], 9216);	// 1+i*9216 because we don't write the flag byte to disk, so we skip 1 byte
					writeblock(block_nr+i, 9216);
				}				/* at this point i = len/9216 */
				if (len%9216) {			/* ################# WRITE LAST (PARTIAL) TRACK (IF ANY) FOR THIS .TAP DATA BLOCK ################### */
					if (verbose) {printf(" Writing     block #%02d to track %2d\n", i+1, block_nr+3+i);}
					memcpy(&track_buffer[0], &tap_file_buffer[1+i*9216], len%9216);	// 1+i*9216 because we don't write the flag byte to disk, so we skip 1 byte
					writeblock(block_nr+i, len%9216);
				}
				if (verbose) {printf("-----------------------------------------------------------------------\n\n");}
				close (diskimg_file_descrptr);
			}
			block_nr += tracks_needed1;	// update current block nr to write for future headed blocks
		}
		if (!write) printf("\n");
		close (tap_file_descrptr);
		if (bad_dlen) {return -6;}
		if (bad_dblk) {return -7;}
		if (headerless && dataless) {return -3;}
		else if (headerless) {return -4;}
		else if (dataless) {return -5;}
		else {return tracks_needed;}
	}


	/* Renames a catalogue entry stored on disk */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error writing to disk image
	 */
	int devilfilerename(char* new_name, unsigned char idx1) {	// idx1 = index of Devil catalogue entry in devildir1_mem[]
		unsigned char idx=0;		// index used for devildir_mem[]
		unsigned char idx0=0;		// index used for devildir0_mem[]
		unsigned int block_length = devildir1_mem[0+idx1*256+12] + devildir1_mem[0+idx1*256+13]*256;
		unsigned char total_dir_entries;

		printf("----------------------------------------\nRenaming catalog entry %d to %s:\n----------------------------------------\n", idx1, new_name);
		for (idx=0; idx<77; idx++) {if (devildir_mem[idx*256]==devildir1_mem[idx1*256]) break;}		// idx = index of Devil catalogue entry "idx1" in devildir_mem[]
		if (devildir1_mem[0+idx1*256+19]) {	// if headerless blocks present, nr of headerless blocks is stored in byte 19 of each dir entry in devildir1_mem[] (see devildirdump())
			printf("This catalog entry contains headerless blocks!! Recalculating total nr. of dir entries:\n");
			for (idx0=idx+1; idx0<77; idx0++) {	// compare 17 bytes including the attribute byte
				if (memcmp(&devildir_mem[0+idx0*256+2], &devildir_mem[0+idx*256+2], 17)) {total_dir_entries=idx0-idx; break;}
			}
			printf("file size = %d, takes up %d tracks, plus %d headerless blocks, %d dir entries to rename.\n", block_length, block_length/9216+((block_length%9216)?1:0), devildir1_mem[0+idx1*256+19], total_dir_entries);
		} else {
			total_dir_entries = block_length/track_length+((block_length%track_length) ? 1 : 0);	// use this as nr. of entries for this catalog entry if no headerless blocks
			printf("file size = %d, takes up %d tracks, so %d dir entries to rename.\n", block_length, total_dir_entries, total_dir_entries);
		}

		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("devilfilerename() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}

		printf("Renaming dir entry %d\n", devildir1_mem[idx1*256]);
		error_code = pwrite (diskimg_file_descrptr, (void*)new_name, 10, track_length*(devildir1_mem[idx1*256]/36)+xlt2[devildir1_mem[idx1*256]%36]*256+1); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}

		unsigned char ren_ent_count=1;
		if (total_dir_entries>1) {
			for (unsigned char dx=idx+1; (dx<idx+total_dir_entries) && (dx<77); dx++) {
				if (!memcmp(&devildir1_mem[0+idx1*256+2], &devildir_mem[0+dx*256+2], 17) && (devildir_mem[0+dx*256+1]==0xE4)) {
					idx0=devildir_mem[0+dx*256];	// index of current dir entry in devildir0_mem[]
					printf("Renaming dir entry %d\n", idx0);
					error_code = pwrite (diskimg_file_descrptr, (void*)new_name, 10, track_length*(idx0/36)+xlt2[idx0%36]*256+1);
					ren_ent_count++;
				}
			}
		}

		printf("----------------------------------------\n");

		fsync (diskimg_file_descrptr);
		close(diskimg_file_descrptr);
		if (error_code == -1) {return 2;} else {return 0;}
	}


	/* Changes attributes of a catalogue entry stored on disk.
	 * Parameters:
	 * 	attr = array of 1 byte (attribute byte with new attributes)
	 * 	idx = index of first dir entry of Devil catalogue entry */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error writing to disk image
	 */
	int devilfilechattr(unsigned char attr, unsigned char idx1) {	// idx1 = index of Devil catalogue entry in devildir1_mem[]
		unsigned char idx=0;		// index used for devildir_mem[]
		unsigned char idx0=0;		// index used for devildir0_mem[]
		unsigned int block_length = devildir1_mem[0+idx1*256+12] + devildir1_mem[0+idx1*256+13]*256;
		unsigned char total_dir_entries;

		printf("----------------------------------------\nChanging attrs for catalog entry %d:\n----------------------------------------\n", idx1);
		for (idx=0; idx<77; idx++) {if (devildir_mem[idx*256]==devildir1_mem[idx1*256]) break;}		// idx = index of Devil catalogue entry "idx1" in devildir_mem[]
		if (devildir1_mem[0+idx1*256+19]) {	// if headerless blocks present, nr of headerless blocks is stored in byte 19 of each dir entry in devildir1_mem[] (see devildirdump())
			printf("This catalog entry contains headerless blocks!! Recalculating total nr. of dir entries:\n");
			for (idx0=idx+1; idx0<77; idx0++) {	// compare 17 bytes including the attribute byte
				if (memcmp(&devildir_mem[0+idx0*256+2], &devildir_mem[0+idx*256+2], 17)) {total_dir_entries=idx0-idx; break;}
			}
			printf("file size = %d, takes up %d tracks, plus %d headerless blocks, %d dir entries to change.\n", block_length, block_length/9216+((block_length%9216)?1:0), devildir1_mem[0+idx1*256+19], total_dir_entries);
		} else {
			total_dir_entries = block_length/track_length+((block_length%track_length) ? 1 : 0);	// use this as nr. of entries for this catalog entry if no headerless blocks
			printf("file size = %d, takes up %d tracks, so %d dir entries to change.\n", block_length, total_dir_entries, total_dir_entries);
		}

		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("devilfilechattr() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}

		printf("Changing attrs for dir entry %d\n", devildir1_mem[idx1*256]);
		error_code = pwrite (diskimg_file_descrptr, (void*)&attr, 1, track_length*(devildir1_mem[idx1*256]/36)+xlt2[devildir1_mem[idx1*256]%36]*256+17); if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno);}

		unsigned char ren_ent_count=1;
		if (total_dir_entries>1) {
			for (unsigned char dx=idx+1; (dx<idx+total_dir_entries) && (dx<77); dx++) {
				if (!memcmp(&devildir1_mem[0+idx1*256+2], &devildir_mem[0+dx*256+2], 17) && (devildir_mem[0+dx*256+1]==0xE4)) {
					idx0=devildir_mem[0+dx*256];	// index of current dir entry in devildir0_mem[]
					printf("Changing attrs for dir entry %d\n", idx0);
					error_code = pwrite (diskimg_file_descrptr, (void*)&attr, 1, track_length*(idx0/36)+xlt2[idx0%36]*256+17);
					ren_ent_count++;
				}
			}
		}

		printf("----------------------------------------\n");

		fsync (diskimg_file_descrptr);
		close(diskimg_file_descrptr);
		if (error_code == -1) {return 2;} else {return 0;}
	}


	/* Deletes a range of catalogue entries stored on disk */
	/* range1 and range2 are catalog entry indexes (as for devildir1_mem[]), but in the function block parsing is done in devildir_mem[]
	 * in order to make it easy to ignore deleted records */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error writing disk image
	 */
	int devilfiledelete(unsigned char range1, unsigned char range2) {

		unsigned char write_buffer[256];	/* save buffer holding 0xE5 bytes */
		for (i=0; i<256; i++) {write_buffer[i]=0xE5;}
		unsigned short data_length /*, data_length */;
		char name[11], name1[11];		/* standard 10 char Spectrum program name */
		unsigned char params[6], params1[6];
		unsigned char idx /*, dir_index*/;
		int i, j, k;

		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("devilfiledelete() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}

		j = range1;
		printf("Deleting catalog entries from %d to %d:\n", range1, range2);
		/* ############################################################################### */
		while (j <= range2) {		/* read all data blocks that make up the catalog range specified */
//			dir_index = devildir1_mem[j*256];
			for (idx=0; idx<77; idx++) {if (devildir_mem[idx*256]==devildir1_mem[j*256]) break;}		// idx = index of first dir entry of Devil catalogue entry "j" in devildir_mem[]
//			dir_index = devildir_mem[idx*256];
			if (verbose) printf ("-----------------------------------------------------------------------\n");
			if (verbose) printf("Deleting data block for catalog entry #%d (starting with dir entry %d):\n", j, devildir_mem[idx*256]);
			switch (devildir_mem[idx*256+1]) {
				case 0 : printf("> Prog: ");
				break;
				case 1 : printf("> NArr: ");
				break;
				case 2 : printf("> CArr: ");
				break;
				case 3 : printf("> Code: ");
			}
			printf("%c%c%c%c%c%c%c%c%c%c =",devildir_mem[idx*256+2], devildir_mem[idx*256+3], devildir_mem[idx*256+4], devildir_mem[idx*256+5], \
							devildir_mem[idx*256+6], devildir_mem[idx*256+7], devildir_mem[idx*256+8], devildir_mem[idx*256+9], \
							devildir_mem[idx*256+10], devildir_mem[idx*256+11]);
			data_length = devildir_mem[idx*256+12]+devildir_mem[idx*256+13]*256;
//			data_length = leftover_size + 2;	/* to be used in the data packing as size (flag byte (FF) plus checksum byte = 2 extra bytes for data block) */
			printf(" %5d bytes", data_length); if (devildir1_mem[j*256+19]) {printf(" + %d headerless blocks\n", devildir1_mem[j*256+19]);} else {printf("\n");}
			sprintf(name, "%c%c%c%c%c%c%c%c%c%c", \
							devildir_mem[idx*256+2], devildir_mem[idx*256+3], devildir_mem[idx*256+4], devildir_mem[idx*256+5], \
							devildir_mem[idx*256+6], devildir_mem[idx*256+7], devildir_mem[idx*256+8], devildir_mem[idx*256+9], \
							devildir_mem[idx*256+10], devildir_mem[idx*256+11]);
			memcpy(&params[0], &devildir_mem[idx*256+12], 6);		/* params = 6 bytes containing length, par1 and par2 */
			i=idx; strcpy(name1, name); memcpy(&params1[0], &params[0], 6);
//			if (debug) {printf ("name = '%s'  name1 = '%s'\n", name, name1);}

			/* ######### delete each dir entry first and then its data (by filling with E5) ######### */

			while ((strcmp(name, name1)==0) && ((i==idx) || ((devildir_mem[i*256+1]==0xE4) && (memcmp(&params[0], &params1[0], 6) == 0)))) {	/* delete all tracks for this data block */
				/* ####################### delete dir entry here ########################### */
				if (verbose) printf ("Deleting dir entry #%02d\n", devildir_mem[i*256]);
				error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, 256, track_length*(devildir_mem[i*256]/36)+xlt2[devildir_mem[i*256]%36]*256);
				if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno); close(diskimg_file_descrptr); return 2;}
				fsync (diskimg_file_descrptr);
				/* ####################### delete block data here ########################## */
				if (verbose) printf ("Deleting data for dir entry #%02d (%5d bytes)\n", devildir_mem[i*256], ((data_length > 9216) ? 9216 : data_length));
				for (k=0; k<36; k++) {
					error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, 256, track_length*(3+devildir_mem[i*256])+k*256);
					if (error_code == -1) {printf("WRITE ERROR %d: %m\n", errno); close(diskimg_file_descrptr); return 2;}
				}
				/* ####################################################################### */
				i++; if (data_length > 9216) {data_length -= 9216;}
				sprintf(name1, "%c%c%c%c%c%c%c%c%c%c", \
							devildir_mem[i*256+2], devildir_mem[i*256+3], devildir_mem[i*256+4], devildir_mem[i*256+5], \
							devildir_mem[i*256+6], devildir_mem[i*256+7], devildir_mem[i*256+8], devildir_mem[i*256+9], \
							devildir_mem[i*256+10], devildir_mem[i*256+11]);
				memcpy(&params1[0], &devildir_mem[i*256+12], 6);
			}
			if (verbose) printf ("-----------------------------------------------------------------------\n");
			j++;
		}

		close(diskimg_file_descrptr);
	/*	diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("devilfiledelete() error: diskimg_file_descrptr ERROR %d: %m\n", errno); exit(0);} */
		return 0;
	}

	/* Changes one attribute (R/W-R/O or DIR-SYS) at a time for a range of catalogue entries stored on disk */
	/* range1 and range2 are catalog entry indexes (as for devildir1_mem[]), but in the function block parsing is done in devildir_mem[]
	 * in order to make it easy to ignore deleted records */
	/* Return codes:
	 * 0 - Success
	 * 1 - Error opening disk image
	 * 2 - Error writing disk image
	 */
	int devilfilesetattr(unsigned char range1, unsigned char range2, char* sys) {

		unsigned char write_buffer[1];		/* save buffer holding sys byte */
		unsigned short data_length		/* data_length */;
		char name[11], name1[11];		/* standard 10 char Spectrum program name */
		unsigned char params[6], params1[6];
		unsigned char idx, /*dir_index,*/ attr_no=0;
		char attr[4];
		if (strcmp(sys,"-r0")==0) {strcpy(attr,"R/W"); attr_no = 0;}
		if (strcmp(sys,"-r1")==0) {strcpy(attr,"R/O"); attr_no = 1;}
		if (strcmp(sys,"-s0")==0) {strcpy(attr,"DIR"); attr_no = 2;}
		if (strcmp(sys,"-s1")==0) {strcpy(attr,"SYS"); attr_no = 3;}
		int i, j;

		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("devilfilesetattr() error: diskimg_file_descrptr ERROR %d: %m\n", errno); return 1;}

		j = range1;
		printf("Setting %s attribute for catalog entries from %d to %d:\n", attr, range1, range2);
		/* ############################################################################### */
		while (j <= range2) {		/* read all data blocks that make up the catalog range specified */
//			dir_index = devildir1_mem[j*256];
			for (idx=0; idx<77; idx++) {if (devildir_mem[idx*256]==devildir1_mem[j*256]) break;}		// idx = index of first dir entry of Devil catalogue entry "j" in devildir_mem[]
//			dir_index = devildir_mem[idx*256];
			if (verbose) {printf("---------------------------------------\n");}
			if (verbose) {printf("Setting %s attribute for catalog entry #%d (starting with dir entry %d):\n", attr, j, devildir_mem[idx*256]);}
			switch (devildir_mem[idx*256+1]) {
				case 0 : printf("> Prog: ");
				break;
				case 1 : printf("> NArr: ");
				break;
				case 2 : printf("> CArr: ");
				break;
				case 3 : printf("> Code: ");
			}
			printf("%c%c%c%c%c%c%c%c%c%c =",devildir_mem[idx*256+2], devildir_mem[idx*256+3], devildir_mem[idx*256+4], devildir_mem[idx*256+5], \
							devildir_mem[idx*256+6], devildir_mem[idx*256+7], devildir_mem[idx*256+8], devildir_mem[idx*256+9], \
							devildir_mem[idx*256+10], devildir_mem[idx*256+11]);
			data_length = devildir_mem[idx*256+12]+devildir_mem[idx*256+13]*256;
			printf(" %5d bytes", data_length); if (devildir1_mem[j*256+19]) {printf(" + %d headerless blocks\n", devildir1_mem[j*256+19]);} else {printf("\n");}
			sprintf(name, "%c%c%c%c%c%c%c%c%c%c", \
							devildir_mem[idx*256+2], devildir_mem[idx*256+3], devildir_mem[idx*256+4], devildir_mem[idx*256+5], \
							devildir_mem[idx*256+6], devildir_mem[idx*256+7], devildir_mem[idx*256+8], devildir_mem[idx*256+9], \
							devildir_mem[idx*256+10], devildir_mem[idx*256+11]);
			memcpy(&params[0], &devildir_mem[idx*256+12], 6);		/* params = 6 bytes containing length, par1 and par2 */
			i=idx; strcpy(name1, name); memcpy(&params1[0], &params[0], 6);
//			if (debug) {printf ("name = '%s'  name1 = '%s'\n", name, name1);}

			/* ######### set attr for each dir entry ######### */

			while ((strcmp(name, name1)==0) && ((i==idx) || ((devildir_mem[i*256+1]==0xE4) && (memcmp(&params[0], &params1[0], 6) == 0)))) {	/* set attr for all entries of this data block */
				write_buffer[0] = devildir_mem[i*256+18];			/* get attribute byte from devildir_mem */
				switch (attr_no) {
					case 0 : write_buffer[0] &= ~(1 << 1); break;		/* reset bit 1 for R/W */
					case 1 : write_buffer[0] |=  (1 << 1); break;		/*   set bit 1 for R/O */
					case 2 : write_buffer[0] &= ~(1 << 0); break;		/* reset bit 0 for DIR */
					case 3 : write_buffer[0] |=  (1 << 0); break;		/*   set bit 0 for SYS */
				}
				/* ####################### set attr of dir entry here ########################### */
				if (verbose) {printf ("Setting %s attribute (%d) for dir entry #%02d\n", attr, write_buffer[0], devildir_mem[i*256]);}
//								       printf ("sector start = %d\n", track_length*(devildir_mem[i*256]/36)+xlt2[devildir_mem[i*256]%36]*256+17);		/* testing */
				error_code = pwrite (diskimg_file_descrptr, (void *) write_buffer, 1, track_length*(devildir_mem[i*256]/36)+xlt2[devildir_mem[i*256]%36]*256+17);
				if (error_code == -1) {printf("devilfilesetattr() error: WRITE ERROR %d: %m\n", errno); close(diskimg_file_descrptr); return 2;}
				fsync (diskimg_file_descrptr);
				devildir0_mem[devildir_mem[i*256]*256+17] = write_buffer[0];	// copy new attr byte back to devildir0_mem[]
				devildir_mem[i*256+18] = write_buffer[0];			// copy new attr byte back to devildir_mem[]
				/* ####################################################################### */
				i++; /* if (leftover_size > 9216) {leftover_size -= 9216;} */
				sprintf(name1, "%c%c%c%c%c%c%c%c%c%c", \
							devildir_mem[i*256+2], devildir_mem[i*256+3], devildir_mem[i*256+4], devildir_mem[i*256+5], \
							devildir_mem[i*256+6], devildir_mem[i*256+7], devildir_mem[i*256+8], devildir_mem[i*256+9], \
							devildir_mem[i*256+10], devildir_mem[i*256+11]);
				memcpy(&params1[0], &devildir_mem[i*256+12], 6);
			}
			if (verbose) {printf ("---------------------------------------\n");}
			j++;
		}

		close(diskimg_file_descrptr);
	/*	diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("devilfilesetattr() error: diskimg_file_descrptr ERROR %d: %m\n", errno); exit(0);} */
		return 0;
	}

	// Interleave the contents of floppy image (".dimg") whose path is given by diskimg_dintl_file_name
	// and save it as a new floppy image (".img") in the same directory (if target_dir is an empty string)
	// or in the directory given by target_dir if not an empty string (target_dir must end in a "/").
	// The target directory should be accessible in R/W mode !
	int interleave(const char* dintl_file_name, const char* target_dir = "") {
		int diskimg_dintl_file_descrptr, diskimg_intl_file_descrptr, i, sect_start;
		char* diskimg_dintl_file_name;	// need this as copy of arg dintl_file_name, for stupid basename() who messes up its arg
		char* diskimg_intl_file_name;
		unsigned char write_buffer[256];

		// =========== allocate memory to	diskimg_dintl_file_name ===================================
		diskimg_dintl_file_name = (char*)malloc((strlen(dintl_file_name)+1)*sizeof(char));	// using malloc() + sprintf() instead of asprintf()
		sprintf(diskimg_dintl_file_name, "%s", dintl_file_name);				// to make fixing easier (1 statement instead of 2)

		// =========== allocate memory to	diskimg_intl_file_name ===================================
		if (strlen(target_dir)==0) {	// if "target_dir" is empty, use the same dir where "diskimg_dintl_file_name" is
			asprintf(&diskimg_intl_file_name, "%s", diskimg_dintl_file_name);
		} else {			// if "target_dir" is not empty, use it as target dir
			diskimg_intl_file_name = (char*)malloc( (strlen(target_dir)+strlen(basename(diskimg_dintl_file_name))+1)*sizeof(char) );
			sprintf(diskimg_dintl_file_name, dintl_file_name);	// fix what basename() screwed up
			sprintf(diskimg_intl_file_name, "%s", target_dir); strcat(diskimg_intl_file_name, basename(diskimg_dintl_file_name));
			sprintf(diskimg_dintl_file_name, dintl_file_name);	// fix what basename() screwed up
		}
		// =========== set ".img" as name extension for diskimg_intl_file_name ===========================
		if (strstr(diskimg_intl_file_name,".dimg")) {
			diskimg_intl_file_name[strstr(diskimg_intl_file_name,".dimg")-&diskimg_intl_file_name[0]]='\0'; strcat(diskimg_intl_file_name, ".img");
		} else {
			printf("'%s': Deinterleaved disk image file name must have the extension .dimg.\n", basename(diskimg_intl_file_name)); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 1;
		}
		printf("Interleaving sectors for image %s:\n", diskimg_dintl_file_name);
		// =========== open deinterleaved floppy image ===================================================
		diskimg_dintl_file_descrptr = open (diskimg_dintl_file_name, O_RDONLY); if (diskimg_dintl_file_descrptr == -1) {printf("Error %d opening deinterleaved disk image: %m\n", errno); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 2;}

		// =========== check if interleaved image file name doesn't already exist ========================
		if (xis_dir(diskimg_intl_file_name)!=-1) {printf("Error: file '%s' already exists. File creation aborted.\n", diskimg_intl_file_name); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 3;}

		// =========== create empty formatted floppy image for interleaved data to copy into =============
		if (verbose) {printf("Attempting to create file for interleaved disk image %s...\n", diskimg_intl_file_name);}
		umask(000);
		// =========== open interleaved floppy image =====================================================
		diskimg_intl_file_descrptr = open (diskimg_intl_file_name, O_RDWR | O_CREAT, 00666);
		if (diskimg_intl_file_descrptr == -1) {printf("Error %d creating interleaved disk image: %m\n", errno); fflush(stdout); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 4;}
		for (i=0; i<72*1024; i++) {write (diskimg_intl_file_descrptr, (void *) buffer, 10);}	/* writing 10 bytes at once */
		if (verbose) {printf("Disk image file %s successfully created\n", diskimg_intl_file_name);}

		// =========== copy sectors interleaved to target floppy image ===================================
		printf("\n-----------------------------------------------------------\nCopying sectors from %s to %s with interleave 2 ...\n", diskimg_dintl_file_name, diskimg_intl_file_name);
		printf(" Copying sector: 0000");
		for (i=0; i<2880; i++) { /* 80 tracks x 2 sides x 18 sectors = 2880 total sectors on a disk*/
			printf("\b\b\b\b%4d", i+1);
			error_code = pread (diskimg_dintl_file_descrptr, (void *) write_buffer, 256, i*256); if (error_code == -1) {printf("\nError %d reading deinterleaved floppy image: %m\n", errno); break;}
			sect_start = (i/18)*sstracklength + xlt[i%18]*256;
			error_code = pwrite (diskimg_intl_file_descrptr, (void *) write_buffer, 256, sect_start); if (error_code == -1) {printf("\nError %d writing to interleaved floppy image: %m\n", errno); break;}
		}
		printf("\n-----------------------------------------------------------\n\n");
		printf("Interleaved image saved as %s\n\n", diskimg_intl_file_name);

		// =========== deallocate memory from	diskimg_(d)intl_file_name ================================
		free(diskimg_intl_file_name); free(diskimg_dintl_file_name);

		close(diskimg_dintl_file_descrptr);
		close(diskimg_intl_file_descrptr);
		return 0;
	}

	// Deinterleave the floppy image currently open
	int deinterleave(const char* target_dir="") {
		int diskimg_dintl_file_descrptr, i, sect_start;
		unsigned char write_buffer[256];
		char* diskimg_dintl_file_name;
		char* diskimg_intl_file_name;	// need this as copy of diskimg_file_name, for stupid basename() who messes with its arg

		// =========== allocate memory to	diskimg_intl_file_name ===================================
		diskimg_intl_file_name = (char*)malloc((strlen(diskimg_file_name)+1)*sizeof(char));	// using malloc() + sprintf() instead of asprintf()
		sprintf(diskimg_intl_file_name, "%s", diskimg_file_name);				// to make fixing easier (1 statement instead of 2)

		// =========== allocate memory to	diskimg_dintl_file_name ==================================
		if (strlen(target_dir)==0) {	// if "target_dir" is empty, use the same dir where "diskimg_file_name" is
			diskimg_dintl_file_name = (char*)malloc((strlen(diskimg_file_name)+2)*sizeof(char));			// diskimg_file_name is a global var
			sprintf(diskimg_dintl_file_name, "%s", diskimg_file_name);
		} else {			// if "target_dir" is not empty, use it as target dir
			diskimg_dintl_file_name = (char*)malloc((strlen(target_dir)+strlen(basename(diskimg_intl_file_name))+2)*sizeof(char));
			sprintf(diskimg_intl_file_name, "%s", diskimg_file_name);	// fix what basename() screwed up
			sprintf(diskimg_dintl_file_name, "%s", target_dir); strcat(diskimg_dintl_file_name, basename(diskimg_intl_file_name));
			sprintf(diskimg_intl_file_name, "%s", diskimg_file_name);	// fix what basename() screwed up
		}
		// =========== set ".dimg" as name extension for diskimg_dintl_file_name =========================
		if (strstr(diskimg_dintl_file_name,".img")) {
			diskimg_dintl_file_name[strstr(diskimg_dintl_file_name,".img")-&diskimg_dintl_file_name[0]]='\0'; strcat(diskimg_dintl_file_name, ".dimg");
		} else {
			printf("'%s': Interleaved disk image file name must have the extension .img.\n", basename(diskimg_intl_file_name)); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 1;
		}

		printf("Deinterleaving sectors for image %s:\n", diskimg_file_name);
		// =========== open interleaved floppy image =====================================================
		diskimg_file_descrptr = open (diskimg_file_name, O_RDONLY); if (diskimg_file_descrptr == -1) {printf("Error %d opening interleaved disk image: %m\n", errno); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 2;}

		// =========== check if deinterleaved image file name doesn't already exist ======================
		if (xis_dir(diskimg_dintl_file_name)!=-1) {printf("Error: file '%s' already exists. File creation aborted.\n", diskimg_dintl_file_name); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 3;}

		// =========== create empty formatted floppy image for deinterleaved data to copy into ===========
		if (verbose) {printf("Attempting to create file for deinterleaved disk image %s...\n", diskimg_dintl_file_name);}
		umask(000);
		// =========== open deinterleaved floppy image ===================================================
		diskimg_dintl_file_descrptr = open (diskimg_dintl_file_name, O_RDWR | O_CREAT, 00666);
		if (diskimg_dintl_file_descrptr == -1) {printf("Error %d creating deinterleaved disk image: %m", errno); free(diskimg_intl_file_name); free(diskimg_dintl_file_name); return 4;}
		for (i=0; i<72*1024; i++) {write (diskimg_dintl_file_descrptr, (void *) buffer, 10);}	/* writing 10 bytes at once */
		if (verbose) {printf("Disk image file %s successfully created\n", diskimg_dintl_file_name);}

		// =========== copy sectors deinterleaved to target floppy image =================================
		printf("\n-----------------------------------------------------------\nCopying sectors from %s to %s without interleave ...\n", diskimg_file_name, diskimg_dintl_file_name);
		printf(" Copying sector: 0001");
		for (i=0; i<2880; i++) { /* 80 tracks x 2 sides x 18 sectors = 2880 total sectors on a disk*/
			printf("\b\b\b\b%4d", i+1);
			sect_start = (i/18)*sstracklength + xlt[i%18]*256;
			error_code = pread (diskimg_file_descrptr, (void *) write_buffer, 256, sect_start); if (error_code == -1) {printf("\ndeinterleave() error: WRITE ERROR %d: %m\n", errno); break;}
			error_code = pwrite (diskimg_dintl_file_descrptr, (void *) write_buffer, 256, i*256); if (error_code == -1) {printf("\ndeinterleave() error: WRITE ERROR %d: %m\n", errno); break;}
		}
		printf("\n-----------------------------------------------------------\n\n");
		printf("Deinterleaved image saved as %s\n\n", diskimg_dintl_file_name);

		// =========== deallocate memory from	diskimg_(d)intl_file_name ================================
		free(diskimg_intl_file_name); free(diskimg_dintl_file_name);

		close(diskimg_file_descrptr);										// diskimg_file_descrptr is a global var
		close(diskimg_dintl_file_descrptr);
		return 0;
	}

	/* ##################################################### PROCEDURE UNDER CONSTRUCTION ################################################################ */
	void diskcheck () {
		unsigned int dir_sect_start;
		unsigned short i, j;	/*, k, crt_block_number, crt_block_value; */
//		unsigned char alloc_blocks[77];		/* map of free and allocated blocks */
		unsigned char rec_check_map[77];
		unsigned char rb[17], rb1[17];		/* comparison buffer */
		unsigned char rb2[256];
		unsigned char warn=0;	/*, warn1=0; */
		int diskimg_file_descrptr;		/* disk image file descriptor */

		/* ################ open disk image file for writing ################### */
		diskimg_file_descrptr = open (diskimg_file_name, O_RDWR); if (diskimg_file_descrptr == -1) {printf("diskimg_file_descrptr ERROR %d: %m\n", errno);}

		/* ################ CHECKING FOR DUPLICATE DIRECTORY RECORDS ################# */
		warn=0;
		for (i = 0; i < 77; i++) {rec_check_map[i]=0;}
		printf("Checking for duplicate directory records ... ");
		for (i = 0; (i < nr_recs) && (rec_check_map[i]==0); i++) {
			memcpy(&rb[0], (&devildir_mem[0] + i*256 + 1), 17);
			for (j=i+1; (j<nr_recs) && (rec_check_map[j]==0); j++) {
				memcpy(&rb1[0], (&devildir_mem[0] + j*256 + 1), 17);
				if (memcmp(&rb[0], &rb1[0], 17) == 0) {rec_check_map[i]=1; rec_check_map[j]=1; printf("\nFound duplicate dir records #%d and #%d for file '%c%c%c%c%c%c%c%c%c%c'", i, j, devildir_mem[0+i*256+2], devildir_mem[0+i*256+3], devildir_mem[0+i*256+4], devildir_mem[0+i*256+5], devildir_mem[0+i*256+6], devildir_mem[0+i*256+7], devildir_mem[0+i*256+8], devildir_mem[0+i*256+9], devildir_mem[0+i*256+10], devildir_mem[0+i*256+11]); warn=1;}
			}
		}
		if (warn == 0) {printf("none found - OK!\n");} else {printf("\n");}

		/* ################ CHECKING FOR WRONG BYTES IN DIRECTORY RECORDS ################# */
		warn=0;
		printf("Checking for wrong bytes in directory records ... ");
		for (i = 0; i < 77; i++) {
			dir_sect_start = (i/18)*sstracklength + xlt[i%18]*256;
			lseek (diskimg_file_descrptr, dir_sect_start, SEEK_SET);
			read (diskimg_file_descrptr, (void *) rb2, 256);
			if ((rb2[0]>3) && (rb2[0]!=0xE4)) {printf("\nWrong flag byte in dir record #%d: %2X", i, rb2[0]); warn=1;}
			if ((rb2[0]<=3)) {for (j=18; j<256; j++) {if (rb2[j]!=0) {printf("\nNon-zero byte #%d in unused dir record #%d area.", j, i); warn=1; break;}}}
		}
		if (warn == 0) {printf("none found - OK!\n");} else {printf("\n");}

		/* ################ close disk image file ################### */
		close(diskimg_file_descrptr);
	}


}


// C++ headers
#include <QApplication>
#include <QFileInfo>
#include <QMainWindow>
//#include "ui_MainWindow.h"

#include "MainWindow/MainWindow.h"

void myMessageOutput(QtMsgType type, const char *msg) {
	/* do nothing, i.e. discard Qt messages completely */
}

int main( int argc, char ** argv ) {
	qInstallMsgHandler(myMessageOutput);
// ########################################## C section ################################################

	if (argc > 2) {

		int ret_code;
		long size = 0;					// size of filename argument provided
		char* diskimg_base_name;			// We use "diskimg_base_name" to avoid repeated calls to basename(DEVIL::diskimg_file_name)
		char* tap_base_name;				// We use "tap_base_name" to avoid repeated calls to basename(DEVIL::diskimg_file_name)

//		if (argc <= 2) {printf (DEVIL::USAGE); return 0;}	// print usage instructions if not enough args

		/* ########################## command line param #1 check - command to execute ############################## */
		if ( ! ((strcmp(argv[1],"-c") == 0) || 
			(strcmp(argv[1],"-g") == 0) || 
			(strcmp(argv[1],"-x") == 0) || 
			(strcmp(argv[1],"-l") == 0) || 
			(strcmp(argv[1],"-t") == 0) || 
			(strcmp(argv[1],"-w") == 0) || 
			(strcmp(argv[1],"-d") == 0) || 
			(strcmp(argv[1],"-k") == 0) || 
			(strcmp(argv[1],"-i") == 0) || 
			(strcmp(argv[1],"-di") == 0) || 
			(strcmp(argv[1],"-r0") == 0) || 
			(strcmp(argv[1],"-r1") == 0) || 
			(strcmp(argv[1],"-s0") == 0) || 
			(strcmp(argv[1],"-s1") == 0)) ) {printf ("Invalid command: %s\n", argv[1]); return 0;}

		if (strcmp(argv[1],"-c") == 0) {DEVIL::create_mode=1;}
		if (strcmp(argv[1],"-g") == 0) {DEVIL::defrag_mode=1;}
		if (strcmp(argv[1],"-l") == 0) {DEVIL::list_mode=1;}
		if (strcmp(argv[1],"-t") == 0) {DEVIL::tap_list_mode=1;}
		if (strcmp(argv[1],"-x") == 0) {DEVIL::extract_mode=1;}
		if (strcmp(argv[1],"-w") == 0) {DEVIL::write_mode=1;}
		if (strcmp(argv[1],"-d") == 0) {DEVIL::delete_mode=1;}
		if (strcmp(argv[1],"-k") == 0) {DEVIL::check_mode=1;}
		if (strcmp(argv[1],"-i") == 0) {DEVIL::interleave_mode=1;}
		if (strcmp(argv[1],"-di") == 0) {DEVIL::deinterleave_mode=1;}
		if ((strcmp(argv[1],"-r0") == 0) || (strcmp(argv[1],"-r1") == 0) || (strcmp(argv[1],"-s0") == 0) || (strcmp(argv[1],"-s1") == 0)) {DEVIL::set_attr_mode=1;}

		/* ########################## command line param #2 check - disk image file name ############################## */
		if  (DEVIL::create_mode ||
		     DEVIL::defrag_mode ||
		     DEVIL::list_mode ||
		     DEVIL::extract_mode ||
		    (DEVIL::write_mode && (DEVIL::xis_dir(argv[2])!=1)) ||
		     DEVIL::delete_mode ||
		     DEVIL::interleave_mode ||
		     DEVIL::deinterleave_mode ||
		     DEVIL::set_attr_mode ||
		     DEVIL::check_mode) {
			asprintf(&DEVIL::diskimg_file_name, "%s", argv[2]); DEVIL::diskimg_file_name_allocated=true;	// initialize "diskimg_file_name"
			asprintf(&diskimg_base_name, "%s", basename(DEVIL::diskimg_file_name));				// We use "diskimg_base_name" to avoid repeated calls to basename()
			free(DEVIL::diskimg_file_name); asprintf(&DEVIL::diskimg_file_name, "%s", argv[2]);		// Fix mess after call to basename()

			if (DEVIL::dotcount(diskimg_base_name) > 1) {printf("'%s': No multiple dots allowed in disk image file name.\n", diskimg_base_name); free(DEVIL::diskimg_file_name); free(diskimg_base_name); return 0;}
			if (DEVIL::dotcount(diskimg_base_name) == 0) {printf("'%s': Disk image file name must have the extension .img.\n", diskimg_base_name); free(DEVIL::diskimg_file_name); free(diskimg_base_name); return 0;}
			if ( (DEVIL::dotcount(diskimg_base_name) == 1U) && (!DEVIL::deinterleave_mode) &&
			    !(strstr(diskimg_base_name, ".img") && (DEVIL::dotpos(diskimg_base_name) == strlen(diskimg_base_name)-4U)) ) {
				printf("'%s': Disk image file name must have the extension .img.\n", diskimg_base_name);
				free(DEVIL::diskimg_file_name); free(diskimg_base_name); return 0;
			}
			if ( (DEVIL::dotcount(diskimg_base_name) == 1U) && (DEVIL::interleave_mode) &&
			    !(strstr(diskimg_base_name, ".dimg") && (DEVIL::dotpos(diskimg_base_name) == strlen(diskimg_base_name)-5U)) ) {
				printf("'%s': Disk image file name must have the extension .dimg.\n", diskimg_base_name);
				free(DEVIL::diskimg_file_name); free(diskimg_base_name); return 0;
			}
			if (!DEVIL::create_mode) {
				DEVIL::test_stream = fopen (DEVIL::diskimg_file_name, "r"); if (DEVIL::test_stream == NULL) {printf ("Error %d opening disk image file: %m\n", errno); free(DEVIL::diskimg_file_name); free(diskimg_base_name); return 0;}
				stat(DEVIL::diskimg_file_name, &DEVIL::st); size = DEVIL::st.st_size;
				fclose(DEVIL::test_stream);
				if (size != 737280) {printf ("Invalid Devil disk image file specified - file size = %ld\n", size); free(DEVIL::diskimg_file_name); free(diskimg_base_name); return 0;}
			}
		}

		/* ########################## command line param #2 check - .TAP file name ############################## */
		if  (DEVIL::tap_list_mode) {
			asprintf(&DEVIL::tap_file_name, "%s", argv[2]); DEVIL::tap_file_name_allocated=true;	// initialize "tap_file_name"
			asprintf(&tap_base_name, "%s", basename(DEVIL::tap_file_name));				// We use "tap_base_name" to avoid repeated calls to basename()
			free(DEVIL::tap_file_name); asprintf(&DEVIL::tap_file_name, "%s", argv[2]);		// Fix mess after call to basename()

			if (DEVIL::dotcount(tap_base_name) > 1) {printf("'%s': No multiple dots allowed in .TAP file name.\n", tap_base_name); free(DEVIL::tap_file_name); free(tap_base_name); return 0;}
			if (DEVIL::dotcount(tap_base_name) == 0) {printf("'%s': .TAP file name must have the extension .tap.\n", tap_base_name); free(DEVIL::tap_file_name); free(tap_base_name); return 0;}
			if ( (DEVIL::dotcount(tap_base_name) == 1U) &&
			    !(strstr(tap_base_name, ".tap") && (DEVIL::dotpos(tap_base_name) == strlen(tap_base_name)-4U)) ) {
				printf("'%s': .TAP file name must have the extension .tap.\n", tap_base_name);
				free(DEVIL::tap_file_name); free(tap_base_name); return 0;
			}
			DEVIL::test_stream = fopen (DEVIL::tap_file_name, "r"); if (DEVIL::test_stream == NULL) {printf ("Error %d opening .TAP image file: %m\n", errno); free(DEVIL::tap_file_name); free(tap_base_name); return 0;}
			fclose(DEVIL::test_stream);
		}

		/* ######################### create empty formatted disk image:   executable -c diskimg_file_name ######################### */
		if (DEVIL::create_mode) {
			if (argc > 3) {printf ("Too many parameters after the -c command\n"); return 0;}
			printf("Attempting to create disk image %s...\n", argv[2]);
			if (DEVIL::xis_dir(argv[2])!=-1) {printf("Error: file '%s' already exists. Disk image creation aborted.\n", argv[2]); return 0;}

//			umask(000);
//			DEVIL::diskimg_file_name = argv[2];		/* initialize diskimg_file_name */
//			DEVIL::diskimg_file_descrptr = open (DEVIL::diskimg_file_name, O_RDWR | O_CREAT, 00666);
//			if (DEVIL::diskimg_file_descrptr == -1) {printf("File creation error %d: %m", errno); return 0;}
//			for (int i=0; i<72*1024; i++) {write (DEVIL::diskimg_file_descrptr, (void *) DEVIL::buffer, 10);}	/* writing 10 bytes at once */
//			close(DEVIL::diskimg_file_descrptr);

			if (DEVIL::create_diskimg(argv[2])) {
				printf("Error creating floppy image %s\n", argv[2]);
			} else {
				printf("Disk image %s successfully created\n", argv[2]);
			}

			return 0;
		}

		/* ############ list command syntax check:   executable -l diskimg_file_name [-deb]   ############# */
		if (DEVIL::list_mode) {	/* cmd line syntax check for list command */
			if (argc == 4) {
				if (strcmp(argv[3], "-deb") != 0) {printf ("Invalid option: %s\n", argv[3]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			if (argc > 4) {printf ("Too many parameters after the -l command\n"); return 0;}
		}

		/* ############ defragment command syntax check:   executable -g diskimg_file_name [-ver [-deb]]   ############# */
		if (DEVIL::defrag_mode) {	/* cmd line syntax check for defragment command */
			if (argc >= 4) {	/* check -ver syntax */
				if (strcmp (argv[3], "-ver") != 0) {printf("Invalid option: %s\n", argv[3]); return 0;}
				else {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc == 5) {	/* check -deb syntax */
				if (strcmp (argv[4], "-deb") != 0) {printf ("Invalid option: %s\n", argv[4]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			if (argc > 5) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ extract file command syntax check:   executable -x diskimg_file_name range [-log|-ver [-deb]]   ############# */
		if (DEVIL::extract_mode) {	/* cmd line syntax check for extract file command */
			if (argc == 3) {printf ("Please specify a file or range of files you want to extract (ex. 5 or 21-34)\n"); return 0;}
			if (argc >= 4) {	/* check range syntax */
				if (DEVIL::range_check(argv[3])) {return 0;}
//				printf("\nExtract command syntax check: Catalog entries to process: %s to %s (%d to %d)\n", range1_c, range2_c, range1, range2);
//				printf("\nExtract command syntax check: Catalog entries to process: %d to %d\n", DEVIL::range1, DEVIL::range2);
			}
			if (argc >= 5) {	/* check -log|-ver syntax */
				if ((strcmp (argv[4], "-log") != 0) && (strcmp (argv[4], "-ver") != 0)) {printf("\nInvalid option: %s\n", argv[4]); return 0;}
				if (strcmp (argv[4], "-log") == 0) {printf ("Operation log will be generated for each file\n"); DEVIL::log_on = 1;}
				if (strcmp (argv[4], "-ver") == 0) {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc == 6) {	/* check -deb syntax */
				if (strcmp (argv[5], "-deb") != 0) {printf ("\nInvalid option: %s\n", argv[5]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			if (argc > 6) {
				printf("\nToo many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ write command syntax check: executable -w diskimg_file_name path/to/tap_file [-ver [-deb]] ############# */
//		if (write_mode && (xis_dir(argv[2])==1)) {
		if (DEVIL::write_mode && (access ( argv[3], F_OK ) == 0)) {
			if (argc >= 5) {
				if (strcmp (argv[4], "-ver") != 0) {printf ("Invalid option: %s\n", argv[4]); return 0;}
				else {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc == 6) {
				if (strcmp (argv[5], "-deb") != 0) {printf ("Invalid option: %s\n", argv[5]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			
			if (argc > 6) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ delete file command syntax check:   executable -d diskimg_file_name range [-ver [-deb]]   ############# */
		if (DEVIL::delete_mode) {	/* cmd line syntax check for delete file command */
			if (argc == 3) {printf ("Please specify a file or range of files you want to delete (ex. 5 or 21-34)\n"); return 0;}
			if (argc >= 4) {	/* check range syntax */
				if (DEVIL::range_check(argv[3])) {return 0;}
//				printf("\nDelete command syntax check: Catalog entries to process: %d to %d\n", DEVIL::range1, DEVIL::range2);
			}
			if (argc >= 5) {	/* check -ver syntax */
				if (strcmp (argv[4], "-ver") != 0) {printf("Invalid option: %s\n", argv[4]); return 0;}
				else {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc == 6) {	/* check -deb syntax */
				if (strcmp (argv[5], "-deb") != 0) {printf ("Invalid option: %s\n", argv[5]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			if (argc > 6) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ set attr command syntax check:   executable -s0|-s1 diskimg_file_name range [-ver [-deb]]   ############# */
		if (DEVIL::set_attr_mode) {	/* cmd line syntax check for set attr command */
			if (argc == 3) {printf ("Please specify a file or range of files you want to delete (ex. 5 or 21-34)\n"); return 0;}
			if (argc >= 4) {	/* check range syntax */
				if (DEVIL::range_check(argv[3])) {return 0;}
//				printf("\nSet attr command syntax check: Catalog entries to process: %d to %d\n", DEVIL::range1, DEVIL::range2);
			}
			if (argc >= 5) {	/* check -ver syntax */
				if (strcmp (argv[4], "-ver") != 0) {printf("Invalid option: %s\n", argv[4]); return 0;}
				else {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc == 6) {	/* check -deb syntax */
				if (strcmp (argv[5], "-deb") != 0) {printf ("Invalid option: %s\n", argv[5]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			if (argc > 6) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ interleave/deinterleave command syntax check:   executable -i diskimg_file_name [-ver]   ############# */
		if (DEVIL::interleave_mode || DEVIL::deinterleave_mode) {	/* cmd line syntax check for interleave/deinterleave command */
			if (argc == 4) {	/* check -ver syntax */
				if (strcmp (argv[3], "-ver") != 0) {printf("Invalid option: %s\n", argv[3]); return 0;}
				else {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc > 4) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ error check command syntax check:   executable -k diskimg_file_name range [-deb]   ############# */
		if (DEVIL::check_mode) {	/* cmd line syntax check for extract file command */
			if (argc == 4) {	/* check -deb syntax */
				if (strcmp (argv[3], "-deb") != 0) {printf("Invalid option: %s\n", argv[3]); return 0;}
				else {printf ("Debug info on.\n"); DEVIL::debug = 1;}
			}
			if (argc > 4) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ .TAP list command syntax check:   executable -t tap_file_name   ############# */
		if (DEVIL::tap_list_mode) {	/* cmd line syntax check for .TAP list command */
			if (argc == 4) {	/* check -ver syntax */
				if (strcmp (argv[3], "-ver") != 0) {printf("Invalid option: %s\n", argv[3]); return 0;}
				else {printf ("Verbose operation.\n"); DEVIL::verbose = 1;}
			}
			if (argc > 4) {
				printf("Too many arguments, exiting.\n"); return 0;
			}
		}

		/* ############ List contents of floppy image or .TAP file in console ####################### */
		if (!DEVIL::interleave_mode) {
			if (!DEVIL::tap_list_mode) {		// if operating on floppy image, list image contents
				ret_code = DEVIL::devildirdump();
				switch (ret_code) {
					case 1: return 0;
					case 2: return 0;
					case 4: return 0;
					case 5: return 0;
					case 6: return 0;
				}
			} else {				// if operating on .TAP file, list .TAP contents
				ret_code = DEVIL::devilfilewrite(argv[2], 0, 0);
				switch (ret_code) {
					case -1: return 0;
					case -2: return 0;
				}
			}
		} else {
			printf("\n");
		}

		if (DEVIL::extract_mode && DEVIL::debug)  {printf("Catalog entries to extract: %s to %s (%d to %d)\n", DEVIL::range1_c, DEVIL::range2_c, DEVIL::range1, DEVIL::range2);}
		if (DEVIL::delete_mode && DEVIL::debug)   {printf("Catalog entries to delete: %s to %s (%d to %d)\n", DEVIL::range1_c, DEVIL::range2_c, DEVIL::range1, DEVIL::range2);}
		if (DEVIL::set_attr_mode && DEVIL::debug) {printf("Catalog entries to set attr: %s to %s (%d to %d)\n", DEVIL::range1_c, DEVIL::range2_c, DEVIL::range1, DEVIL::range2);}


		/* ##################################################################################################################### */
		/* ########################################### DEFRAGMENT COMMAND ###################################################### */

		if (DEVIL::defrag_mode && (argc >= 3)) {
			DEVIL::defrag_disk();
		}

		/* ##################################################################################################################### */
		/* ########################################### EXTRACTION MODE COMMANDS ################################################ */

		if (DEVIL::extract_mode && (argc >= 4)) {	// if minimum arguments required exist
			DEVIL::path_to_dest_dir_base_allocated=true;
			DEVIL::path_to_dest_dir_base=(char*)malloc(2);		// allocate "path_to_dest_dir_base" ...
			asprintf(&DEVIL::path_to_dest_dir_base,".");		// ... and set it to current directory (.)
			DEVIL::diskimg_dir ();					// set floppy image base saving dir path (in "dest_dir_base")
			if (DEVIL::xis_dir(DEVIL::dest_dir_base) == -2) {printf ("Error: '%s' already exists but is not a directory\n", DEVIL::dest_dir_base); return 0;}
			DEVIL::devilfilecopy(DEVIL::range1, DEVIL::range2, "", 1);
		}

		/* ##################################################################################################################### */
		/* ########################################### WRITE MODE COMMANDS ##################################################### */

		if (DEVIL::write_mode && (argc >= 3)) {	/* if minimum arguments required exist */
			DEVIL::blocks_needed = DEVIL::devilfilewrite(argv[3], 0, 0);
			if (DEVIL::blocks_needed == -1) {printf("TAP file too short, exiting.\n"); return 0;} else
			if (DEVIL::blocks_needed == -2) {printf("Error opening TAP file, exiting.\n"); return 0;} else
			if (DEVIL::blocks_needed == -3) {printf("Headerless and dataless blocks found in TAP file, exiting.\n"); return 0;} else
			if (DEVIL::blocks_needed == -4) {printf("Headerless blocks found in TAP file, exiting.\n"); return 0;} else
			if (DEVIL::blocks_needed == -5) {printf("Dataless headers found in TAP file, exiting.\n"); return 0;} else
			if (DEVIL::blocks_needed > 0) {
				printf("Total tracks required: %d\n-------------------------\n\n", DEVIL::blocks_needed);
				int i=0, j=0, k;
				while (i<77) {
					while ((i<77) && (DEVIL::alloc_dir[i]==1)) {i++;}
					if (i==76) {printf("Disk image full, writing aborted.\n"); return 0;}
					else {
						k=i;
						while ((i<77) && (DEVIL::alloc_dir[i]==0) && (j < DEVIL::blocks_needed)) {i++; j++;}
						if (j == DEVIL::blocks_needed) {
							printf("Found enough continuous free space at block %d (track %d)\n\n", k, k+3);
							DEVIL::devilfilewrite(argv[3], 1, k); return 0;
						} else {j=0;}
					}
				}
				printf ("Not enough continuous free space available on disk, writing aborted.\n"); return 0;
			}
		}

		/* ################################################################################################################ */
		/* ########################################### DELETE COMMAND ##################################################### */

		if (DEVIL::delete_mode && (argc >= 4)) {	/* if minimum arguments required exist */
			DEVIL::devilfiledelete(DEVIL::range1, DEVIL::range2);
		}

		/* ################################################################################################################ */
		/* ########################################### INTERLEAVE COMMAND ################################################# */
		
		if (DEVIL::interleave_mode && (argc >= 3)) {
			DEVIL::interleave(DEVIL::diskimg_file_name);	// save interleaved image in same dir as argv[2]
		}

		/* ################################################################################################################ */
		/* ########################################### DEINTERLEAVE COMMAND ############################################### */
		
		if (DEVIL::deinterleave_mode && (argc >= 3)) {
			DEVIL::deinterleave();				// save deinterleaved image in same dir as argv[2]
		}

		/* ################################################################################################################ */
		/* ########################################### SET ATTR COMMAND ################################################### */
	
		if (DEVIL::set_attr_mode && (argc >= 4)) {	/* if minimum arguments required exist */
			DEVIL::devilfilesetattr(DEVIL::range1, DEVIL::range2, argv[1]);
		}

		/* ################################################################################################################ */
		/* ########################################### CHECK COMMAND ###################################################### */

		if (DEVIL::check_mode && (argc >= 3)) {	/* if correct nr of arguments required */
			DEVIL::diskcheck();
		}

		return(0);


	} else {	/* print usage instructions if less than 2 args passed and start GUI */

		printf (DEVIL::USAGE);

	}

// ######################### Qt section - only executed if less than 2 arguments passed on command line or program launched by mouse click #############################

	getcwd ( DEVIL::exe_full_path, FILENAME_MAX );				// if run from command line with relative path and less than 2 args
	strcat(DEVIL::exe_full_path,"/"); strcat(DEVIL::exe_full_path,argv[0]);	// we determine absolute path and store it in DEVIL::exe_full_path
//	printf ("exe_full_path = %s\n", DEVIL::exe_full_path);

	QFileInfo fi (QString(DEVIL::exe_full_path));

	QApplication a( argc, argv );
	if (! fi.exists()) {	// if run by mouse click, getcwd will always return home dir, so replace with QApplication::argv()[0]
		sprintf(DEVIL::exe_full_path, "%s", a.argv()[0]);
		printf ("replacing exe_full_path with %s\n", a.argv()[0]);
	}
	MainWindow w; /* QMainWindow maindlg(0); w.setupUi(&maindlg);
	maindlg.show(); */
	w.show();

	a.connect( &a, SIGNAL( lastWindowClosed() ), &a, SLOT( quit() ) );
	return a.exec();


}
